c
c-----------------------------------------------------------------------
c
c     Does one-step Newton Point-GS relaxation and returns norm of
C     running ('instantaneous') residuals---checkerboard ordering.
c
c     Routine (as well as RELAXL and LOP) is also designed to have
c     functions G, DG passed to it for the purposes of solving
c
c     Laplacian(u) + G(u) = rhs
c
c     In this version, G(u) = 0 is hard wired, for vectorization
c     purposes, in all routines.  In any case, modify code RRESL = ...,
c     JACEL = ..., appropriately when changing G. (JACEL --> diagonal
c     Jacobian element --> d(RRESL)/d U(I,J)). Also make corresponding
c     changes in RELAXL, LOP and CLCRHS.
c
c-----------------------------------------------------------------------
c
      real*8 function relaxc(rres,u,rhs,mask,nx,ny,nz,h,
     *                       q,qx,qy,x,y,z,omega)
! here qx is the  x derivative of q, and qy is the y derivative of q
c
         implicit       none
c

c
c
         integer        nx,              ny,            nz
         real*8          rres(nx,ny,nz),  u(nx,ny,nz),
     *                  rhs(nx,ny,nz),mask(nx,ny,nz)
         real*8     x(nx),y(ny),z(nz)
         real*8          h
c
         real*8          hm2,      m6hm2,    jacel,    rresl, hi, hi2
         real*8          omega
         real*8          q(nx,ny,nz),qx(nx,ny,nz),qy(nx,ny,nz)
         real*8         r1,r2,a1,a2
         real*8         cxx,cyy,czz,cx,cy,cxy
         integer        i,        j,        k,
     *                  isw,      ksw,      pass
      do i =1,nx
      do j =1,ny
      do k =1,nz
      rres(i,j,k) =0.0d0
      end do
      end do
      end do
c
C         call d3lbs(rres,0.0d0,nx,ny,nz)

         relaxc = 0.0d0
         hm2 = 1.0d0 / (h * h)
         hi = 0.5d0 /  h
         hi2= hi * hi 
c
         ksw = 2
         do pass = 1 , 2
            isw = ksw
            do k = 2 , nz - 1
               do j = 2 , ny - 1
                  do i = isw + 1 , nx - 1, 2

                     a1 = exp(-2*q(i,j,k))
                     cxx = ( x(i)**2*a1+y(j)**2 ) 
                     cyy = ( y(j)**2*a1+x(i)**2 ) 
                     czz = ( y(j)**2+x(i)**2 ) * a1
                     cx  = ( x(i)*(a1-1)+2*y(j)*(y(j)*qx(i,j,k) -
     *                                           x(i)*qy(i,j,k) )  )
                     cy  = ( y(j)*(a1-1)+2*x(i)*(x(i)*qy(i,j,k) -
     *                                           y(j)*qx(i,j,k) )  )
                     cxy = (2*x(i)*y(j)*(a1-1) ) 

! now we just finite difference ..... (try with no self adjoint form.....)

!                     rresl =cxx*hm2*(u(i+1,j,k)+u(i-1,j,k)-2*u(i,j,k)) +
!     *                      cyy*hm2*(u(i,j+1,k)+u(i,j-1,k)-2*u(i,j,k)) +
!     *                      czz*hm2*(u(i,j,k+1)+u(i,j,k-1)-2*u(i,j,k)) +
!     *                      cx *hi*( u(i+1,j,k) - u(i-1,j,k) ) +
!     *                      cy *hi*( u(i,j+1,k) - u(i,j-1,k) ) +
!     *                      cxy*hi2*( u(i+1,j+1,k) + u(i-1,j-1,k) - 
!     *                                u(i+1,j-1,k) - u(i-1,j+1,k) )  -
!     *                      rhs(i,j,k)
!                     jacel = -2*cxx*hm2-2*cyy*hm2-2*czz*hm2

! make this simple for a scharwschild hole at first .....

							 rresl = mask(i,j,k) * (
     *                       hm2 * (u(i+1,j,k) + u(i-1,j,k) +
     *                              u(i,j+1,k) + u(i,j-1,k) +
     *                              u(i,j,k+1) + u(i,j,k-1) -
     *                              6.0d0 * u(i,j,k)) -
     *                       rhs(i,j,k)  )

							jacel = -6.0d0 * hm2


                     rresl = mask(i,j,k)*rresl
                     jacel = (1.0d0-mask(i,j,k)) + jacel*mask(i,j,k)
							if (jacel.eq.(0.0d0)) then 
								rresl = 0.0d0
								jacel = 1.0d0
							end if

                     u(i,j,k)=omega*(u(i,j,k)-rresl/jacel)+ 
     *                        (1-omega)*u(i,j,k)

                     rres(i,j,k) = rresl
                     relaxc = relaxc + rresl * rresl
                  end do
                  isw = 3 - isw
               end do
            end do
            ksw = 3 - ksw
      end do
      relaxc = sqrt(relaxc / (nx * ny * nz))
!		write(6,*) relaxc
23		format(3i3,9e12.2)

      return
      end
