!-----------------------------------------------------------------------
!
!  Creation date: Wed Jul 13 09:39:26 CDT 1994
!   Written by Reid Guenther and Mijan Huq.
!   Ref: Berger and Rigoutsos, IEEE Trans. Vol. 21, No. 5, 1991.
!
!
! Routine to carry out Clustering in 3d ad using 
!
!
! NX/Y/Z   := Size of the box.
! box_max  := Maximum number of boxes.
! min_eff  := Threshold value for efficiency. (Real*8)
! box     := 3D integer Array containing characteristic 
!            function. Size NX*NY*NZ.
! box_stk := Integer Stack containing box characterstics. Size 6*box_max
!             box_stk(1,i) = x coordinate of lower most corner of box i
!             box_stk(2,i) = y coordinate of lower most corner of box i
!             box_stk(3,i) = z coordinate of lower most corner of box i
!             box_stk(4,i) = x coordinate of upper most corner of box i
!             box_stk(5,i) = y coordinate of upper most corner of box i
!             box_stk(6,i) = z coordinate of upper most corner of box i
! wksp     := 1D integer Work Array. Size 2*NX + 2*NY + 2*NZ
!
!-----------------------------------------------------------------------
!
! this routine is a DAGH wrapper for the cluster wrap routine below.
!
!-----------------------------------------------------------------------
      subroutine dagh_cls_3d(box, lb, ub, shape, blk_wid, buf_wid,    
     &                    box_max, box_no, box_stk_lb, box_stk_ub,    
     &                    min_eff)

      implicit none

! input variables
         integer lb(3),  ub(3), shape(3)
         integer box(shape(1),shape(2),shape(3))
         integer blk_wid, buf_wid, box_max
         real*8  min_eff

! output variables
         integer box_no, box_stk_lb(3,box_max), box_stk_ub(3,box_max)

! routine variables
         integer clstr, nx, ny, nz, box_stk_max
         parameter (box_stk_max = 100)

         integer box_stk(6,100)
         integer nmax_wksp
         parameter (nmax_wksp = 5000)
         integer wksp(nmax_wksp)

         integer stride_x, stride_y, stride_z

         integer box_pt, i, j, k


         nx = shape(1)
         ny = shape(2)
         nz = shape(3)
         if (6*max(nx,ny,nz).gt.nmax_wksp) then
            write(8,*)'DAGH_CLS>>> Warning 6*max(shape(*)) is ',     
     &                ' greater than nmax_wksp ',nmax_wksp
            write(8,*)'DAGH_CLS>>> shape(1) ',shape(1)
            write(8,*)'DAGH_CLS>>> shape(2) ',shape(2)
            write(8,*)'DAGH_CLS>>> shape(3) ',shape(3)
            box_no = 0
            return
         end if

         if (box_max.gt.box_stk_max) then
            write(8,*)'DAGH_CLS>>> Warning box_stk ',box_stk,      
     &                ' greater than box_stk_max ',box_stk_max
            return
         end if

!
! first run through, let blk_wid = 1, but try other blk_wid, too!

! Initialize box_stk
			do i = 1 , 6
				do j = 1 , 100
               box_stk(i,j)=0
				end do
			end do

         box_no = clstr(nx,ny,nz,box_max,blk_wid,                   
     &                  box,box_stk,wksp,min_eff)

      goto 333

!
!   reinitiate box with buf_wid added.
         do i = 1, nx
            do j = 1, ny
               do k = 1, nz
                  box(i,j,k) = 0
               end do
            end do
         end do

         do box_pt = 1, box_no
            do i = max(1,box_stk(1,box_pt)-buf_wid),              
     &             min(nx,box_stk(4,box_pt)+buf_wid)
               do j = max(1,box_stk(2,box_pt)-buf_wid),            
     &                min(nx,box_stk(5,box_pt)+buf_wid)
                  do k = max(1,box_stk(3,box_pt)-buf_wid),        
     &                   min(nx,box_stk(6,box_pt)+buf_wid)
                     box(i,j,k) = 1
                  end do
               end do
            end do
         end do

         box_no = clstr(nx,ny,nz,box_max,blk_wid,                  
     &                  box,box_stk,wksp,min_eff)

 333  continue

!
!   now define box_stk_lb and box_stk_ub in terms of global coordinates
!   defined by lb and ub and shape
         stride_x = (ub(1) - lb(1))/(shape(1) - 1)
         stride_y = (ub(2) - lb(2))/(shape(2) - 1)
         stride_z = (ub(3) - lb(3))/(shape(3) - 1)

     
         do box_pt = 1, box_no
            box_stk_lb(1,box_pt) = lb(1) +                           
     &                            (box_stk(1,box_pt)-1)*stride_x
            box_stk_lb(2,box_pt) = lb(2) +                          
     &                            (box_stk(2,box_pt)-1)*stride_y
            box_stk_lb(3,box_pt) = lb(3) +                           
     &                            (box_stk(3,box_pt)-1)*stride_z

            box_stk_ub(1,box_pt) = lb(1) +                        
     &                            (box_stk(4,box_pt)-1)*stride_x
            box_stk_ub(2,box_pt) = lb(2) +                         
     &                            (box_stk(5,box_pt)-1)*stride_y
            box_stk_ub(3,box_pt) = lb(3) +                           
     &                            (box_stk(6,box_pt)-1)*stride_z
         end do


      return
      end

!-----------------------------------------------------------------------
!
!  this routine is a wrapper for the clustering algorithm and sets
!   up fortran workspace.
!
!-----------------------------------------------------------------------
      integer function clstr(nx,ny,nz,box_max,blk_wd,                
     &                      box,box_stk,wksp,min_eff)

         implicit none

         integer cmp_clstr

         integer  nx, ny, nz, box_max, blk_wd
         integer  box_stk(6,box_max), wksp(2*nx+2*ny+2*nz)
         real*8   min_eff
         integer  box(nx,ny,nz)
         integer  i,j

         logical trace
         parameter (trace=.false.)

!
! Memory allocation of : sigx, sigy, lplx, lply.
!
         integer sigx, sigy, sigz, lplx, lply, lplz

         sigx = 1
         sigy = nx + 1
         sigz = nx + ny + 1
         lplx = nx + ny + nz + 1
         lply = 2*nx + ny + nz + 1
         lplz = 2*nx + 2*ny + nz + 1


         if (trace) then
            write(8,*)'nx =',nx
            write(8,*)'ny =',ny
            write(8,*)'nz =',nz
            write(8,*)'box_max =',box_max
            write(8,*)'blk_wd =',blk_wd
            write(8,*)'sigx =',sigx
            write(8,*)'sigy =',sigy
            write(8,*)'sigz =',sigz
            write(8,*)'lplx =',lplx
            write(8,*)'lply =',lply
            write(8,*)'lplz =',lplz
            write(8,*)'min_eff =',min_eff
            write(8,*)'box(1,1,1):',box(1,1,1)
            write(8,*)'box(11,1,1):',box(11,1,1)
            write(8,*)'box(4,7,9):',box(4,7,9)
            write(8,*)'box(6,10,12):',box(6,10,12)
            write(8,*)'box(3,6,4):',box(3,6,4)
            write(8,*)'box(2,13,5):',box(2,13,5)
            write(8,*)'box(7,7,7):',box(7,7,7)
            write(8,*)'box(1,11,1):',box(1,11,1)
            write(8,*)'box(1,1,11):',box(1,1,11)
            write(8,*)'box(nx,ny,nz):',box(nx,ny,nz)
         end if

!
!  Call the cluster function which returns the number of boxs found.
!
         clstr = cmp_clstr(nx,ny,nz,box_max,blk_wd, box, box_stk,   
     &              wksp(sigx),wksp(sigy),wksp(sigz),              
     &              wksp(lplx),wksp(lply),wksp(lplz),min_eff)

         return

      end



!-----------------------------------------------------------------------
!
! Internal cluster routine to compute boxs as described by the
!  algorithm given in the Berger & Rigoutsos paper (IEEE Transactions on
!  systems, man, and cybernetics. VOL 21, NO. 5, Sept/Oct 1991).
!
!-----------------------------------------------------------------------
      integer function cmp_clstr(nx,ny,nz,box_max,blk_wd,box,box_stk,   &
     &                     sigx,sigy,sigz,lplx, lply,lplz, min_eff)

         implicit none

         integer  nx, ny, nz, box_max, blk_wd
         integer  box_stk(6,box_max)
         integer  sigx(nx),sigy(ny),sigz(nz),lplx(nx),lply(ny),lplz(nz)
         real*8   min_eff
         integer  box(nx,ny,nz)

!
! local variables
!
         integer  i, flag, no_box, j, ind, x_wd, y_wd, z_wd, k
         integer  l,m,n
         real*8   box_eff
         real*8   cur_box_eff
         character*20 fname

         logical trace, trace_gr
         parameter (trace=.false.,trace_gr = .false.)


         call InitStack(box_stk, 1, 1, 1, nx, ny, nz, no_box, box_max)
         i=1
 100     continue
            cur_box_eff = box_eff(box, box_stk(1,i), nx, ny, nz)
            if (trace) then
               write(8,*)'after computing cur_box_eff'
               write(8,*)'i =',i
               write(8,*)'cur_box_eff ',cur_box_eff
               write(8,*)'min_eff ',min_eff
               write(8,*)'box_stk(1,i) ',box_stk(1,i)
               write(8,*)'box_stk(2,i) ',box_stk(2,i)
               write(8,*)'box_stk(3,i) ',box_stk(3,i)
               write(8,*)'box_stk(4,i) ',box_stk(4,i)
               write(8,*)'box_stk(5,i) ',box_stk(5,i)
               write(8,*)'box_stk(6,i) ',box_stk(6,i)
               write(8,*)'box(1,1,1):',box(1,1,1)
               write(8,*)'box(5,5,5):',box(5,5,5)
               write(8,*)'box(11,1,1):',box(11,1,1)
               write(8,*)'box(1,11,1):',box(1,11,1)
               write(8,*)'box(1,1,11):',box(1,1,11)
               write(8,*)'box(nx,ny,nz):',box(nx,ny,nz)
            end if
!
!  If the efficiency of the current box is less than the
!   threshold try to improve by splitting.
!
            if (cur_box_eff.lt.min_eff) then

               call cmp_sig_lpl(box,sigx,sigy,sigz,                
     &                      lplx,lply,lplz,box_stk(1,i),nx,ny,nz,i)

               if (trace) then
                  write(8,*)'i =',i
                  write(8,*)'cur_box_eff =',cur_box_eff
                  write(8,*)'nx_lo =',box_stk(1,i)
                  write(8,*)'ny_lo =',box_stk(2,i)
                  write(8,*)'nz_lo =',box_stk(3,i)
                  write(8,*)'nx_hi =',box_stk(4,i)
                  write(8,*)'ny_hi =',box_stk(5,i)
                  write(8,*)'nz_hi =',box_stk(6,i)
                  do k = box_stk(1,i),box_stk(4,i)
                     write(8,10) 'sigx(',k,')=',sigx(k),          
     &                          ' lplx(',k,')=',lplx(k)
                  enddo
                  do j = box_stk(2,i),box_stk(5,i)
                     write(8,10) 'sigy(',j,')=',sigy(j),           
     &                          ' lply(',j,')=',lply(j)
                  end do
               end if

               flag = 0
               call find_split(nx,ny,nz,no_box,i,blk_wd,         
     &                        box_stk,                           
     &                        sigx, sigy, sigz, lplx, lply,       
     &                        lplz, flag, box_max)

               if (flag.eq.0) then
                   i = i + 1
               end if

            else
               i = i + 1
            end if


!
!  keep trying to divide the inefficient boxes until done or
!   run out of space in the stack
!
         if (i .le. no_box .and. no_box.lt.box_max) then
            goto 100
         end if

         cmp_clstr = no_box

!
! this is needed to conform to Manish Parishar's demand for boxes
! with integer widths of blk_wd.  The strange form of the if statement
! occurs because the smooshing of the boxes finds the correct leftmost
! boundaries of the box and they do not need to be moved by one point to
! the left.

      goto 199

         if (blk_wd.gt.1) then
            do i = 1,no_box
               if (mod(box_stk(1,i)-2,blk_wd).eq.0.and.              
     &             box_stk(1,i).ne.1) then
                  box_stk(1,i) = box_stk(1,i) -1
               end if
               if (mod(box_stk(2,i)-2,blk_wd).eq.0.and.               
     &             box_stk(2,i).ne.1) then
                  box_stk(2,i) = box_stk(2,i) -1
               end if
               if (mod(box_stk(3,i)-2,blk_wd).eq.0.and.              
     &             box_stk(3,i).ne.1) then
                  box_stk(3,i) = box_stk(3,i) -1
               end if
            end do
         end if

 199  continue 

      return
 10   format(a,i3,a,i6,a,i3,a,i8)
 20   format(i4,f10.4,i6,8i4)
      end
!-----------------------------------------------------------------------
!
!  subroutine that computes the signatures and laplacians
!     of the given box\
!
!-----------------------------------------------------------------------
      subroutine cmp_sig_lpl(box, sigx, sigy, sigz, lplx, lply,     
     &                       lplz, box_stk, nx, ny, nz, a)

      implicit none

         integer nx, ny, nz
         integer sigx(nx),sigy(ny),sigz(nz),box_stk(6)
         integer lplx(nx),lply(ny),lplz(nz)
         integer  box(nx,ny,nz)

         integer i, j, k,x_lo, y_lo, z_lo, x_hi, y_hi, z_hi, tmpl, tmpr
         integer a

         logical trace
         parameter (trace = .false.)


         call ivls(sigx,0,nx)
         call ivls(sigy,0,ny)
         call ivls(sigz,0,nz)
         x_lo = box_stk(1)
         y_lo = box_stk(2)
         z_lo = box_stk(3)
         x_hi = box_stk(4)
         y_hi = box_stk(5)
         z_hi = box_stk(6)

         if (trace) then
            write(8,*)'a= ',a
            write(8,*)'nx ',nx
            write(8,*)'ny ',ny
            write(8,*)'nz ',nz
            write(8,*)'x_lo ',x_lo
            write(8,*)'y_lo ',y_lo
            write(8,*)'z_lo ',z_lo
            write(8,*)'x_hi ',x_hi
            write(8,*)'y_hi ',y_hi
            write(8,*)'z_hi ',z_hi
         end if

         do i = x_lo, x_hi
            do j = y_lo, y_hi
               do k = z_lo, z_hi
                  sigx(i) = sigx(i) + box(i,j,k)
               end do
            end do
         end do

         do j = y_lo, y_hi
            do i = x_lo, x_hi
               do k = z_lo, z_hi
                  sigy(j) = sigy(j) + box(i,j,k)
               end do
            end do
         end do

         do k = z_lo, z_hi
            do j = y_lo, y_hi
               do i = x_lo, x_hi
                  sigz(k) = sigz(k) + box(i,j,k)
               end do
            end do
         end do

! make sure outer laplacians are always zero otherwise the (i/j)+/-1
!  references will pollute the algorithm
!         lplx(x_lo) = 0
!         lplx(x_hi) = 0
!         lply(y_lo) = 0
!         lply(y_hi) = 0
!         lply(z_lo) = 0
!         lply(z_hi) = 0
!         do i = x_lo+1,x_hi -1
!            lplx(i) = sigx(i+1)-2*sigx(i)+sigx(i-1)
!         end do
!
!         do j = y_lo+1,y_hi -1
!            lply(j) = sigy(j+1)-2*sigy(j)+sigy(j-1)
!         end do
!
!         do j = z_lo+1,z_hi -1
!            lplz(j) = sigz(j+1)-2*sigz(j)+sigz(j-1)
!         end do

!
! compute laplacians by the difference of 
! the absolute value of the gradients
!  - this should take care of some of 
! the problems of using the above laplacian.
         call ivls(lplx,0,nx)
         call ivls(lply,0,ny)
         call ivls(lplz,0,nz)

         do i = x_lo, x_hi-1
            do j = y_lo, y_hi
               do k = z_lo, z_hi
                  lplx(i) = lplx(i) + abs(box(i,j,k)-box(i+1,j,k))
               end do
            end do
         end do

         do j = y_lo, y_hi-1
            do i = x_lo, x_hi
               do k = z_lo, z_hi
                  lply(j) = lply(j) + abs(box(i,j,k)-box(i,j+1,k))
               end do
            end do
         end do

         do k = z_lo, z_hi-1
            do j = y_lo, y_hi
               do i = x_lo, x_hi
                  lplz(k) = lplz(k) + abs(box(i,j,k)-box(i,j,k+1))
               end do
            end do
         end do

         do i = x_hi-1,x_lo+1,-1
            lplx(i) = lplx(i)-lplx(i-1)
         end do

         do j = y_hi-1,y_lo+1,-1
            lply(j) = lply(j)-lply(j-1)
         end do

         do k = z_hi-1,z_lo+1,-1
            lplz(k) = lplz(k)-lplz(k-1)
         end do


         lplx(x_lo) = 0
         lply(y_lo) = 0
         lplz(z_lo) = 0
         lplx(x_hi) = 0
         lply(y_hi) = 0
         lplz(z_hi) = 0

      return
      end
!-----------------------------------------------------------------------
!
!  subroutine that computes the efficiency of the given box
!   box_eff = (number of flagged mesh points in the box)/
!              (total number of mesh points in the box)
!
!-----------------------------------------------------------------------
      double precision function box_eff(box, box_stk, nx, ny, nz)

      implicit none

         integer nx, ny, nz
         integer box_stk(6)
         integer box(nx,ny,nz)

         integer flg_pts, i, j, k, x_lo, y_lo, z_lo, x_hi, y_hi, z_hi

         x_lo = box_stk(1)
         y_lo = box_stk(2)
         z_lo = box_stk(3)
         x_hi = box_stk(4)
         y_hi = box_stk(5)
         z_hi = box_stk(6)

         flg_pts = 0

         do i = x_lo, x_hi
            do j = y_lo, y_hi
               do k = z_lo, z_hi
                  flg_pts = flg_pts + (box(i,j,k))
               end do
            end do
         end do

         box_eff = 1.0d0*flg_pts/                                
     &               (x_hi-x_lo+1.0d0)/                           
     &               (y_hi-y_lo+1.0d0)/(z_hi-z_lo+1.0d0)

      return
      end
!-----------------------------------------------------------------------
!
!  Subrotine that tries to reduce the size of the box if there are empty
!  regions along the sides of the boxes.
!
!-----------------------------------------------------------------------
      subroutine reduce(stkid,box_stk,blk_wd,sigx,sigy,sigz,nx,ny,nz)

      implicit none

         integer stkid, blk_wd, nx, ny, nz
         integer box_stk(6), sigx(nx), sigy(ny), sigz(nz)
         integer i, j, k, nx_lo, nx_hi, ny_lo, ny_hi, nz_lo, nz_hi

         logical trace
         parameter (trace = .false.)

         nx_lo = box_stk(1)
         ny_lo = box_stk(2)
         nz_lo = box_stk(3)
         nx_hi = box_stk(4)
         ny_hi = box_stk(5)
         nz_hi = box_stk(6)

         if (trace) then
            i = stkid
            write(8,*)'REDUCE>> before running'
            write(8,*)'stkid =',i
            write(8,*)'nx_lo =',box_stk(1)
            write(8,*)'ny_lo =',box_stk(2)
            write(8,*)'nz_lo =',box_stk(3)
            write(8,*)'nx_hi =',box_stk(4)
            write(8,*)'ny_hi =',box_stk(5)
            write(8,*)'nz_hi =',box_stk(6)
!            do j = box_stk(2),box_stk(5)
!               write(8,*)'sigy(',j,')=',sigy(j)
!            end do
         end if
! first see if there the box can be reduced
         i = nx_lo
 11      if (sigx(i).eq.0.and.(nx_hi-nx_lo).ge.blk_wd) then
!           if (mod(i,blk_wd).eq.0.and.i.ne.1) then
               nx_lo = i + 1
!           end if
            i = i + 1
            goto 11
         end if

         i = nx_hi
 16      if (sigx(i).eq.0.and.(nx_hi-nx_lo).ge.blk_wd) then
!           if (mod(i-2,blk_wd).eq.0) then
               nx_hi = i - 1
!           end if
            i = i - 1
            goto 16
         end if

         j = ny_lo
  21     if (sigy(j).eq.0.and.(ny_hi-ny_lo).ge.blk_wd) then
!           if (mod(j,blk_wd).eq.0.and.j.ne.1) then
               ny_lo = j + 1
!           end if
            j = j + 1
            goto 21
         end if

         j = ny_hi
  26     if (sigy(j).eq.0.and.(ny_hi-ny_lo).ge.blk_wd) then
!           if (mod(j-2,blk_wd).eq.0) then
               ny_hi = j - 1
!           end if
            j = j - 1
            goto 26
         end if

         k = nz_lo
  31     if (sigz(k).eq.0.and.(nz_hi-nz_lo).ge.blk_wd) then
!           if (mod(k,blk_wd).eq.0.and.k.ne.1) then
               nz_lo = k + 1
!           end if
            k = k + 1
            goto 31
         end if

         k = nz_hi
  36     if (sigz(k).eq.0.and.(nz_hi-nz_lo).ge.blk_wd) then
!           if (mod(k-2,blk_wd).eq.0) then
               nz_hi = k - 1
!           end if
            k = k - 1
            goto 36
         end if

         box_stk(1) = nx_lo
         box_stk(2) = ny_lo
         box_stk(3) = nz_lo
         box_stk(4) = nx_hi
         box_stk(5) = ny_hi
         box_stk(6) = nz_hi

      return
      end


!-----------------------------------------------------------------------
!
!  This routine tries to split the inefficient box first by 
!   seeing if there are any zeros in the signatures (lines in the
!   box mesh with no flagged points) and if that does not
!   work then it computes the laplacian of the signatures and looks
!   for inflection points (i.e. where the laplacian goes to zero).
!
!-----------------------------------------------------------------------
      subroutine find_split(nx,ny,nz,no_box,stkid,blk_wd,               &
     &                     box_stk,                                     &
     &                     sigx, sigy, sigz, lplx, lply, lplz,          &
     &                     flag, box_max)

      implicit none

         integer Find_Holes, Find_Inflx
         integer no_box, nx, ny, nz, stkid, blk_wd
         integer flag, box_max
         integer    sigx(nx), sigy(ny), sigz(nz)
         integer lplx(nx), lply(ny), lplz(nz)
         integer box_stk(6,box_max)

!
!       Look for Holes
!       Search for holes and split box and store in box_stk.
!       flag returns 1 if holes were found. 0 if not.
!       Routine has to add to NO_box if splitting is done.
!
         flag = find_holes(no_box,stkid,blk_wd,sigx,sigy,sigz,          &
     &                      box_stk,box_max)

!
!       If no holes are found then look for inflextion points.
!           Search for inflextion points and split if possible.
!         If splitting is possible add to stack. Return flag = 1
!         if splitting is carried out. else return 1
!       Routine has to add to NO_box if splitting is done.
!
         if ( flag .eq. 0 )then
            flag = Find_Inflx(nx,ny,nz,no_box,stkid,blk_wd,sigx,         &
     &                  sigy,sigz,lplx,lply,lplz,box_stk,box_max)
         end if

      return
      end

!-----------------------------------------------------------------------
!
! Integer function that attempts to find a hole (zero in the signature
!  array) in the box mesh and if it does then it splits the
!  box into two at that hole
!
!-----------------------------------------------------------------------
      integer function find_holes(no_box,stkid,blk_wd,                  &
     &                           sigx, sigy, sigz, box_stk,box_max)

      implicit none

         integer no_box, stkid, box_max, blk_wd
         integer sigx(*), sigy(*), sigz(*), box_stk(6,*)
         integer i, j, k, nx_lo, nx_hi, ny_lo, ny_hi, nz_lo, nz_hi

         logical trace
         parameter (trace = .false.)

         find_holes = 0
         nx_lo = box_stk(1,stkid)
         ny_lo = box_stk(2,stkid)
         nz_lo = box_stk(3,stkid)
         nx_hi = box_stk(4,stkid)
         ny_hi = box_stk(5,stkid)
         nz_hi = box_stk(6,stkid)

         if (trace) then
            i = stkid
            write(8,*)'Find_Holes before'
            write(8,*)'stkid =',i
            write(8,*)'nx_lo =',box_stk(1,i)
            write(8,*)'ny_lo =',box_stk(2,i)
            write(8,*)'nz_lo =',box_stk(3,i)
            write(8,*)'nx_hi =',box_stk(4,i)
            write(8,*)'ny_hi =',box_stk(5,i)
            write(8,*)'nz_hi =',box_stk(6,i)
            do j = box_stk(2,i),box_stk(5,i)
               write(8,*)'sigy(',j,')=',sigy(j)
            end do
         end if

! first see if there the box can be reduced
         i = nx_lo
 10      if (sigx(i).eq.0.and.(nx_hi-nx_lo).ge.blk_wd) then
!           if (mod(i,blk_wd).eq.0.and.i.ne.1) then
               nx_lo = i + 1
               find_holes = 1
!           end if
            i = i + 1
            goto 10
         end if

         i = nx_hi
 15      if (sigx(i).eq.0.and.(nx_hi-nx_lo).ge.blk_wd) then
!           if (mod(i-2,blk_wd).eq.0) then
               nx_hi = i - 1
               find_holes = 1
!           end if
            i = i - 1
            goto 15
         end if

         j = ny_lo
  20     if (sigy(j).eq.0.and.(ny_hi-ny_lo).ge.blk_wd) then
!           if (mod(j,blk_wd).eq.0.and.j.ne.1) then
               ny_lo = j + 1
               find_holes = 1
!           end if
            j = j + 1
            goto 20
         end if

         j = ny_hi
  25     if (sigy(j).eq.0.and.(ny_hi-ny_lo).ge.blk_wd) then
!           if (mod(j-2,blk_wd).eq.0) then
               ny_hi = j - 1
               find_holes = 1
!           end if
            j = j - 1
            goto 25
         end if

         k = nz_lo
  30     if (sigz(k).eq.0.and.(nz_hi-nz_lo).ge.blk_wd) then
!           if (mod(k,blk_wd).eq.0.and.k.ne.1) then
               nz_lo = k + 1
               find_holes = 1
!           end if
            k = k + 1
            goto 30
         end if

         k = nz_hi
  35     if (sigz(k).eq.0.and.(nz_hi-nz_lo).ge.blk_wd) then
!           if (mod(k-2,blk_wd).eq.0) then
               nz_hi = k - 1
               find_holes = 1
!           end if
            k = k - 1
            goto 35
         end if

         box_stk(1,stkid) = nx_lo
         box_stk(2,stkid) = ny_lo
         box_stk(3,stkid) = nz_lo
         box_stk(4,stkid) = nx_hi
         box_stk(5,stkid) = ny_hi
         box_stk(6,stkid) = nz_hi

! now see if the current box can be divided

         do i = nx_lo+(blk_wd),nx_hi-(blk_wd)
!           if (sigx(i).eq.0.and.(mod(i-2,blk_wd)).eq.0) then
            if (sigx(i).eq.0) then

               if (trace) then
                  write(8,*)'stkid =',stkid
                  write(8,*)'nx_lo =',nx_lo
                  write(8,*)'ny_lo =',ny_lo
                  write(8,*)'nz_lo =',nz_lo
                  write(8,*)'i-1 =',i-1
                  write(8,*)'ny_hi =',ny_hi
                  write(8,*)'i+1 =',i+1
                  write(8,*)'nx_hi =',nx_hi
                  write(8,*)'nz_hi =',nz_hi
                  write(8,*)'box_stk(1) =',box_stk(1,stkid)
                  write(8,*)'box_stk(2) =',box_stk(2,stkid)
                  write(8,*)'box_stk(3) =',box_stk(3,stkid)
                  write(8,*)'box_stk(4) =',box_stk(4,stkid)
                  write(8,*)'box_stk(5) =',box_stk(5,stkid)
                  write(8,*)'box_stk(5) =',box_stk(5,stkid)
               end if

               call boxStack(stkid, nx_lo, ny_lo, nz_lo, i-1, ny_hi,    &
     &                        nz_hi, i+1,ny_lo, nz_lo, nx_hi, ny_hi,    &
     &                        nz_hi, box_stk, no_box, box_max)

               if (trace) then
                  write(8,*)'box_stk(1,id) =',box_stk(1,stkid)
                  write(8,*)'box_stk(2,id) =',box_stk(2,stkid)
                  write(8,*)'box_stk(3,id) =',box_stk(3,stkid)
                  write(8,*)'box_stk(4,id) =',box_stk(4,stkid)
                  write(8,*)'box_stk(5,id) =',box_stk(5,stkid)
                  write(8,*)'box_stk(6,id) =',box_stk(6,stkid)
                  write(8,*)'box_stk(1) =',box_stk(1,no_box)
                  write(8,*)'box_stk(2) =',box_stk(2,no_box)
                  write(8,*)'box_stk(3) =',box_stk(3,no_box)
                  write(8,*)'box_stk(4) =',box_stk(4,no_box)
                  write(8,*)'box_stk(5) =',box_stk(5,no_box)
                  write(8,*)'box_stk(6) =',box_stk(6,no_box)
               end if

               Find_Holes = 1
               goto 200
            end if
         end do

         do j = ny_lo+(blk_wd),ny_hi-(blk_wd)
!           if (sigy(j).eq.0.and.(mod(j-2,blk_wd)).eq.0) then
            if (sigy(j).eq.0) then
               if (trace) then
                write(8,*)'Find_Holes intermediate'
                write(8,*)'stkid =',stkid
                write(8,*)'nx_lo =',box_stk(1,stkid)
                write(8,*)'ny_lo =',box_stk(2,stkid)
                write(8,*)'nz_lo =',box_stk(3,stkid)
                write(8,*)'nx_hi =',box_stk(4,stkid)
                write(8,*)'ny_hi =',box_stk(5,stkid)
                write(8,*)'nz_hi =',box_stk(6,stkid)
                  do i = box_stk(2,stkid),box_stk(5,stkid)
                     write(8,*)'sigy(',i,')=',sigy(i)
                  end do
               end if

               call boxStack(stkid, nx_lo, ny_lo, nz_lo, nx_hi, j-1,  
     &                        nz_hi,nx_lo,j+1, nz_lo,nx_hi, ny_hi,     
     &                        nz_hi, box_stk, no_box, box_max)
               Find_Holes = 1
               if (trace) then
                write(8,*)'Find_Holes after'
                write(8,*)'stkid =',stkid
                write(8,*)'nx_lo =',box_stk(1,stkid)
                write(8,*)'ny_lo =',box_stk(2,stkid)
                write(8,*)'nz_lo =',box_stk(3,stkid)
                write(8,*)'nx_hi =',box_stk(4,stkid)
                write(8,*)'ny_hi =',box_stk(5,stkid)
                write(8,*)'nz_hi =',box_stk(6,stkid)
                  do i = box_stk(2,stkid),box_stk(5,stkid)
                     write(8,*)'sigy(',i,')=',sigy(i)
                  end do
               end if

               goto 200
            end if
         end do

         do k = nz_lo+(blk_wd),nz_hi-(blk_wd)
!           if (sigz(k).eq.0.and.(mod(k-2,blk_wd)).eq.0) then
            if (sigz(k).eq.0) then
               call boxStack(stkid, nx_lo, ny_lo, nz_lo, nx_hi, ny_hi, 
     &                        k-1,nx_lo,ny_lo, k+1,nx_hi, ny_hi,      
     &                        nz_hi, box_stk, no_box, box_max)
               Find_Holes = 1
               goto 200
            end if
         end do
200   continue

      return
      end


          
!-----------------------------------------------------------------------
!
!  Routine: boxStack 
!   Internal Routine to Add and Delete boxs from the 
!    box stack.
!   
!  Input:
!       stkid := Current stack number. The one to delete.
!       stlo1x/y/z := First box begining coordinates. 
!       stlo2x/y/z := Second box begining coordinates. 
!       sthi1x/y/z := First box extensions. 
!       sthi2x/y/z := Second box extensions. 
!       box_stk := box Stack.
!       NO_box  := Number of boxs thus far. Will add 1 to this.
!       box_max     := Maximum number of boxs allowed.
!  
!  Output:
!       Routine will update the value of NO_box and add in an
!       additional box and replace the current one in box_stk.
!
!-----------------------------------------------------------------------
      subroutine boxStack(stkid, stlo1x, stlo1y, stlo1z,sthi1x, sthi1y,   &
     &           sthi1z,stlo2x, stlo2y, stlo2z, sthi2x, sthi2y, sthi2z,   &
     &           box_stk, NO_box, box_max)

      implicit none

         integer NO_box, box_max
         integer stkid
         integer stlo1x, stlo1y, stlo1z, sthi1x, sthi1y, sthi1z
         integer stlo2x, stlo2y, stlo2z, sthi2x, sthi2y, sthi2z
         integer box_stk(6,box_max)
         integer i,j

!     Check if adding one more box will exceed the maximum
!     stack size. 

         if( (NO_box + 1) .gt. box_max )then
          write(8,*)'Error in boxStack:'
          write(8,*)'     Number of boxs will ',                           &
     &          'exceed maximum  stack size'
          stop
        end if

!     Begin by replacing the current stack with the first box.

         box_stk( 1, stkid) = stlo1x
         box_stk( 2, stkid) = stlo1y
         box_stk( 3, stkid) = stlo1z
         box_stk( 4, stkid) = sthi1x
         box_stk( 5, stkid) = sthi1y
         box_stk( 6, stkid) = sthi1z

!     Next add one more stack at the end.

         box_stk( 1, NO_box+1) = stlo2x
         box_stk( 2, NO_box+1) = stlo2y
         box_stk( 3, NO_box+1) = stlo2z
         box_stk( 4, NO_box+1) = sthi2x
         box_stk( 5, NO_box+1) = sthi2y
         box_stk( 6, NO_box+1) = sthi2z

!     Add one to the number of boxs

         NO_box = NO_box + 1


         return
       end
!-----------------------------------------------------------------------
! Routine: InitStack
!
! Simple Routine that initializes the box stack.
!
! Input: 
!      box_stk := box Stack array.
!       stbgx/y := box begining coordinates.
!       stexx/y := box extensions.
!-----------------------------------------------------------------------
      subroutine InitStack(box_stk, stbgx, stbgy, stbgz,stexx, stexy,    &
     &           stexz,NO_box, box_max)

         implicit none

         integer box_max, NO_box,                                        &
     &           box_stk(6,box_max),                                     &
     &           stbgx, stbgy, stbgz,stexx, stexy,stexz


!     Initialize NO_box

         NO_box = 1

!     Set up first stack element.

         box_stk( 1, 1) = stbgx
         box_stk( 2, 1) = stbgy
         box_stk( 3, 1) = stbgz
         box_stk( 4, 1) = stexx
         box_stk( 5, 1) = stexy
         box_stk( 6, 1) = stexz

         return
        end
!-----------------------------------------------------------------------
!
!   Routine: Find_Inflx
!   Integer function that calculates the laplacian given the signature
!   arrays and locates inflextion points or zeros in the laplacian.
!   With these inflextion points a choice is made as to how to split the
!   box.
!
!   Laplacian array begins at lox+1 and ends at hix-1. 
!
!-----------------------------------------------------------------------
      integer function find_inflx(nx,ny,nz,no_box,stkid,blk_wd,sigx,    &
     &                      sigy,sigz,lplx,lply,lplz,box_stk, box_max)

      implicit none

         integer no_box,stkid,box_max,nx,ny,nz,blk_wd
         integer box_stk(6,box_max)
         integer sigx(nx), sigy(ny), sigz(nz)
         integer lplx(nx), lply(ny), lplz(nz)

         integer nx_lo, ny_lo, nz_lo, nx_hi, ny_hi, nz_hi
         integer nx_lo1, ny_lo1, nz_lo1,nx_hi1, ny_hi1,nz_hi1
         integer nx_lo2, ny_lo2, nz_lo2, nx_hi2, ny_hi2, nz_hi2

         integer i,j, k, hi_lplx, infptx, hi_lply, infpty, tmp
         integer hi_lplz, infptz

         logical trace
         parameter (trace = .false.)

         nx_lo = box_stk(1,stkid)
         ny_lo = box_stk(2,stkid)
         nz_lo = box_stk(3,stkid)
         nx_hi = box_stk(4,stkid)
         ny_hi = box_stk(5,stkid)
         nz_hi = box_stk(6,stkid)

!
! looking for two possible cases:
!   two succesive laplacians have opposite signs
         hi_lplx = 0
         infptx = 0
         do i = nx_lo+blk_wd,nx_hi-blk_wd
            if ((lplx(i).gt.0.and.lplx(i+1).lt.0).or.           
     &          (lplx(i).lt.0.and.lplx(i+1).gt.0)) then
!              if (mod(i-1,blk_wd).eq.0) then
                  tmp = abs(lplx(i)-lplx(i+1))
                  if (tmp.gt.hi_lplx) then
                     hi_lplx = tmp
                     infptx  = i
                  end if
!              end if
            end if
         end do

         hi_lply = 0
         infpty = 0
         do j = ny_lo+blk_wd,ny_hi-blk_wd
            if ((lply(j).gt.0.and.lply(j+1).lt.0).or.            
     &          (lply(j).lt.0.and.lply(j+1).gt.0)) then
!              if (mod(j-1,blk_wd).eq.0) then
                  tmp = abs(lply(j)-lply(j+1))
                  if (tmp.gt.hi_lply) then
                     hi_lply = tmp
                     infpty  = j
                  end if
!              end if
            end if
         end do

         hi_lplz = 0
         infptz = 0
         do k = nz_lo+blk_wd,nz_hi-blk_wd
            if ((lplz(k).gt.0.and.lplz(k+1).lt.0).or.           
     &          (lplz(k).lt.0.and.lplz(k+1).gt.0)) then
!              if (mod(k-1,blk_wd).eq.0) then
                  tmp = abs(lplz(k)-lplz(k+1))
                  if (tmp.gt.hi_lplz) then
                     hi_lplz = tmp
                     infptz  = k
                  end if
!              end if
            end if
         end do

         if (infpty.ne.0.or.infptx.ne.0.or.infptz.ne.0) then
            if (hi_lplx.ge.hi_lply.and.hi_lplx.ge.hi_lplz) then
               nx_lo1 = nx_lo
               ny_lo1 = ny_lo
               nz_lo1 = nz_lo
               nx_hi1 = infptx
               ny_hi1 = ny_hi
               nz_hi1 = nz_hi

               nx_lo2 = infptx+1
               ny_lo2 = ny_lo
               nz_lo2 = nz_lo
               nx_hi2 = nx_hi
               ny_hi2 = ny_hi
               nz_hi2 = nz_hi
            else if (hi_lply.ge.hi_lplx.and.hi_lply.ge.hi_lplz) then
               nx_lo1 = nx_lo
               ny_lo1 = ny_lo
               nz_lo1 = nz_lo
               nx_hi1 = nx_hi
               ny_hi1 = infpty
               nz_hi1 = nz_hi

               nx_lo2 = nx_lo
               ny_lo2 = infpty+1
               nz_lo2 = nz_lo
               nx_hi2 = nx_hi
               ny_hi2 = ny_hi
               nz_hi2 = nz_hi
            else
               nx_lo1 = nx_lo
               ny_lo1 = ny_lo
               nz_lo1 = nz_lo
               nx_hi1 = nx_hi
               ny_hi1 = ny_hi
               nz_hi1 = infptz

               nx_lo2 = nx_lo
               ny_lo2 = ny_lo
               nz_lo2 = infptz+1
               nx_hi2 = nx_hi
               ny_hi2 = ny_hi
               nz_hi2 = nz_hi
            end if
               
            call boxStack(stkid, nx_lo1, ny_lo1, nz_lo1,nx_hi1, ny_hi1,  &
     &           nz_hi1, nx_lo2, ny_lo2, nz_lo2,nx_hi2, ny_hi2, nz_hi2,  &
     &           box_stk, NO_box, box_max)
            find_inflx = 1
         else
            find_inflx = 0
         end if

      return
 10   format(a,i3,a,i4)
      end
!------------------------------------------------------------------------
!
! Routine that outputs what the clustering algorithm has done to the
!  data given initially
!
!   box_stk(1-3,i) contains nx_lo, ny_lo, nz_lo - of the ith box
!   box_stk(4-6,i) contains nx_hi, ny_hi, nz_hi - of the ith box
!
!   fnamef - contains the flagged points.
!   fnameb - contains the box info.
!
!-----------------------------------------------------------------------
      subroutine out_clstr(nx,ny,nz,no_boxes,big_box,box_stk,
     &                     level, 
     &                     cx,cy,cz,fnamef,fnameb)

      implicit none

         integer ind, indlnb
         integer nx,ny,nz,no_boxes, i,j,k, level
         character*(*) fnamef, fnameb

         real*8 xmin,xmax,ymin,ymax,zmin,zmax
         integer bnx,bny,bnz

         integer box_stk(6,no_boxes)
!
!  hdf variables needed
!
         integer    gf3_rc, gf3_write_hdf_f77
         integer    dshape(3),       cshape(3)
         real*8     big_box(nx,ny,nz),cx(nx),cy(ny),cz(nz)

         if (nx.gt.100.or.ny.gt.100.or.nz.gt.100) then
            call error('Out_clstr: box size too large. > 100.')
         end if

         open(unit=7,file=fnameb(1:indlnb(fnameb)),form='formatted',   
     &           err=900)
         write(7,10)0.0d0,0.0d0,0.0d0, 0.0d0,0.0d0,0.0d0, 0, 0, 0, 0
!         xmin = 1.0d0
!         ymin = 1.0d0
!         zmin = 1.0d0
!         xmax = 1.0d0*nx
!         ymax = 1.0d0*ny
!         zmax = 1.0d0*nz
!         write(7,10)xmin,xmax,ymin,ymax,zmin,zmax,nx,ny,nz,0
         do i = 1,no_boxes
            xmin = (box_stk(1,i)) *1.0d0 +1.0d0
            xmax = (box_stk(4,i)) *1.0d0 +1.0d0
            ymin = (box_stk(2,i)) *1.0d0 +1.0d0
            ymax = (box_stk(5,i)) *1.0d0 +1.0d0
            zmin = (box_stk(3,i)) *1.0d0 +1.0d0
            zmax = (box_stk(6,i))*1.0d0
!            bnx  = box_stk(4,i)-box_stk(1,i) + 1
!            bny  = box_stk(5,i)-box_stk(2,i) + 1
!            bnz  = box_stk(6,i)-box_stk(3,i) + 1
				bnx = 5
				bny = 5
				bnz = 5
            write(7,10)xmin,xmax,ymin,ymax,zmin,zmax,bnx,bny,bnz,i
         end do
         close(7)

! flag file
         cshape(1) = nx
         cshape(2) = ny
         cshape(3) = nz

         dshape(1) = nx
         dshape(2) = ny
         dshape(3) = nz

			do i = 1 ,nx
				cx(i) = 1.0d0*i 
			end do
			do j = 1 ,ny
				cy(j) = 1.0d0*j
			end do
			do k = 1 ,nz
				cz(k) = 1.0d0*k
			end do
         gf3_rc = gf3_write_hdf_f77(fnamef(1:indlnb(fnamef)),        
     &            'flagged_pts',cshape, cx, cy, cz, dshape, big_box)
         if( gf3_rc .ne. 0 ) then
            call error('Out_clstr: Error writing flag file')
         end if

      return
  10  format(1p,6e13.5,4i4)
 900  continue
      write(8,*)'Out_clstr: error opening ',fnameb(1:indlnb(fnameb))
      stop
      end

!---------------------------------------------------------------------
!
!     Load vector with scalar.
!
!---------------------------------------------------------------------

      SUBROUTINE IVLS(V1,S1,N)
 
         INTEGER     V1(1)
         INTEGER     S1
         INTEGER     I, N
 
         DO 10 I = 1 , N
            V1(I) = S1
 10      CONTINUE
 
         RETURN
 
      END
		subroutine error(name)
		implicit none
		character*(*) name
		write(0,*) name
		return
		end
