! Copyright Scott A. Klasky NPAC 1996
!============================================================================
      module def_grid
         type grd
            integer                             :: parent
            integer                             :: child
            integer                             :: neighbor
            integer                             :: sibling
            integer,dimension(3)                :: lb,ub
         end type grd
         integer, dimension(:),allocatable      :: lhead 
         integer, parameter                     :: UNDEF = -1
         integer, dimension(:),allocatable      :: lnext
      end module def_grid
!============================================================================
      subroutine cls_3d(msk,shp,blk_overlap, &
                        box_max,box_no,box_stk_lb, &
                        box_stk_ub,min_eff)
      USE def_grid
      implicit none
      integer, dimension(3)                      :: shp
      integer, dimension(shp(1),shp(2),shp(3))   :: msk
      integer                                    :: blk_overlap
      integer                                    :: box_max, box_no
      integer, dimension(3,box_max)              :: box_stk_lb, box_stk_ub 
      integer                                    :: box_number 
      logical                                    :: not_done
      real*8                                     :: comput_eff,eff
      real*8                                     :: min_eff
      integer                                    :: level, grid
      type(grd), dimension(box_max)              :: grid_ptr
      integer                                    :: init_tree
      integer,dimension(:,:),allocatable         :: sigmax,sigmay,sigmaz
      integer,dimension(:,:),allocatable         :: delx,dely,delz
      integer                                    :: maxsize
      logical                                    :: found_hole, find_hole

      allocate(lhead(box_max))
      allocate( sigmax(shp(2),shp(3)),sigmay(shp(1),shp(3)) )
      allocate( sigmaz(shp(1),shp(2)) )
      allocate( delx(shp(2),shp(3)),dely(shp(1),shp(3)) )
      allocate( delz(shp(1),shp(2)) )
      allocate(lnext(6*box_max))

      not_done = .true.
      lnext    = UNDEF
      level    =  init_tree(grid_ptr,box_max,shp)

      do while (not_done) 
100   continue 

         grid = lhead(level)
         if (grid == UNDEF) not_done = .true.

200      continue

            if (grid == UNDEF) then
               level = level + 1
               goto 100 
            end if

            eff = comput_eff(grid,grid_ptr,msk,shp,box_max) 

            if (eff.ge.min_eff) then
               grid = grid_ptr(grid)%neighbor
               goto 200 
            end if

            call compute_sigma(grid,grid_ptr,msk,shp,box_max,  &
                               sigmax,sigmay,sigmaz )

            found_hole = find_hole(grid,grid_ptr,sigmax,sigmay,sigmaz)

   end do

! now we put in the buffer zones, while extracting the usefull information
! into the output arrays

      return
      end
!============================================================================
      logical function find_hole(grid,grid_ptr,sigmax,sigmay,sigmaz)
      USE def_grid
      implicit none
      type(grd), dimension(box_max)              :: grid_ptr
      integer                                    :: grid
      integer, dimension(:,:)                    :: sigmax
      integer, dimension(:,:)                    :: sigmay
      integer, dimension(:,:)                    :: sigmaz
		integer i,j,k

      find_hole = .false.


		do j = grid_ptr(grid)%lb(2),grid_ptr(grid)%ub(2)
		   do k = grid_ptr(grid)%lb(3),grid_ptr(grid)%ub(3)
				if (sigmax(j,k).eq.0) then
					find_hole = .true.

				end if
			end do
		end do

      do k = grid_ptr(grid)%lb(3),grid_ptr(grid)%ub(3)
         do i = grid_ptr(grid)%lb(1),grid_ptr(grid)%ub(1)
            if (sigmay(k,i).eq.0) then
					find_hole = .true.

            end if
         end do
      end do

      do i = grid_ptr(grid)%lb(1),grid_ptr(grid)%ub(1)
         do j = grid_ptr(grid)%lb(2),grid_ptr(grid)%ub(2)
            if (sigmaz(i,j).eq.0) then
					find_hole = .true.

            end if
         end do
      end do

		


      return
      end
!============================================================================
      subroutine compute_sigma(grid,grid_ptr,msk,shp,box_max,      &
                               sigmax,sigmay,sigmaz)
      USE def_grid
      implicit none
      integer                                    :: box_max
      type(grd), dimension(box_max)              :: grid_ptr
      integer, dimension(3)                      :: shp 
      integer, dimension(shp(1),shp(2),shp(3))   :: msk
      integer                                    :: grid
      integer, dimension(:,:)                    :: sigmax 
      integer, dimension(:,:)                    :: sigmay 
      integer, dimension(:,:)                    :: sigmaz

! compute sigma in all 3 directions
      sigmax = 0
      sigmay = 0
      sigmaz = 0

      sigmax=                                                      &
         sum( msk( grid_ptr(grid)%lb(1):grid_ptr(grid)%ub(1),      &
                   grid_ptr(grid)%lb(2):grid_ptr(grid)%ub(2),      &
                   grid_ptr(grid)%lb(3):grid_ptr(grid)%ub(3) ), DIM = 1 )

      sigmay=                                                      &
         sum( msk( grid_ptr(grid)%lb(1):grid_ptr(grid)%ub(1),      &
                   grid_ptr(grid)%lb(2):grid_ptr(grid)%ub(2),      &
                   grid_ptr(grid)%lb(3):grid_ptr(grid)%ub(3) ), DIM = 2 )

      sigmaz=                                                      &
         sum( msk( grid_ptr(grid)%lb(1):grid_ptr(grid)%ub(1),      &
                   grid_ptr(grid)%lb(2):grid_ptr(grid)%ub(2),      &
                   grid_ptr(grid)%lb(3):grid_ptr(grid)%ub(3) ), DIM = 3 )

      return
      end
!============================================================================
      integer function init_tree(grid_ptr,box_max,shp)
      USE def_grid
      implicit none
      integer                                    :: box_max
      type(grd), dimension(box_max)              :: grid_ptr
      integer, dimension(3)                      :: shp 

      grid_ptr%parent   = UNDEF
      grid_ptr%child    = UNDEF
      grid_ptr%neighbor = UNDEF
      grid_ptr%sibling  = UNDEF
      lhead             = UNDEF
      grid_ptr%lb(1)       = UNDEF
      grid_ptr%lb(2)       = UNDEF
      grid_ptr%lb(3)       = UNDEF
      grid_ptr%ub(1)       = UNDEF
      grid_ptr%ub(2)       = UNDEF
      grid_ptr%ub(3)       = UNDEF
      
      lhead(1)           = 1
      grid_ptr(1)%lb(1)  = 1
      grid_ptr(1)%lb(2)  = 1
      grid_ptr(1)%lb(3)  = 1
      grid_ptr(1)%ub(1)  = shp(1) 
      grid_ptr(1)%ub(2)  = shp(2) 
      grid_ptr(1)%ub(3)  = shp(3) 
      lnext(1) = 1

      init_tree = 1

      return
      end
!============================================================================
      real*8 function comput_eff(grid,grid_ptr,msk,shp,box_max)
      USE def_grid
      implicit none
      integer                                    :: grid
      integer, dimension(3)                      :: shp
      integer, dimension(shp(1),shp(2),shp(3))   :: msk
      integer                                    :: box_max
      type(grd), dimension(box_max)              :: grid_ptr


      comput_eff = sum(msk( grid_ptr(grid)%lb(1):grid_ptr(grid)%ub(1),      &
                            grid_ptr(grid)%lb(2):grid_ptr(grid)%ub(2),      &
                            grid_ptr(grid)%lb(3):grid_ptr(grid)%ub(3) ) )   & 
                       / ( (grid_ptr(grid)%ub(1)-grid_ptr(grid)%lb(1)+1) *  &
                           (grid_ptr(grid)%ub(2)-grid_ptr(grid)%lb(2)+1) *  &
                           (grid_ptr(grid)%ub(3)-grid_ptr(grid)%lb(3)+1) ) 

      return
      end
