! Copyright Scott A. Klasky NPAC 1996
!============================================================================
      module def_grid
         type grd
            integer                             :: parent
            integer                             :: child
            integer                             :: neighbor
            integer                             :: sibling
            integer,dimension(3)                :: lb,ub
         end type grd
         integer, dimension(:),allocatable      :: lhead 
         integer, parameter                     :: UNDEF = -1
         integer, parameter                     :: max_stor = 10000
         integer, dimension(max_stor)           :: lnext
      end module def_grid
!============================================================================
      subroutine cls_3d(msk,shp,blk_overlap, &
                        box_no,box_stk_lb, &
                        box_stk_ub,min_eff,minlen2d)
      USE def_grid
      implicit none
      integer, dimension(3)                      :: shp
      integer, dimension(shp(1),shp(2),shp(3))   :: msk
      integer                                    :: blk_overlap
      integer,parameter                          :: box_max = 10000
      integer                                    :: box_no
      integer, dimension(3,box_max)              :: box_stk_lb, box_stk_ub 
      integer                                    :: box_number 
      logical                                    :: not_done
      real*8                                     :: comput_eff,eff
      real*8                                     :: min_eff
      integer                                    :: level, grid,lmax
      type(grd), dimension(box_max)              :: grid_ptr
      integer                                    :: init_tree
      integer,dimension(shp(1))                  :: sigmax
      integer,dimension(shp(2))                  :: sigmay
      integer,dimension(shp(3))                  :: sigmaz
      integer,dimension(shp(1))                  :: delx
      integer,dimension(shp(2))                  :: dely
      integer,dimension(shp(3))                  :: delz
      integer                                    :: maxsize
      logical                                    :: found_hole, find_hole
      integer                                    :: minlen2d
      integer                                    :: max_lev  !used for debugging
      logical,parameter                          :: ltrace = .true.

      allocate(lhead(box_max))

      not_done = .true.
      lnext    = UNDEF
      level    =  init_tree(grid_ptr,box_max,shp)

      do while (not_done) 
100   continue 

         grid = lhead(level)
         if (grid == UNDEF) exit 

200      continue

            if (grid == UNDEF) then
               level = level + 1
               goto 100 
            end if

            eff = comput_eff(grid,grid_ptr,msk,shp,box_max,minlen2d) 

            if (eff.ge.min_eff) then
               grid = grid_ptr(grid)%neighbor
               goto 200 
            end if

            call compute_sigma(grid,grid_ptr,msk,shp,box_max,  &
                               sigmax,sigmay,sigmaz) 

            found_hole = find_hole(box_max,level,grid,grid_ptr,&
                                   sigmax,sigmay,sigmaz,shp,   &
                                   msk,minlen2d,min_eff) 

            if (.not.found_hole) then

               call compute_delta(shp,grid,grid_ptr,box_max,  &
                                  sigmax,sigmay,sigmaz,delx,dely,delz)
               call find_great_infl(level,grid,grid_ptr,box_max,  &
                                    shp,delx,dely,delz,minlen2d) 

            end if

            grid = grid_ptr(grid)%neighbor
            goto 200
   end do

! now we put in the buffer zones, while extracting the usefull information
! into the output arrays

      lmax = level - 1
      box_no = 0

      do level = 1 , lmax
         grid = lhead(level)
300      continue
         if (grid.eq.UNDEF) cycle

         if (grid_ptr(grid)%child.eq.UNDEF) then 
            box_no = box_no + 1

! don't forget to add the buffer zone
            box_stk_lb(1,box_no) = grid_ptr(grid)%lb(1) - blk_overlap
            box_stk_lb(2,box_no) = grid_ptr(grid)%lb(2) - blk_overlap
            box_stk_lb(3,box_no) = grid_ptr(grid)%lb(3) - blk_overlap 
            box_stk_ub(1,box_no) = grid_ptr(grid)%ub(1) + blk_overlap 
            box_stk_ub(2,box_no) = grid_ptr(grid)%ub(2) + blk_overlap
            box_stk_ub(3,box_no) = grid_ptr(grid)%ub(3) + blk_overlap

            eff = comput_eff(grid,grid_ptr,msk,shp,box_max,1)

            if (ltrace) &
            write(67,66)grid_ptr(grid)%lb(1),                 &
                       grid_ptr(grid)%lb(2),grid_ptr(grid)%lb(3), &
                       grid_ptr(grid)%ub(1),grid_ptr(grid)%ub(2), &
                       grid_ptr(grid)%ub(3),grid,eff, &
                       (( grid_ptr(grid)%ub(1)-grid_ptr(grid)%lb(1)+1)*   &
                       ( grid_ptr(grid)%ub(2)-grid_ptr(grid)%lb(2)+1)*   &
                       ( grid_ptr(grid)%ub(3)-grid_ptr(grid)%lb(3)+1))*eff, &
                       (( grid_ptr(grid)%ub(1)-grid_ptr(grid)%lb(1)+1)*   &
                       ( grid_ptr(grid)%ub(2)-grid_ptr(grid)%lb(2)+1)*   &
                       ( grid_ptr(grid)%ub(3)-grid_ptr(grid)%lb(3)+1))*1.0d0
66 format(7i5,1P,3e12.2) 


         end if

         grid = grid_ptr(grid)%neighbor
         goto 300
      end do


      return
      end
!============================================================================
      subroutine find_great_infl(level,grid,grid_ptr,box_max,  &
                                 shp,delx,dely,delz,minlen2d)
      USE def_grid
      implicit none
      integer                                    :: box_max
      integer                                    :: grid,level
      integer, dimension(3)                      :: shp
      type(grd), dimension(box_max)              :: grid_ptr
      integer, dimension(shp(1))                 :: delx
      integer, dimension(shp(2))                 :: dely
      integer, dimension(shp(3))                 :: delz
      integer                                    :: dir,location
      integer                                    :: change,maxchange
      integer, dimension(3)                      :: nlb,nub
      integer                                    :: i,j,k
      integer, external                          :: sign
      integer                                    :: s1,s2,s3,s4,s5,s6
      logical                                    :: t1,t2,t3,t4,t5,t6
      integer                                    :: minlen2d     

      dir      = 1
      location = 1
      maxchange = 0

      if ( grid_ptr(grid)%lb(1)+1 < grid_ptr(grid)%ub(1) - 1 ) then
         do i = grid_ptr(grid)%lb(1)+1, grid_ptr(grid)%ub(1) - 1
            change = abs(delx(i) - delx(i+1))
            if (change.gt.maxchange.and.                                  &
                sign(delx(i)).ne.sign(delx(i+1))) then
               maxchange = change
               location  = i
               dir       = 1
            end if
         end do
      end if

      if ( grid_ptr(grid)%lb(2)+1 < grid_ptr(grid)%ub(2) - 1 ) then
         do j = grid_ptr(grid)%lb(2)+1, grid_ptr(grid)%ub(2) - 1
            change = abs(dely(j) - dely(j+1))
            if (change.gt.maxchange.and.                                  &
                sign(dely(j)).ne.sign(dely(j+1))) then
               maxchange = change
               location  = j
               dir       = 2
            end if
         end do
      end if

      if ( grid_ptr(grid)%lb(3)+1 < grid_ptr(grid)%ub(3) - 1 ) then
         do k = grid_ptr(grid)%lb(3)+1, grid_ptr(grid)%ub(3) - 1
            change = abs(delz(k) - delz(k+1))
            if (change.gt.maxchange.and.                                  &
                sign(delz(k)).ne.sign(delz(k+1))) then
               maxchange = change
               location  = k
               dir       = 3
            end if
         end do
      end if

   
! now we can make a two new grids
! the s1-s6,t1-t6 arrays help make sure that a box can't be made to small this
! way
      if (maxchange.gt.0) then
         if (dir.eq.1) then
            nlb(1) = grid_ptr(grid)%lb(1)
            nlb(2) = grid_ptr(grid)%lb(2)
            nlb(3) = grid_ptr(grid)%lb(3)
            nub(1) = location 
            nub(2) = grid_ptr(grid)%ub(2)
            nub(3) = grid_ptr(grid)%ub(3)

            s1 = nub(1)-nlb(1)+1 
            s2 = nub(2)-nlb(2)+1 
            s3 = nub(3)-nlb(3)+1 
            s4 = grid_ptr(grid)%ub(1) - (location + 1) 
            s5 = s2 
            s6 = s3 
            t1 = s1.ge.minlen2d
            t2 = s2.ge.minlen2d
            t3 = s3.ge.minlen2d
            t4 = s4.ge.minlen2d
            t5 = t2 
            t6 = t3 

            if ( ((t1.and.t2).or.(t1.and.t3).or.(t2.and.t3)).and.      &
                 ((t4.and.t5).or.(t4.and.t6).or.(t5.and.t6)))  then 
               call make_grid(level,grid,grid_ptr,nlb,nub,box_max) 
   
               nlb(1) = location + 1 
               nlb(2) = grid_ptr(grid)%lb(2)
               nlb(3) = grid_ptr(grid)%lb(3)
               nub(1) = grid_ptr(grid)%ub(1)
               nub(2) = grid_ptr(grid)%ub(2)
               nub(3) = grid_ptr(grid)%ub(3)
               call make_grid(level,grid,grid_ptr,nlb,nub,box_max)
            end if
   
         else if (dir.eq.2) then
   
            nlb(1) = grid_ptr(grid)%lb(1)
            nlb(2) = grid_ptr(grid)%lb(2)
            nlb(3) = grid_ptr(grid)%lb(3)
            nub(1) = grid_ptr(grid)%ub(1)
            nub(2) = location
            nub(3) = grid_ptr(grid)%ub(3)
            s1 = nub(1)-nlb(1)+1 
            s2 = nub(2)-nlb(2)+1 
            s3 = nub(3)-nlb(3)+1 
            s4 = s1 
            s5 = grid_ptr(grid)%ub(2) - (location + 1) 
            s6 = s3 
            t1 = s1.ge.minlen2d
            t2 = s2.ge.minlen2d
            t3 = s3.ge.minlen2d
            t4 = s4.ge.minlen2d
            t5 = t2 
            t6 = t3 
            if ( ((t1.and.t2).or.(t1.and.t3).or.(t2.and.t3)).and.      &
                 ((t4.and.t5).or.(t4.and.t6).or.(t5.and.t6)))   then 
               call make_grid(level,grid,grid_ptr,nlb,nub,box_max)
   
               nlb(1) = grid_ptr(grid)%lb(1)
               nlb(2) = location + 1
               nlb(3) = grid_ptr(grid)%lb(3)
               nub(1) = grid_ptr(grid)%ub(1)
               nub(2) = grid_ptr(grid)%ub(2)
               nub(3) = grid_ptr(grid)%ub(3)
               call make_grid(level,grid,grid_ptr,nlb,nub,box_max)
            end if
   
         else
   
            nlb(1) = grid_ptr(grid)%lb(1)
            nlb(2) = grid_ptr(grid)%lb(2)
            nlb(3) = grid_ptr(grid)%lb(3)
            nub(1) = grid_ptr(grid)%ub(1)
            nub(2) = grid_ptr(grid)%ub(2)
            nub(3) = location
            s1 = nub(1)-nlb(1)+1 
            s2 = nub(2)-nlb(2)+1 
            s3 = nub(3)-nlb(3)+1 
            s4 = s1 
            s5 = s2
            s6 = grid_ptr(grid)%ub(3) - (location + 1) 
            t1 = s1.ge.minlen2d
            t2 = s2.ge.minlen2d
            t3 = s3.ge.minlen2d
            t4 = s4.ge.minlen2d
            t5 = t2 
            t6 = t3 
            if ( ((t1.and.t2).or.(t1.and.t3).or.(t2.and.t3)).and.      &
                 ((t4.and.t5).or.(t4.and.t6).or.(t5.and.t6)))   then 
               call make_grid(level,grid,grid_ptr,nlb,nub,box_max)
   
               nlb(1) = grid_ptr(grid)%lb(1)
               nlb(2) = grid_ptr(grid)%lb(2)
               nlb(3) = location + 1
               nub(1) = grid_ptr(grid)%ub(1)
               nub(2) = grid_ptr(grid)%ub(2)
               nub(3) = grid_ptr(grid)%ub(3)
               call make_grid(level,grid,grid_ptr,nlb,nub,box_max)
            end if
      
         end if
   
      end if

      return
      end
!============================================================================
      subroutine compute_delta(shp,grid,grid_ptr,box_max,  &
                               sigmax,sigmay,sigmaz,delx,dely,delz)
      USE def_grid
      implicit none
      integer                                    :: box_max
      integer                                    :: grid
      type(grd), dimension(box_max)              :: grid_ptr
      integer, dimension(3)                      :: shp
      integer, dimension(shp(1))                 :: sigmax
      integer, dimension(shp(2))                 :: sigmay
      integer, dimension(shp(3))                 :: sigmaz
      integer, dimension(shp(1))                 :: delx 
      integer, dimension(shp(2))                 :: dely
      integer, dimension(shp(3))                 :: delz

! compute sigma in all 3 directions
      delx = 0
      dely = 0
      delz = 0

      delx= cshift(sigmax,1,1) + cshift(sigmax,-1,1) - 2*cshift(sigmax,0,1)
      dely= cshift(sigmay,1,1) + cshift(sigmay,-1,1) - 2*cshift(sigmay,0,1)
      delz= cshift(sigmaz,1,1) + cshift(sigmaz,-1,1) - 2*cshift(sigmaz,0,1)

      return
      end
!============================================================================
      logical function find_hole(box_max,level,grid,grid_ptr,   &
                                 sigmax,sigmay,sigmaz,shp,      &
                                 msk,minlen2d,min_eff) 
      USE def_grid
      implicit none
      integer                                    :: box_max
      type(grd), dimension(box_max)              :: grid_ptr
      integer                                    :: grid,level
      integer, dimension(3)                      :: shp
      integer, dimension(shp(1),shp(2),shp(3))   :: msk
      integer, dimension(shp(1))                 :: sigmax
      integer, dimension(shp(2))                 :: sigmay
      integer, dimension(shp(3))                 :: sigmaz
      integer                                    :: i,j,k,l
      integer, dimension(3)                      :: nlb,nub
      logical                                    :: start
      real*8                                     :: comput_eff,eff
		integer                                    :: min_eff,minlen2d

      find_hole = .false.

      if (grid_ptr(grid)%lb(1)   < grid_ptr(grid)%ub(1) ) then 
         if ( sigmax(grid_ptr(grid)%lb(1)) == 0 ) then
            do i = grid_ptr(grid)%lb(1),grid_ptr(grid)%ub(1)
               if (sigmax(i).ne.0) then 
                  grid_ptr(grid)%lb(1) = i
						find_hole = .true.
                  exit
               end if
            end do
         end if
         if ( sigmax(grid_ptr(grid)%ub(1)) == 0 ) then
            do i = grid_ptr(grid)%ub(1),grid_ptr(grid)%lb(1),-1
               if (sigmax(i).ne.0) then 
                  grid_ptr(grid)%ub(1) = i
						find_hole = .true.
                  exit
               end if
            end do
         end if
      end if

      if (grid_ptr(grid)%lb(2)   < grid_ptr(grid)%ub(2) ) then  
         if ( sigmay(grid_ptr(grid)%lb(2)) == 0 ) then
            do j = grid_ptr(grid)%lb(2),grid_ptr(grid)%ub(2)
               if (sigmay(j).ne.0) then 
                  grid_ptr(grid)%lb(2) = j
						find_hole = .true.
                  exit
               end if
            end do
         end if

         if ( sigmay(grid_ptr(grid)%ub(2)) == 0 ) then
            do j = grid_ptr(grid)%ub(2),grid_ptr(grid)%lb(2),-1
               if (sigmay(j).ne.0) then
                  grid_ptr(grid)%ub(2) = j
						find_hole = .true.
                  exit
               end if
            end do
         end if
      end if
      
   
      if (grid_ptr(grid)%lb(3)   < grid_ptr(grid)%ub(3) ) then  
         if ( sigmaz(grid_ptr(grid)%lb(3)) == 0 ) then
            do k = grid_ptr(grid)%lb(3),grid_ptr(grid)%ub(3)
               if (sigmaz(k).ne.0) then
                  grid_ptr(grid)%lb(3) = k
						find_hole = .true.
                  exit
               end if
            end do
         end if
   
         if ( sigmaz(grid_ptr(grid)%ub(3)) == 0 ) then
            do k = grid_ptr(grid)%ub(3),grid_ptr(grid)%lb(3),-1
               if (sigmaz(k).ne.0) then
                  grid_ptr(grid)%ub(3) = k
						find_hole = .true.
                  exit
               end if
            end do
         end if
      end if

		if (find_hole) then
         eff = comput_eff(grid,grid_ptr,msk,shp,box_max,minlen2d)
			if (eff.ge.min_eff) then
				return
			end if
		end if

		find_hole = .false.
		

! now find any more holes  ...
      if (grid_ptr(grid)%lb(1)   < grid_ptr(grid)%ub(1) ) then

         nlb(3) = grid_ptr(grid)%lb(3)
         nub(3) = grid_ptr(grid)%ub(3)
         nlb(2) = grid_ptr(grid)%lb(2)
         nub(2) = grid_ptr(grid)%ub(2)
         nlb(1) = grid_ptr(grid)%lb(1)
         start  = .true.
   
         do i= grid_ptr(grid)%lb(1),grid_ptr(grid)%ub(1)
            if (start.and.sigmax(i).eq.0) then
               find_hole = .true.
               nub(1) = i - 1
               call make_grid(level,grid,grid_ptr,nlb,nub,box_max)
               start = .false.
            else if (.not.start.and.sigmax(i).ne.0) then
               nlb(1) = i
               start = .true.
            end if
         end do
   
         if (start.and.nlb(1).ne.grid_ptr(grid)%lb(1)) then
            nub(1) = grid_ptr(grid)%ub(1)
            call make_grid(level,grid,grid_ptr,nlb,nub,box_max)   
         end if
   
      end if


         
      if (grid_ptr(grid)%lb(2)   < grid_ptr(grid)%ub(2) ) then
         nlb(1) = grid_ptr(grid)%lb(1)
         nub(1) = grid_ptr(grid)%ub(1)
         nlb(3) = grid_ptr(grid)%lb(3)
         nub(3) = grid_ptr(grid)%ub(3)
         nlb(2) = grid_ptr(grid)%lb(2)
         start  = .true.

         do j = grid_ptr(grid)%lb(2),grid_ptr(grid)%ub(2)
            if (start.and.sigmay(j).eq.0) then
               find_hole = .true.
               nub(2) = j - 1
               call make_grid(level,grid,grid_ptr,nlb,nub,box_max) 
               start = .false.
            else if (.not.start.and.sigmay(j).ne.0) then
               nlb(2) = j
               start = .true.
            end if
         end do
   
         if (start.and.nlb(2).ne.grid_ptr(grid)%lb(2)) then
            nub(2) = grid_ptr(grid)%ub(2)
            call make_grid(level,grid,grid_ptr,nlb,nub,box_max) 
         end if

      end if

      if (grid_ptr(grid)%lb(3)   < grid_ptr(grid)%ub(3) ) then
         nlb(1) = grid_ptr(grid)%lb(1)
         nub(1) = grid_ptr(grid)%ub(1)
         nlb(2) = grid_ptr(grid)%lb(2)
         nub(2) = grid_ptr(grid)%ub(2)
         nlb(3) = grid_ptr(grid)%lb(3)
         start  = .true.

         do k = grid_ptr(grid)%lb(3),grid_ptr(grid)%ub(3)
            if (start.and.sigmaz(k).eq.0) then
               find_hole = .true.
               nub(3) = k - 1
               call make_grid(level,grid,grid_ptr,nlb,nub,box_max)
               start = .false.
            else if (.not.start.and.sigmaz(k).ne.0) then
               nlb(3) = k
               start = .true.
            end if
         end do
   
         if (start.and.nlb(3).ne.grid_ptr(grid)%lb(3)) then
            nub(3) = grid_ptr(grid)%ub(3)
            call make_grid(level,grid,grid_ptr,nlb,nub,box_max)
         end if

      end if
   
      return
      end
!============================================================================
      subroutine make_grid(level,grid,grid_ptr,nlb,nub,box_max)
      USE def_grid
      implicit none
      integer                                    :: grid,level
      integer                                    :: box_max
      type(grd), dimension(box_max)              :: grid_ptr
      integer, dimension(3)                      :: nlb,nub
      integer                                    :: new_grid_num
      integer                                    :: get_grid

      new_grid_num = get_grid()

      if ( lhead(level+1) == UNDEF )  then

         grid_ptr(grid)%child           = new_grid_num
         grid_ptr(new_grid_num)%parent  = grid 
         lhead(level+1)                 = new_grid_num

      else

         grid_ptr(grid)%child            = new_grid_num
         grid_ptr(new_grid_num)%parent   = grid 
         grid_ptr(new_grid_num)%neighbor = lhead(level+1)
         lhead(level+1)                  = new_grid_num

      end if

      grid_ptr(new_grid_num)%lb(1)  = nlb(1)
      grid_ptr(new_grid_num)%lb(2)  = nlb(2)
      grid_ptr(new_grid_num)%lb(3)  = nlb(3)
      grid_ptr(new_grid_num)%ub(1)  = nub(1)
      grid_ptr(new_grid_num)%ub(2)  = nub(2)
      grid_ptr(new_grid_num)%ub(3)  = nub(3)

      return
      end
!============================================================================
      integer function  get_grid() 
      USE def_grid
      implicit none
      integer i

      do i = 1 , max_stor
         if (lnext(i).eq.UNDEF) then
            lnext(i) = 1 
            get_grid = i
            return
         end if
      end do

      write(0,*) 'MUST increase max_stor in the module that is in clustering' 
      stop
      end
!============================================================================
      subroutine compute_sigma(grid,grid_ptr,msk,shp,box_max,      &
                               sigmax,sigmay,sigmaz) 
      USE def_grid
      implicit none
      integer                                           :: box_max
      type(grd), dimension(box_max)                     :: grid_ptr
      integer, dimension(3)                             :: shp 
      integer, dimension(shp(1),shp(2),shp(3))          :: msk
      integer                                           :: grid
      integer, dimension(shp(1))                        :: sigmax 
      integer, dimension(shp(2))                        :: sigmay 
      integer, dimension(shp(3))                        :: sigmaz
      integer                                           :: i,j,k

! compute sigma in all 3 directions
      sigmax = 0
      sigmay = 0
      sigmaz = 0

      do i = grid_ptr(grid)%lb(1),grid_ptr(grid)%ub(1)
         do j = grid_ptr(grid)%lb(2),grid_ptr(grid)%ub(2)
            do k = grid_ptr(grid)%lb(3),grid_ptr(grid)%ub(3)
               sigmax(i) = sigmax(i) + msk(i,j,k)
            end do
         end do
      end do

      do j = grid_ptr(grid)%lb(2),grid_ptr(grid)%ub(2)
         do k = grid_ptr(grid)%lb(3),grid_ptr(grid)%ub(3)
            do i = grid_ptr(grid)%lb(1),grid_ptr(grid)%ub(1)
               sigmay(j) = sigmay(j) + msk(i,j,k)
            end do
         end do
      end do

      do k = grid_ptr(grid)%lb(3),grid_ptr(grid)%ub(3)
         do j = grid_ptr(grid)%lb(2),grid_ptr(grid)%ub(2)
            do i = grid_ptr(grid)%lb(1),grid_ptr(grid)%ub(1)
               sigmaz(k) = sigmaz(k) + msk(i,j,k)
            end do
         end do
      end do

      return
      end
!============================================================================
      integer function init_tree(grid_ptr,box_max,shp)
      USE def_grid
      implicit none
      integer                                    :: box_max
      type(grd), dimension(box_max)              :: grid_ptr
      integer, dimension(3)                      :: shp 

      grid_ptr%parent      = UNDEF
      grid_ptr%child       = UNDEF
      grid_ptr%neighbor    = UNDEF
      grid_ptr%sibling     = UNDEF
      lhead                = UNDEF
      grid_ptr%lb(1)       = UNDEF
      grid_ptr%lb(2)       = UNDEF
      grid_ptr%lb(3)       = UNDEF
      grid_ptr%ub(1)       = UNDEF
      grid_ptr%ub(2)       = UNDEF
      grid_ptr%ub(3)       = UNDEF
      
      lhead(1)             = 1
      grid_ptr(1)%lb(1)    = 1
      grid_ptr(1)%lb(2)    = 1
      grid_ptr(1)%lb(3)    = 1
      grid_ptr(1)%ub(1)    = shp(1) 
      grid_ptr(1)%ub(2)    = shp(2) 
      grid_ptr(1)%ub(3)    = shp(3) 
      lnext(1) = 1

      init_tree = 1

      return
      end
!============================================================================
      real*8 function comput_eff(grid,grid_ptr,msk,shp,box_max,minlen2d)
      USE def_grid
      implicit none
      integer                                    :: grid
      integer, dimension(3)                      :: shp
      integer, dimension(shp(1),shp(2),shp(3))   :: msk
      integer                                    :: box_max
      type(grd), dimension(box_max)              :: grid_ptr
      integer                                    :: i,j,k
      integer                                    :: minlen2d 
      integer                                    :: s1,s2,s3
      logical                                    :: t1,t2,t3

      s1 = grid_ptr(grid)%ub(1)-grid_ptr(grid)%lb(1)+1
      s2 = grid_ptr(grid)%ub(2)-grid_ptr(grid)%lb(2)+1
      s3 = grid_ptr(grid)%ub(3)-grid_ptr(grid)%lb(3)+1


      comput_eff = 1.0d0*sum(msk(grid_ptr(grid)%lb(1):grid_ptr(grid)%ub(1), &
                           grid_ptr(grid)%lb(2):grid_ptr(grid)%ub(2),       &
                           grid_ptr(grid)%lb(3):grid_ptr(grid)%ub(3) ) )  / &
                           (s1*s2*s3) *1.0d0 

      t1 = s1.lt.minlen2d
      t2 = s2.lt.minlen2d
      t3 = s3.lt.minlen2d

      if ( (t1.and.t2).or.(t1.and.t3).or.(t2.and.t3) ) comput_eff = 1.0d0

      return
      end
!============================================================================
      integer function sign(num)
      implicit none
      integer num

      if (num.lt.0) then
         sign = -1
      else if (num.gt.0) then
         sign = +1
      else 
         sign = 0
      end if 

      return
      end 
