c-----------------------------------------------------------------------
c
c     Front end to mgv0 which invokes mgv0  with 
c
c     preswf = presw = preswp
c     pstswf = pstsw = pstswp
c
c-----------------------------------------------------------------------

      real*8 function mgvcyc
     *   (uf,rhsf,nxf,nyf,nzf, hf, q,sizeq,
     *    preswp,pstswp,ncycle)

         implicit      none

         real*8        mgv0

         integer       nxf,       nyf,     nzf,      sizeq
         real*8        uf(nxf,nyf,nzf),    rhsf(nxf,nyf,nzf),
     *                 q(sizeq)
         real*8        hf
         integer       preswp,    pstswp,  ncycle

         mgvcyc = mgv0(uf,rhsf,nxf,nyf,nzf,hf,q,sizeq,
     *                 preswp,pstswp,preswp,pstswp,ncycle)
         
         return

      end

c-----------------------------------------------------------------------
c
c     Basic v-cycle multi grid solver for 
c
c     Laplacian(u) = rhs(u)
c
c     in (x,y,z) coordinates on rectangular domain. 
c-----------------------------------------------------------------------
c     NOTE: ncycle is input *and* output parameter.  Do *not* call 
c     with constant (i.e. 0) as actual parameter!  Call with ncycle = 0
c     the first time routine is called with given rhs, ncycle will
c     then be automatically updated so that if routine is rd-called
c     without changing rhs, ncycle, prd-smoothing sweeps will be 
c     skipped ...
c-----------------------------------------------------------------------

      real*8 function mgv0
     *   (uf,rhsf,nxf,nyf,nzf, hf, q, sizeq,
     *    preswf,pstswf,presw,pstsw,ncycle)

         implicit      none

         real*8        d3nrm2,    relaxc
         integer       qalloc,    lilog2
         logical       lp2np1

         integer       nxf,       nyf,     nzf,      sizeq
         real*8        uf(nxf,nyf,nzf),    rhsf(nxf,nyf,nzf),
     *                 q(sizeq)
         real*8        hf
         integer       preswf,    pstswf,  presw,    pstsw,
     *                 ncycle
c
c        Multi-grid parameters, data structures and pointers 
c
         real*8        epsic
         parameter   ( epsic = 1.0d-10 )

         integer       lmin,      lmax
         parameter   ( lmin = 1,  lmax = 20 )

         integer       nx(lmin:lmax),      ny(lmin:lmax),
     *                 nz(lmin:lmax)
         real*8        h(lmin:lmax)

         integer       u(lmin:lmax),       rhs(lmin:lmax),
     *                 mres(lmin:lmax),    du(lmin:lmax)

         integer       imposs
         parameter   ( imposs = -2 000 000 000 )

         integer          p,      pceil
         common  / comp / p,      pceil
c
c        Locals ...
c
         integer       lxf,       lyf,       lzf,
     *                 l,         lf,        lc,       nl,
     *                 sweep

         real*8        resnrm

         logical       ltrace 
         parameter   ( ltrace = .false. )

         save

         mgv0 = -1.0d0
c
c        Parameter check ...
c

         if( ltrace ) then
            write(0,*) '>>> mgv0: Parameter dump: '
            write(0,*) nxf, nyf, nzf, hf
            write(0,*) preswf, pstswf, presw, pstsw, ncycle
         end if
         if( .not. lp2np1(nxf)  .or.  .not. lp2np1(nyf)  .or.
     *       .not. lp2np1(nzf) ) then
            write(0,*) '>>> mgv0: nxf, nyf, nxf must be of form '//
     *                 '2^n + 1.'
            write(0,*) nxf, nyf, nzf
            mgv0 = -1.0d0
            return
         end if

         lxf = lilog2(nxf)
         lyf = lilog2(nyf)
         lzf = lilog2(nzf)

         lf = min(lxf,min(lyf,lzf))

         if( lf .gt. lmax ) then
            write(0,*) '>>> mgv0: lf > lmax'
            write(0,*) lf, lmax
            mgv0 = -1.0d0
            return
         end if

         lc = lmin
         nl = lf - lc + 1

         if( ltrace ) then
            write(0,*) '>>> mgv0: lxf, lyf, lzf, lc, lf, nl'
            write(0,*) lxf, lyf, lzf, lc, lf, nl
         end if

         if( ncycle .eq. 0 ) then
c
c           Set up grid structure, allocate storage ...
c
            call qreset(sizeq)
            nx(lf) = nxf
            ny(lf) = nyf
            nz(lf) = nzf
            h(lf)  = hf
            u(lf)    = imposs
            rhs(lf)  = imposs
            mres(lf) = qalloc(nx(lf)*ny(lf)*nz(lf))
            du(lf)   = mres(lf)
            do 1000 l = lf - 1 , lc , -1
               nx(l) = (nx(l+1) - 1) / 2 + 1
               ny(l) = (ny(l+1) - 1) / 2 + 1
               nz(l) = (nz(l+1) - 1) / 2 + 1
               h(l) = 2.0d0 * h(l+1)

               u(l)    = qalloc(nx(l)*ny(l)*nz(l))
               rhs(l)  = qalloc(nx(l)*ny(l)*nz(l))
               mres(l) = qalloc(nx(l)*ny(l)*nz(l))
               du(l)   = mres(l)
 1000       continue
            if( ltrace ) then
               call ivdump(nx,nl,'nx',0)
               call ivdump(ny,nl,'ny',0)
               call ivdump(nz,nl,'nz',0)
               call dvdump(h,nl,'h',0)
               call ivdump(u,nl,'u pointers',0)
               call ivdump(rhs,nl,'rhs pointers',0)
               call ivdump(mres,nl,'mres pointers',0)
               call ivdump(du,nl,'du pointers',0)
            end if
         end if
c
c        Cycle from fine to coarse ...
c
         do l = lf , lc , -1
            if( l .ne. lf  .or.  ncycle .eq. 0 ) then
               if(      l .eq. lc ) then
 100              continue
                     resnrm = relaxc(q(u(l)),q(rhs(l)),
     *                               nx(l),ny(l),nz(l),h(l))
                     if( ltrace ) then
                        call trace_res('mgv0: pre sweep',l,resnrm)
                     end if
                  if( resnrm .gt. epsic ) go to 100
               else if( l .eq. lf ) then
                  do sweep = 1 , preswf
                     resnrm = relaxc(uf,rhsf,
     *                               nx(l),ny(l),nz(l),h(l))
                     call cmres(q(mres(l)),uf,rhsf,
     *                          nx(l),ny(l),nz(l),h(l))
                     if( ltrace ) then
                        call trace_res('mgv0: pre sweep',l,resnrm)
                     end if
                  end do
               else 
                  do sweep = 1 , presw 
                     resnrm = relaxc(q(u(l)),q(rhs(l)),
     *                               nx(l),ny(l),nz(l),h(l))
                     if( ltrace ) then
                        call trace_res('mgv0: pre sweep',l,resnrm)
                     end if
                  end do
               end if
            end if
c
c           Set up coarse grid problem ...
c
            if(      l .eq. lf ) then
               call cmres(q(mres(l)),uf,rhsf,nx(l),ny(l),nz(l),h(l))
            else if( l .ne. lc ) then
               call cmres(q(mres(l)),q(u(l)),q(rhs(l)),
     *                    nx(l),ny(l),nz(l),h(l))
            end if
            if( l .ne. lc ) then
               call d3rshw(q(mres(l)),q(rhs(l-1)),
     *                     nx(l-1),ny(l-1),nz(l-1))
               call d3ls(q(u(l-1)),0.0d0,
     *                   nx(l-1),ny(l-1),nz(l-1))
            end if

         end do 
c
c        Cycle from coarse to fine ...
c

         do l = lc + 1 , lf 
c
c           Interpolate correction ...
c
            call d3lint(q(u(l-1)),q(du(l)),nx(l-1),ny(l-1),nz(l-1))
            if( l .eq. lf ) then
               call d33a(uf,q(du(l)),uf,nx(l),ny(l),nz(l))
            else 
               call d33a(q(u(l)),q(du(l)),q(u(l)),nx(l),ny(l),nz(l))
            end if

            if( l .eq. lf ) then
               do sweep = 1 , pstswf
                  resnrm = relaxc(uf,rhsf,nx(l),ny(l),nz(l),h(l))
                  if( ltrace ) then
                     call trace_res('mgv0: post sweep',l,resnrm)
                  end if
               end do
            else 
               do sweep = 1 , pstsw 
                  resnrm = relaxc(q(u(l)),q(rhs(l)),
     *                            nx(l),ny(l),nz(l),h(l))
                  if( ltrace ) then
                     call trace_res('mgv0: post sweep',l,resnrm)
                  end if
               end do
            end if

         end do

         ncycle = ncycle + 1
         mgv0 = resnrm

         return

      end

c-----------------------------------------------------------------------
c
c     Resets pointer into main storage array.
c
c-----------------------------------------------------------------------

      subroutine qreset(ceil) 

         implicit         none

         integer          ceil

         integer          p,      pceil
         common  / comp / p,      pceil

         pceil = ceil
         p = 1

         return

      end

c-----------------------------------------------------------------------
c
c     Returns next available location in main storage array and updates 
c     pointer.
c
c-----------------------------------------------------------------------

      integer function qalloc(size) 

         integer          size

         integer          p,      pceil
         common  / comp / p,      pceil

         if( size .gt. 0 ) then
            if( p + size .le. pceil ) then
               qalloc = p
               p = p + size
            else 
               write(0,*) '>>> qalloc: Out of memory'
               stop
            end if
         else 
            write(0,*) '>>> qalloc: size ', size, ' <= 0.'
            qalloc = -1
         end if
         
         return

      end

c-----------------------------------------------------------------------
c
c     Returns number of words of memory used.
c
c-----------------------------------------------------------------------

      integer function qused()

         integer          p,      pceil
         common  / comp / p,      pceil

         qused = p

         return

      end
c-----------------------------------------------------------------------
c
c     Returns number of words of available memory.
c
c-----------------------------------------------------------------------

      integer function qleft()

         integer          p,      pceil
         common  / comp / p,      pceil

         qleft = pceil - p + 1

         return

      end

c-----------------------------------------------------------------------
c
c     Does ond-step Point-GS relaxation and returns norm of
c     running ('instantaneous') residuals---checkerboard ordering.
c
c-----------------------------------------------------------------------
c
      real*8 function relaxc(u,rhs,nx,ny,nz,h)
c
         implicit       none
c
         integer        nx,              ny,            nz
         real*8         u(nx,ny,nz),     rhs(nx,ny,nz)
         real*8         h
c
         real*8         hm2,      mjelm1,   rresl
         integer        i,        j,        k,
     *                  isw,      ksw,      pass
c
         relaxc = 0.0d0
         hm2 = 1.0d0 / (h * h)
         mjelm1 =  h * h / 6.0d0
c
         ksw = 2
         do 40 pass = 1 , 2
            isw = ksw
            do 30 k = 2 , nz - 1
               do 20 j = 2 , ny - 1
                  do 10 i = isw + 1 , nx - 1, 2
                     rresl = hm2 * (u(i+1,j,k) + u(i-1,j,k) +
     *                              u(i,j+1,k) + u(i,j-1,k) +
     *                              u(i,j,k+1) + u(i,j,k-1) -
     *                              6.0d0 * u(i,j,k)) -
     *                       rhs(i,j,k)
                     u(i,j,k) = u(i,j,k) + mjelm1 *  rresl 
                     relaxc = relaxc + rresl * rresl
 10               continue
                  isw = 3 - isw
 20            continue
 30         continue
            ksw = 3 - ksw
 40      continue
         relaxc = sqrt(relaxc / (nx * ny * nz))
c
         return
c
      end
 
c-----------------------------------------------------------------------
c
c     Computes negative of residuals ...
c
c-----------------------------------------------------------------------
 
      subroutine cmres(mres,u,rhs,nx,ny,nz,h)
c
         implicit       none
c
         integer        nx,               ny,           nz
         real*8         mres(nx,ny,nz),   u(nx,ny,nz),
     *                  rhs(nx,ny,nz)
c
         real*8         h
c
         integer        i,                j,            k
         real*8         hm2
c
         hm2 = 1.0d0 / (h * h)
c
         do 30 k = 2 , nz - 1
            do 20 j = 2 , ny - 1
               do 10 i = 2 , nx - 1
                  mres(i,j,k) = rhs(i,j,k) - 
     *                          hm2 * (u(i+1,j,k) + u(i-1,j,k) +
     *                                 u(i,j+1,k) + u(i,j-1,k) +
     *                                 u(i,j,k+1) + u(i,j,k-1) -
     *                                   6.0d0*u(i,j,k))
 10            continue
 20         continue
 30      continue
         call d3lbs(mres,0.0d0,nx,ny,nz)
c
         return
c
      end

c-----------------------------------------------------------------------
c
c     Tracing routine
c
c-----------------------------------------------------------------------

      subroutine trace_res(string,l,resnrm)
   
         implicit      none

         character*(*) string 
         integer       l
         real*8        resnrm

         write(0,1000) string, l, resnrm
1000     format(a,' Level: ',i2,'  |res.|: ',1pe12.3)

         return

      end

C-----------------------------------------------------------------------
C
C     Calculates exact solution u(x,y) = sin(l_x pi x) sin(l_y pi y)
C     sin(l_z pi y) and corresponding right hand side for
C     laplacian(u) = rhs.
C
C-----------------------------------------------------------------------
C
      SUBROUTINE CLCU(U,H,NX,NY,NZ)
C
         IMPLICIT       NONE
C
         INCLUDE       'mgprob.inc'
C
         INTEGER        NX,       NY,       NZ
         real*8         U(NX,NY,NZ)
C
         real*8         H
C
         INTEGER        I,        J,        K
C
         DO 30 K = 1 , NZ
            DO 20 J = 1 , NY
               DO 10 I = 1 , NX
                  U(I,J,K) = SIN(PI * LX * (I - 1) * H) *
     *                       SIN(PI * LY * (J - 1) * H) *
     *                       SIN(PI * LZ * (K - 1) * H)
 10            CONTINUE
 20         CONTINUE
 30      CONTINUE
         CALL D3LBS(U,0.0d0,NX,NY,NZ)
C
         RETURN
C
      END
C
C-----------------------------------------------------------------------
C
      SUBROUTINE CLCRHS(RHS,H,NX,NY,NZ)
C
         IMPLICIT       NONE
C
         INCLUDE       'mgprob.inc'
C
         INTEGER        NX,       NY,      NZ
         real*8         RHS(NX,NY,NZ)
C
         real*8         H
C
         real*8         UIJK
         INTEGER        I,        J,       K
C
         DO 30 K = 1 , NZ
            DO 20 J = 1 , NY
               DO 10 I = 1 , NX
                  UIJK =
     *                       SIN(PI * LX * (I - 1) * H) *
     *                       SIN(PI * LY * (J - 1) * H) *
     *                       SIN(PI * LZ * (K - 1) * H)
                  RHS(I,J,K) = -(LX * LX + LY * LY + LZ * LZ)
     *                           * PI * PI * UIJK
 10            CONTINUE
 20         CONTINUE
 30      CONTINUE
         CALL D3LBS(RHS,0.0d0,NX,NY,NZ)
C
         RETURN
C
      END

c-----------------------------------------------------------------------
c
c     Basic v-cycle multi grid solver for 
c
c     Laplacian(u) = rhs(u)
c
c     in (x,y,z) coordinates on rectangular domain. 
c
c     This version allows nxf, nyf, nzy which are of form
c     2^p * <small prime> + 1
c
c-----------------------------------------------------------------------

      real*8 function mgvr1
     *   (uf,rhsf,xf,yf,zf,nxf,nyf,nzf, hf, q, sizeq,
     *    preswf,pstswf,presw,pstsw,ncycle,option)

         implicit      none

         real*8        d3nrm2,    relaxc
         integer       qalloc,    lilog2
         logical       lp2np1

         integer       nxf,       nyf,     nzf,      sizeq
         real*8        uf(nxf,nyf,nzf),    rhsf(nxf,nyf,nzf),
     *                 q(sizeq)

         real*8          xf(nxf),   yf(nyf),  zf(nzf)
         real*8        hf
         integer       preswf,    pstswf,  presw,    pstsw,
     *                 ncycle

         integer      option, resb

         integer      bsweeps, nbsweeps
			real*8         beta
			parameter    ( beta = -1.0d0 )
			real*8         alpha

			real*8         work, work_norm
         

         real*8         resnrmob,  relax2bv1 
c
c        Multi-grid parameters, data structures and pointers 
c
         real*8        epsic
         parameter   ( epsic = 1.0d-10 )

         integer       lmin,      lmax
         parameter   ( lmin = 1,  lmax = 20 )

         integer       nx(lmin:lmax),      ny(lmin:lmax),
     *                 nz(lmin:lmax)
         real*8        h(lmin:lmax)

         integer       u(lmin:lmax),       rhs(lmin:lmax),
     *                 mres(lmin:lmax),    du(lmin:lmax)

         integer       x(lmin:lmax),   y(lmin:lmax),  z(lmin:lmax)

         integer       imposs
         parameter   ( imposs = -2 000 000 000 )

         integer          p,      pceil
         common  / comp / p,      pceil
c
c        For parameter (nxf, nyf, nzf) checking 
c
         integer       checkfacts

         logical       badargs

         integer       nfact_x,   nx0
         integer       nfact_y,   ny0
         integer       nfact_z,   nz0

         integer       maxn0
         parameter   ( maxn0 = 11 )
c
c        Locals ...
c
         integer       lxf,       lyf,       lzf,
     *                 l,         lf,        lc,       nl,
     *                 sweep

         real*8        resnrm

         logical       ltrace,    lfulltrace
         parameter   ( ltrace = .false. , lfulltrace = .false. )

         save

         mgvr1 = -1.0d0

			work_norm = 1.0d0/nxf/nyf/nzf

C			open(88,file='bsweeps.dat')
C			read(88,*) nbsweeps
C			read(88,*) alpha
C			close(88)

			nbsweeps = 1

			resb = 1

			alpha = 0.30d0

c
c        Parameter check ...
c

         if( ltrace ) then
            write(0,*) '>>> mgvr1: Parameter dump: '
            write(0,*) nxf, nyf, nzf, hf
            write(0,*) preswf, pstswf, presw, pstsw, ncycle
         end if

         badargs = .false.
         if( checkfacts(nxf - 1,maxn0,nx0,nfact_x) .lt. 0 ) then
            write(*,*) 'mgvr1: Bad nxf value: ', nxf
            badargs = .true.
         end if
         if( checkfacts(nyf - 1,maxn0,ny0,nfact_y) .lt. 0 ) then
            write(*,*) 'mgvr1: Bad nyf value: ', nyf
            badargs = .true.
         end if
         if( checkfacts(nzf - 1,maxn0,nz0,nfact_z) .lt. 0 ) then
            write(*,*) 'mgvr1: Bad nzf value: ', nzf
            badargs = .true.
         end if

         if( badargs) then
            return
         end if

         nl = min(nfact_x,nfact_y,nfact_z) + 1
         lc = 1
         lf = lc + nl - 1

         nx0 = (nxf - 1) / (2 ** (nl - 1)) + 1
         ny0 = (nyf - 1) / (2 ** (nl - 1)) + 1
         nz0 = (nzf - 1) / (2 ** (nl - 1)) + 1
C we must have at least 5 points on the coarsest grid

			do l = 1, 3

				if (nx0.lt.5.or.ny0.lt.5.or.nz0.lt.5) then
					lc = lc + 1
					nx0 = 2*nx0 - 1
					ny0 = 2*ny0 - 1
					nz0 = 2*nz0 - 1
				end if
			end do




         if( ltrace ) then
            write(*,*) 'mgvr1: ', nl, ' levels.'
            write(*,*) 'mgvr1:  Coarse grid ', nx0, ' x ', ny0,
     *                 ' x ', nz0
         end if
         
         if( ncycle .eq. 0 ) then
c
c           Set up grid structure, allocate storage ...
c
            call qreset(sizeq)
            nx(lf) = nxf
            ny(lf) = nyf
            nz(lf) = nzf
            h(lf)  = hf
            u(lf)    = imposs
            rhs(lf)  = imposs
            x(lf)    = imposs
            y(lf)    = imposs
            z(lf)    = imposs
            mres(lf) = qalloc(nx(lf)*ny(lf)*nz(lf))
            du(lf)   = mres(lf)
            do 1000 l = lf - 1 , lc , -1
               nx(l) = (nx(l+1) - 1) / 2 + 1
               ny(l) = (ny(l+1) - 1) / 2 + 1
               nz(l) = (nz(l+1) - 1) / 2 + 1
               h(l) = 2.0d0 * h(l+1)

               u(l)    = qalloc(nx(l)*ny(l)*nz(l))
               rhs(l)  = qalloc(nx(l)*ny(l)*nz(l))
               mres(l) = qalloc(nx(l)*ny(l)*nz(l))
               du(l)   = mres(l)
               x(l)    = qalloc(nx(l))
               y(l)    = qalloc(ny(l))
               z(l)    = qalloc(nz(l))
 1000       continue
            if( ltrace ) then
               call ivdump(nx,nl,'nx',0)
               call ivdump(ny,nl,'ny',0)
               call ivdump(nz,nl,'nz',0)
               call dvdump(h,nl,'h',0)
               call ivdump(u,nl,'u pointers',0)
               call ivdump(rhs,nl,'rhs pointers',0)
               call ivdump(mres,nl,'mres pointers',0)
               call ivdump(du,nl,'du pointers',0)
					call dvdump(xf,nxf,'xf',0)
					call dvdump(yf,nyf,'yf',0)
					call dvdump(zf,nzf,'zf',0)

            end if
         end if


C compute all of x , y  , z

         do l = lf -1 , lc , -1

            call dvumesh(xf(1),xf(nxf),q(x(l)),nx(l),h(l))
            call dvumesh(yf(1),yf(nyf),q(y(l)),ny(l),h(l))
            call dvumesh(zf(1),zf(nzf),q(z(l)),nz(l),h(l))

         end do

c
c        Cycle from fine to coarse ...
c
         do l = lf , lc , -1
            if( l .ne. lf  .or.  ncycle .eq. 0 ) then
               if(      l .eq. lc ) then
 100              continue
                     resnrm = relaxc(q(u(l)),q(rhs(l)),
     *                               nx(l),ny(l),nz(l),h(l))

							work = work + nx(l)*ny(l)*nz(l)*work_norm

                  if (option.eq.1) then
C                     do bsweeps = 1 , nbsweeps 
                        resnrmob=relax2bv1( q(u(l)),q(rhs(l)),
     *                            nx(l),ny(l),nz(l),
     *                            q(x(l)),q(y(l)),q(z(l)),
     *                            h(l),beta,1.0d0 )
C                     end do
                  end if

                     if( lfulltrace ) then
                        call trace_res('mgvr1: pre sweep',l,resnrm)

                     end if
                  if( resnrm .gt. epsic ) go to 100
               else if( l .eq. lf ) then
                  do sweep = 1 , preswf
                     resnrm = relaxc(uf,rhsf,
     *                               nx(l),ny(l),nz(l),h(l))
							work = work + nx(l)*ny(l)*nz(l)*work_norm

							write(57,'(1P,2e10.2)') work,resnrm

                  if (option.eq.1) then
                     do bsweeps = 1 , nbsweeps 
                        resnrmob=relax2bv1( uf,rhsf,
     *                                  nx(l),ny(l),nz(l),
     *                                  xf,yf,zf,
     *                                  h(l),beta,alpha )
                     end do
                  end if

                     call cmres(q(mres(l)),uf,rhsf,
     *                          nx(l),ny(l),nz(l),h(l))

                     if (option.eq.1) then
                     call lu2bv1(q(mres(l)),uf,rhsf,nx(l),ny(l),nz(l),
     *                           xf,yf,zf,h(l),beta)
                     end if

                     if( ltrace ) then
                        call trace_res('mgvr1: pre sweep',l,resnrm)
                     end if
                  end do
               else 
                  do sweep = 1 , presw 
                     resnrm = relaxc(q(u(l)),q(rhs(l)),
     *                               nx(l),ny(l),nz(l),h(l))
							work = work + nx(l)*ny(l)*nz(l)*work_norm

                  if (option.eq.1) then
                     do bsweeps = 1 , 3
                        resnrmob=relax2bv1( q(u(l)),q(rhs(l)),
     *                                      nx(l),ny(l),nz(l),
     *                                      q(x(l)),q(y(l)),q(z(l)),
     *                                      h(l),beta,alpha )
                     end do
                  end if

                     if( ltrace ) then
                        call trace_res('mgvr1: pre sweep',l,resnrm)

                     end if
                  end do
               end if
            end if
c
c           Set up coarse grid problem ...
c
            if(      l .eq. lf ) then
               call cmres(q(mres(l)),uf,rhsf,nx(l),ny(l),nz(l),h(l))
               if (option.eq.1) then
                  call lu2bv1(q(mres(l)),uf,rhsf,nx(l),ny(l),nz(l),
     *                        xf,yf,zf,h(l),beta)
               end if

            else if( l .ne. lc ) then
               call cmres(q(mres(l)),q(u(l)),q(rhs(l)),
     *                    nx(l),ny(l),nz(l),h(l))
               if (option.eq.1) then
                  call lu2bv1(q(mres(l)),q(u(l)),q(rhs(l)),
     *                        nx(l),ny(l),nz(l),
     *                        q(x(l)),q(y(l)),q(z(l)),h(l),beta)
               end if

            end if
            if( l .ne. lc ) then
               call d3rshw(q(mres(l)),q(rhs(l-1)),
     *                     nx(l-1),ny(l-1),nz(l-1))
C put the special boundary restriction in here ....

					if (option.eq.1.and.resb.eq.1) then

						call restrictb( q(rhs(l-1)), q(mres(l)),
     *                 nx(l-1),ny(l-1),nz(l-1),nx(l),ny(l),nz(l))
					end if


C
               call d3ls(q(u(l-1)),0.0d0,
     *                   nx(l-1),ny(l-1),nz(l-1))
            end if

         end do 
c
c        Cycle from coarse to fine ...
c

         do l = lc + 1 , lf 
c
c           Interpolate correction ...
c
            call d3lint(q(u(l-1)),q(du(l)),nx(l-1),ny(l-1),nz(l-1))
            if( l .eq. lf ) then
               call d33a(uf,q(du(l)),uf,nx(l),ny(l),nz(l))
            else 
               call d33a(q(u(l)),q(du(l)),q(u(l)),nx(l),ny(l),nz(l))
            end if

            if( l .eq. lf ) then
               do sweep = 1 , pstswf
                  resnrm = relaxc(uf,rhsf,nx(l),ny(l),nz(l),h(l))
							work = work + nx(l)*ny(l)*nz(l)*work_norm
							write(57,'(1P,2e10.2)') work,resnrm
						if (option.eq.1) then
                     do bsweeps = 1 , 3
                        resnrmob=relax2bv1( uf,rhsf,
     *                                  nx(l),ny(l),nz(l),
     *                                  xf,yf,zf,
     *                                  h(l),beta,alpha )
							end do
						end if
                  if( ltrace ) then
                     call trace_res('mgvr1: post sweep',l,resnrm)

                  end if
               end do
            else 
               do sweep = 1 , pstsw 
                  resnrm = relaxc(q(u(l)),q(rhs(l)),
     *                            nx(l),ny(l),nz(l),h(l))
							work = work + nx(l)*ny(l)*nz(l)*work_norm
                  if (option.eq.1) then
                     do bsweeps = 1 , 3
                        resnrmob=relax2bv1( q(u(l)),q(rhs(l)),
     *                                  nx(l),ny(l),nz(l),
     *                                  q(x(l)),q(y(l)),q(z(l)),
     *                                  h(l),beta,alpha )
                     end do
                  end if

                  if( ltrace ) then
                     call trace_res('mgvr1: post sweep',l,resnrm)

                  end if
               end do
            end if

         end do

         ncycle = ncycle + 1
         mgvr1 = resnrm

         return

      end

c-----------------------------------------------------------------------
c
c     Helper routine for argument checking ...
c
c-----------------------------------------------------------------------

      integer function checkfacts(num,maxbase,base,npow2)

         implicit      none

         integer       ivprimedecomp
         
         integer       num,        maxbase,     base,       npow2

         integer       maxnfact
         parameter   ( maxnfact = 100 )
         integer       vfact(maxnfact)
         integer       nfact
         
         if( ivprimedecomp(num,vfact,nfact) .gt. 0 ) then
            if( vfact(nfact) .le. maxbase ) then
               if( vfact(nfact-1) .eq. 2 ) then
                  base = vfact(nfact) + 1
                  npow2 = nfact - 1
                  checkfacts = npow2
               else 
                  checkfacts = -1
               end if
            else 
               checkfacts = -2
            end if
         else 
            checkfacts = -3
         end if

         return

      end

c==============================================================================
c                         Robin outer boundary package
c                                     by:
c                                 Scott A. Klasky
c                                    9/22/94
c==============================================================================
c      The routine in this package are needed in order to use robin outer
c     boundary conditions.
c     Version 0 will also have slow routines (unoptimized)
c==============================================================================
c    relax2bv1: this routine relaxes the outer robin boundary to second order
c
c==============================================================================
      real*8   function relax2bv1( u, f, nx, ny, nz,
     .                             x, y, z, h,beta, alpha )

      implicit       none
      integer        nx,              ny,            nz
      real*8         u(nx,ny,nz),
     *               f(nx,ny,nz)
      real*8         x(nx), y(ny) , z(nz)
      real*8            h , beta
      integer        i,        j,        k
      integer icb,ioff,joff,koff
      integer ihalf , jhalf, khalf
      real*8   hinv
      real*8   alpha

      real*8   normb
      real*8   u_e

      real*8   uold,dem,num,uold2

      ihalf = (nx+1)/2
      jhalf = (ny+1)/2
      khalf = (nz+1)/2

      hinv = 1.0d0/h
c=======================vary i=================================================
      koff = 1
      do icb = 1 , 2
         joff = koff
cmic$   do all
cmic$*  shared(u, f, x,y,z, nx, ny, nz, alpha, hinv, icb,beta,rhs)
cmic$*  private(i, j, k, ioff,koff, joff )
         do k = 2 , nz - 1

				if ((k/2)*2.eq.k) then
					joff=3-icb
				else
               joff = icb
				end if
cdir$ ivdep
            do j = 1 + joff , ny-1,2

               i = 1

               u(i,j,k)   = u(i,j,k) + alpha * (
     .                    1.0d0/( x(i)*hinv *25.0d0/12.0d0 +beta ) *
     .                    ( y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k) )
     .                    + z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1) )
     .                    + x(i)*hinv*( 4.0d0/3.0d0* u(i+3,j,k)
     .                    - 0.25d0*u(i+4,j,k)
     .                    - 3.0d0*u(i+2,j,k) + 4.0d0*u(i+1,j,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )

               i = nx

               u(i,j,k)   = u(i,j,k) + alpha * (
     .                    1.0d0/(-x(i)*hinv *25.0d0/12.0d0 +beta ) *
     .                    ( y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k) )
     .                    + z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1) )
     .                    + x(i)*hinv*(-4.0d0/3.0d0* u(i-3,j,k)
     .                    + 0.25d0*u(i-4,j,k)
     .                    + 3.0d0*u(i-2,j,k) - 4.0d0*u(i-1,j,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )

            end do
            joff = 3 - joff
         end do
         koff = 3 - koff
      end do
c=======================vary j=================================================
      koff = 1
      do icb = 1 , 2
         ioff = koff
cmic$   do all
cmic$*  shared(u, f, x,y,z, nx, ny, nz, alpha, hinv, icb,beta,rhs)
cmic$*  private(i, j, k, ioff,koff, joff )
         do k = 2  , nz-1

            if ((k/2)*2.eq.k) then
               ioff=3-icb
            else
               ioff = icb
            end if

cdir$ ivdep
            do i = 1+ioff , nx-1  , 2
               j = 1

               u(i,j,k)   = u(i,j,k) + alpha * (
     .                    1.0d0/( y(j)*hinv *25.0d0/12.0d0 +beta ) *
     .                    ( x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k) )
     .                    + z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1) )
     .                    + y(j)*hinv*( 4.0d0/3.0d0* u(i,j+3,k)
     .                    - 0.25d0*u(i,j+4,k)
     .                    - 3.0d0*u(i,j+2,k) + 4.0d0*u(i,j+1,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )

               j = ny

               u(i,j,k)   = u(i,j,k) + alpha * (
     .                    1.0d0/(-y(j)*hinv *25.0d0/12.0d0 +beta ) *
     .                    ( x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k) )
     .                    + z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1) )
     .                    + y(j)*hinv*(-4.0d0/3.0d0* u(i,j-3,k)
     .                    + 0.25d0*u(i,j-4,k)
     .                    + 3.0d0*u(i,j-2,k) - 4.0d0*u(i,j-1,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )

            end do
            ioff = 3 - ioff
         end do
         koff = 3 - koff
      end do
c=======================vary k=================================================
      joff = 1
      do icb = 1 , 2
         ioff = joff
cmic$   do all
cmic$*  shared(u, f, x,y,z, nx, ny, nz, alpha, hinv, icb,beta,rhs)
cmic$*  private(i, j, k, ioff,koff, joff )
         do j = 2  , ny-1
            if ((j/2)*2.eq.j) then
               ioff=3-icb
            else
               ioff = icb
            end if


cdir$ ivdep
            do i = 1+ioff , nx-1 , 2

               k = 1

               u(i,j,k)   = u(i,j,k) + alpha * (
     .                    1.0d0/( z(k)*hinv *25.0d0/12.0d0 +beta ) *
     .                    ( x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k) )
     .                    + y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k) )
     .                    + z(k)*hinv*( 4.0d0/3.0d0* u(i,j,k+3)
     .                    - 0.25d0*u(i,j,k+4)
     .                    - 3.0d0*u(i,j,k+2) + 4.0d0*u(i,j,k+1) )
     .                    - f(i,j,k) ) - u(i,j,k) )

               k = nz

               u(i,j,k)   = u(i,j,k) + alpha * (
     .                    1.0d0/(-z(k)*hinv *25.0d0/12.0d0 +beta ) *
     .                    ( x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k) )
     .                    + y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k) )
     .                    + z(k)*hinv*(-4.0d0/3.0d0* u(i,j,k-3)
     .                    + 0.25d0*u(i,j,k-4)
     .                    + 3.0d0*u(i,j,k-2) - 4.0d0*u(i,j,k-1) )
     .                    - f(i,j,k) ) - u(i,j,k) )


            end do
            ioff = 3 - ioff
         end do
         joff = 3 - joff
      end do

c==============================================================================
c
c 12 lines ....
c  now do the lines a,b,c,d, e,f,g,h, m,n,o,p

      ioff = 1
      do icb = 1 , 2
         do i = 1 + ioff , ny-1 , 2
            j = 1
            k = 1

               u(i,j,k) = u(i,j,k) + alpha * (
     .                    1.0d0/( y(j)*hinv *25.0d0/12.0d0
     .                           +z(k)*hinv *25.0d0/12.0d0 +beta) *
     .                    ( x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k) )
     .                    + z(k)*hinv*( 4.0d0/3.0d0* u(i,j,k+3)
     .                    - 0.25d0*u(i,j,k+4)
     .                    - 3.0d0*u(i,j,k+2) + 4.0d0*u(i,j,k+1) )
     .                    + y(j)*hinv*( 4.0d0/3.0d0* u(i,j+3,k)
     .                    - 0.25d0*u(i,j+4,k)
     .                    - 3.0d0*u(i,j+2,k) + 4.0d0*u(i,j+1,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )


            j = ny
            k = 1

               u(i,j,k) = u(i,j,k) + alpha * (
     .                    1.0d0/(-y(j)*hinv *25.0d0/12.0d0
     .                           +z(k)*hinv *25.0d0/12.0d0 +beta) *
     .                    ( x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k) )
     .                    + z(k)*hinv*( 4.0d0/3.0d0* u(i,j,k+3)
     .                    - 0.25d0*u(i,j,k+4)
     .                    - 3.0d0*u(i,j,k+2) + 4.0d0*u(i,j,k+1) )
     .                    + y(j)*hinv*(-4.0d0/3.0d0* u(i,j-3,k)
     .                    + 0.25d0*u(i,j-4,k)
     .                    + 3.0d0*u(i,j-2,k) - 4.0d0*u(i,j-1,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )




            j = 1
            k = nz

               u(i,j,k) = u(i,j,k) + alpha * (
     .                    1.0d0/( y(j)*hinv *25.0d0/12.0d0
     .                           -z(k)*hinv *25.0d0/12.0d0 +beta) *
     .                    ( x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k) )
     .                    + z(k)*hinv*(-4.0d0/3.0d0* u(i,j,k-3)
     .                    + 0.25d0*u(i,j,k-4)
     .                    + 3.0d0*u(i,j,k-2) - 4.0d0*u(i,j,k-1) )
     .                    + y(j)*hinv*( 4.0d0/3.0d0* u(i,j+3,k)
     .                    - 0.25d0*u(i,j+4,k)
     .                    - 3.0d0*u(i,j+2,k) + 4.0d0*u(i,j+1,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )



            j = ny
            k = nz

               u(i,j,k) = u(i,j,k) + alpha * (
     .                    1.0d0/(-y(j)*hinv *25.0d0/12.0d0
     .                           -z(k)*hinv *25.0d0/12.0d0 +beta) *
     .                    ( x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k) )
     .                    + z(k)*hinv*(-4.0d0/3.0d0* u(i,j,k-3)
     .                    + 0.25d0*u(i,j,k-4)
     .                    + 3.0d0*u(i,j,k-2) - 4.0d0*u(i,j,k-1) )
     .                    + y(j)*hinv*(-4.0d0/3.0d0* u(i,j-3,k)
     .                    + 0.25d0*u(i,j-4,k)
     .                    + 3.0d0*u(i,j-2,k) - 4.0d0*u(i,j-1,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )



         end do
         ioff = 3 - ioff
      end do
c--------------------------------j lines
      joff = 1
      do icb = 1 , 2
         do j = 1 + joff , ny-1 , 2
            i = 1
            k = 1

               u(i,j,k) = u(i,j,k) + alpha * (
     .                    1.0d0/( x(i)*hinv *25.0d0/12.0d0
     .                           +z(k)*hinv *25.0d0/12.0d0 +beta) *
     .                    ( y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k) )
     .                    + z(k)*hinv*( 4.0d0/3.0d0* u(i,j,k+3)
     .                    - 0.25d0*u(i,j,k+4)
     .                    - 3.0d0*u(i,j,k+2) + 4.0d0*u(i,j,k+1) )
     .                    + x(i)*hinv*( 4.0d0/3.0d0* u(i+3,j,k)
     .                    - 0.25d0*u(i+4,j,k)
     .                    - 3.0d0*u(i+2,j,k) + 4.0d0*u(i+1,j,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )



            i = nx
            k = 1

               u(i,j,k) = u(i,j,k) + alpha * (
     .                    1.0d0/(-x(i)*hinv *25.0d0/12.0d0
     .                           +z(k)*hinv *25.0d0/12.0d0 +beta) *
     .                    ( y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k) )
     .                    + z(k)*hinv*( 4.0d0/3.0d0* u(i,j,k+3)
     .                    - 0.25d0*u(i,j,k+4)
     .                    - 3.0d0*u(i,j,k+2) + 4.0d0*u(i,j,k+1) )
     .                    + x(i)*hinv*(-4.0d0/3.0d0* u(i-3,j,k)
     .                    + 0.25d0*u(i-4,j,k)
     .                    + 3.0d0*u(i-2,j,k) - 4.0d0*u(i-1,j,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )



            i = 1
            k = nz

               u(i,j,k) = u(i,j,k) + alpha * (
     .                    1.0d0/( x(i)*hinv *25.0d0/12.0d0
     .                           -z(k)*hinv *25.0d0/12.0d0 +beta) *
     .                    ( y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k) )
     .                    + z(k)*hinv*(-4.0d0/3.0d0* u(i,j,k-3)
     .                    + 0.25d0*u(i,j,k-4)
     .                    + 3.0d0*u(i,j,k-2) - 4.0d0*u(i,j,k-1) )
     .                    + x(i)*hinv*( 4.0d0/3.0d0* u(i+3,j,k)
     .                    - 0.25d0*u(i+4,j,k)
     .                    - 3.0d0*u(i+2,j,k) + 4.0d0*u(i+1,j,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )


            i = nx
            k = nz

               u(i,j,k) = u(i,j,k) + alpha * (
     .                    1.0d0/(-x(i)*hinv *25.0d0/12.0d0
     .                           -z(k)*hinv *25.0d0/12.0d0 +beta) *
     .                    ( y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k) )
     .                    + z(k)*hinv*(-4.0d0/3.0d0* u(i,j,k-3)
     .                    + 0.25d0*u(i,j,k-4)
     .                    + 3.0d0*u(i,j,k-2) - 4.0d0*u(i,j,k-1) )
     .                    + x(i)*hinv*(-4.0d0/3.0d0* u(i-3,j,k)
     .                    + 0.25d0*u(i-4,j,k)
     .                    + 3.0d0*u(i-2,j,k) - 4.0d0*u(i-1,j,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )


         end do
         joff = 3 - joff
      end do
c---------------------------do the k lines

      koff = 1
      do icb = 1 , 2
         do k = 1 + koff , nz-1 , 2
            i = 1
            j = 1

               u(i,j,k) = u(i,j,k) + alpha * (
     .                    1.0d0/( x(i)*hinv *25.0d0/12.0d0
     .                           +y(j)*hinv *25.0d0/12.0d0 +beta) *
     .                    ( z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1) )
     .                    + y(j)*hinv*( 4.0d0/3.0d0* u(i,j+3,k)
     .                    - 0.25d0*u(i,j+4,k)
     .                    - 3.0d0*u(i,j+2,k) + 4.0d0*u(i,j+1,k) )
     .                    + x(i)*hinv*( 4.0d0/3.0d0* u(i+3,j,k)
     .                    - 0.25d0*u(i+4,j,k)
     .                    - 3.0d0*u(i+2,j,k) + 4.0d0*u(i+1,j,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )



            i = nx
            j = 1

               u(i,j,k) = u(i,j,k) + alpha * (
     .                    1.0d0/(-x(i)*hinv *25.0d0/12.0d0
     .                           +y(j)*hinv *25.0d0/12.0d0 +beta) *
     .                    ( z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1) )
     .                    + y(j)*hinv*( 4.0d0/3.0d0* u(i,j+3,k)
     .                    - 0.25d0*u(i,j+4,k)
     .                    - 3.0d0*u(i,j+2,k) + 4.0d0*u(i,j+1,k) )
     .                    + x(i)*hinv*(-4.0d0/3.0d0* u(i-3,j,k)
     .                    + 0.25d0*u(i-4,j,k)
     .                    + 3.0d0*u(i-2,j,k) - 4.0d0*u(i-1,j,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )

            i = 1
            j = ny

               u(i,j,k) = u(i,j,k) + alpha * (
     .                    1.0d0/( x(i)*hinv *25.0d0/12.0d0
     .                           -y(j)*hinv *25.0d0/12.0d0 +beta) *
     .                    ( z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1) )
     .                    + y(j)*hinv*(-4.0d0/3.0d0* u(i,j-3,k)
     .                    + 0.25d0*u(i,j-4,k)
     .                    + 3.0d0*u(i,j-2,k) - 4.0d0*u(i,j-1,k) )
     .                    + x(i)*hinv*( 4.0d0/3.0d0* u(i+3,j,k)
     .                    - 0.25d0*u(i+4,j,k)
     .                    - 3.0d0*u(i+2,j,k) + 4.0d0*u(i+1,j,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )

            i = nx
            j = ny

               u(i,j,k) = u(i,j,k) + alpha * (
     .                    1.0d0/(-x(i)*hinv *25.0d0/12.0d0
     .                           -y(j)*hinv *25.0d0/12.0d0 +beta) *
     .                    ( z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1) )
     .                    + y(j)*hinv*(-4.0d0/3.0d0* u(i,j-3,k)
     .                    + 0.25d0*u(i,j-4,k)
     .                    + 3.0d0*u(i,j-2,k) - 4.0d0*u(i,j-1,k) )
     .                    + x(i)*hinv*(-4.0d0/3.0d0* u(i-3,j,k)
     .                    + 0.25d0*u(i-4,j,k)
     .                    + 3.0d0*u(i-2,j,k) - 4.0d0*u(i-1,j,k) )
     .                    - f(i,j,k) ) - u(i,j,k) )



         end do
         koff = 3 - koff
      end do

c now we must do the eight points ............

      i = 1
      j = 1
      k = 1
               u(i,j,k)   = u(i,j,k)  + alpha*(
     .                    u_e(nx,ny,nz,i,j,k,h,x,y,z,beta,f,u)
     .                    -u(i,j,k))

      i = 1
      j = 1
      k = nz


               u(i,j,k)   = u(i,j,k)  + alpha*(
     .                    u_e(nx,ny,nz,i,j,k,h,x,y,z,beta,f,u)
     .                     -u(i,j,k))


      i = 1
      j = ny
      k = 1


               u(i,j,k)   = u(i,j,k)  + alpha*(
     .                    u_e(nx,ny,nz,i,j,k,h,x,y,z,beta,f,u)
     .                    -u(i,j,k))

      i = 1
      j = ny
      k = nz

               u(i,j,k)   = u(i,j,k)  + alpha*(
     .                    u_e(nx,ny,nz,i,j,k,h,x,y,z,beta,f,u)
     .                   -u(i,j,k))

      i = nx
      j = 1
      k = 1

               u(i,j,k)   = u(i,j,k)  + alpha*(
     .                    u_e(nx,ny,nz,i,j,k,h,x,y,z,beta,f,u)
     .                   -u(i,j,k))


      i = nx
      j = 1
      k = nz

               u(i,j,k)   = u(i,j,k)  + alpha*(
     .                    u_e(nx,ny,nz,i,j,k,h,x,y,z,beta,f,u)
     .                   -u(i,j,k))

      i = nx
      j = ny
      k = 1

               u(i,j,k)   = u(i,j,k)  + alpha*(
     .                    u_e(nx,ny,nz,i,j,k,h,x,y,z,beta,f,u)
     .                   -u(i,j,k))


      i = nx
      j = ny
      k = nz


               u(i,j,k)   = u(i,j,k)  + alpha*(
     .                    u_e(nx,ny,nz,i,j,k,h,x,y,z,beta,f,u)
     .                   -u(i,j,k))

      relax2bv1 = 0.0d0

      return
      end
c==============================================================================
      real*8   function u_e( nx,ny,nz,i,j,k,h,x,y,z,beta,f,u)
      implicit none
      integer nx,ny,nz
      integer i , j, k
      real*8   h,hinv
      real*8   x(nx),y(ny),z(nz)
      real*8   beta
      real*8   f(nx,ny,nz),u(nx,ny,nz)

      real*8   num , dem
      real*8   eval_w1i,eval_w2i,eval_w1j,eval_w2j,eval_w1k,eval_w2k


      hinv = 1.0d0/h
      num = 0.0d0
      dem = 0.0d0

c place an if here ...........

      if (i.eq.1) then
         num = num + x(i)*hinv*( 4.0d0/3.0d0*u(i+3,j,k)
     .             - 0.25d0*u(i+4,j,k)
     .             - 3.0d0*u(i+2,j,k) + 4.0d0*u(i+1,j,k) )
         dem = dem + x(i)*hinv * 25.0d0/12.0d0
      else if (i.eq.nx) then
         num = num + x(i)*hinv*(-4.0d0/3.0d0*u(i-3,j,k)
     .             + 0.25d0*u(i-4,j,k)
     .             + 3.0d0*u(i-2,j,k) - 4.0d0*u(i-1,j,k) )
         dem = dem - x(i)*hinv * 25.0d0/12.0d0

      else
         num = num + 0.5d0*hinv * x(i) *( u(i+1,j,k) - u(i-1,j,k) )
         dem = dem + 0.0d0

      end if

      if (j.eq.1) then
         num = num + y(j)*hinv*( 4.0d0/3.0d0*u(i,j+3,k)
     .             - 0.25d0*u(i,j+4,k)
     .             - 3.0d0*u(i,j+2,k) + 4.0d0*u(i,j+1,k) )
         dem = dem + y(j)*hinv * 25.0d0/12.0d0
      else if (j.eq.ny) then
         num = num + y(j)*hinv*(-4.0d0/3.0d0*u(i,j-3,k)
     .             + 0.25d0*u(i,j-4,k)
     .             + 3.0d0*u(i,j-2,k) - 4.0d0*u(i,j-1,k) )
         dem = dem - y(j)*hinv * 25.0d0/12.0d0

      else
         num = num + 0.5d0*hinv * y(j) * ( u(i,j+1,k) - u(i,j-1,k) )
         dem = dem + 0.0d0
      end if

      if (k.eq.1) then
         num = num + z(k)*hinv*( 4.0d0/3.0d0*u(i,j,k+3)
     .             - 0.25d0*u(i,j,k+4)
     .             - 3.0d0*u(i,j,k+2) + 4.0d0*u(i,j,k+1) )
         dem = dem + z(k)*hinv * 25.0d0/12.0d0
      else if (k.eq.nz) then
         num = num + z(k)*hinv*(-4.0d0/3.0d0*u(i,j,k-3)
     .             + 0.25d0*u(i,j,k-4)
     .             + 3.0d0*u(i,j,k-2) - 4.0d0*u(i,j,k-1) )
         dem = dem - z(k)*hinv * 25.0d0/12.0d0

      else
         num = num + 0.5d0*hinv * z(k) * ( u(i,j,k+1) - u(i,j,k-1) )
         dem = dem + 0.0d0
      end if

      num = num - f(i,j,k)

      dem = dem + beta

      u_e = num / dem

      return
      end
c==============================================================================
c    lub2: this routine computes - the residual on the outer boundary 
c
c==============================================================================
      subroutine lu2bv1( res, u, rhs,nx, ny, nz,
     .                      x, y, z, h,beta )

      implicit       none
      integer        nx,              ny,            nz
      real*8            res(nx,ny,nz),  u(nx,ny,nz), rhs(nx,ny,nz)
      real*8           x(nx), y(ny) , z(nz)
      real*8            h , beta
      integer        i,        j,        k
      integer icb,ioff,joff,koff
      real*8   hinv

      real*8   normb
      real*8   lub_e

      hinv = 1.0d0/h
c=======================vary i=================================================
cmic$   do all 
cmic$*  shared(res, u,x,y,z, nx, ny, nz, hinv, beta,rhs,rhs)
cmic$*  private(i, j, k)
         do k = 2 , nz - 1
cdir$ ivdep
            do j = 2 , ny-1

               i = 1
               res(i,j,k) = rhs(i,j,k) - (
     .                    (-x(i)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k))
     .                    + z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1))
     .                    + x(i)*hinv*( 4.0d0/3.0d0* u(i+3,j,k)
     .                    - 0.25d0*u(i+4,j,k)
     .                    - 3.0d0*u(i+2,j,k) + 4.0d0*u(i+1,j,k) ) )

               i = nx

               res(i,j,k) =rhs(i,j,k) - (
     .                    (+x(i)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k))
     .                    + z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1))
     .                    + x(i)*hinv*(-4.0d0/3.0d0* u(i-3,j,k)
     .                    + 0.25d0*u(i-4,j,k)
     .                    + 3.0d0*u(i-2,j,k) - 4.0d0*u(i-1,j,k) ) )


            end do
         end do
c=======================vary j=================================================
cmic$   do all 
cmic$*  shared(res, u,x,y,z, nx, ny, nz, hinv, beta,rhs,rhs)
cmic$*  private(i, j, k)
         do k = 2  , nz-1
            do i = 2, nx-1
               j = 1

               res(i,j,k) = rhs(i,j,k) - (
     .                    (-y(j)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k))
     .                    + z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1))
     .                    + y(j)*hinv*( 4.0d0/3.0d0* u(i,j+3,k)
     .                    - 0.25d0*u(i,j+4,k)
     .                    - 3.0d0*u(i,j+2,k) + 4.0d0*u(i,j+1,k) ) )

               j = ny
               res(i,j,k) = rhs(i,j,k) - (
     .                    (+y(j)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k))
     .                    + z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1))
     .                    + y(j)*hinv*(-4.0d0/3.0d0* u(i,j-3,k)
     .                    + 0.25d0*u(i,j-4,k)
     .                    + 3.0d0*u(i,j-2,k) - 4.0d0*u(i,j-1,k) ) )

            end do
         end do
c=======================vary k=================================================
cmic$   do all 
cmic$*  shared(res, u,x,y,z, nx, ny, nz, hinv, beta,rhs)
cmic$*  private(i, j, k)
         do j = 2  , ny-1
            do i = 2, nx-1

               k = 1
               res(i,j,k) = rhs(i,j,k) -  (
     .                    (-z(k)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k))
     .                    + y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k))
     .                    + z(k)*hinv*( 4.0d0/3.0d0* u(i,j,k+3)
     .                    - 0.25d0*u(i,j,k+4)
     .                    - 3.0d0*u(i,j,k+2) + 4.0d0*u(i,j,k+1) ) )

               k = nz
               res(i,j,k) =rhs(i,j,k) - (
     .                    (+z(k)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k))
     .                    + y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k))
     .                    + z(k)*hinv*(-4.0d0/3.0d0* u(i,j,k-3)
     .                    + 0.25d0*u(i,j,k-4)
     .                    + 3.0d0*u(i,j,k-2) - 4.0d0*u(i,j,k-1) ) )

            end do
         end do
c==============================================================================
c
c 12 lines ....
c  now do the lines a,b,c,d, e,f,g,h, m,n,o,p

         do i = 2 , ny-1
            j = 1
            k = 1

               res(i,j,k) =rhs(i,j,k) - ( 
     .                    (-z(k)*hinv *25.0d0/12.0d0
     .                      -y(j)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k))
     .                    + y(j)*hinv*( 4.0d0/3.0d0* u(i,j+3,k)
     .                    - 0.25d0*u(i,j+4,k)
     .                    - 3.0d0*u(i,j+2,k) + 4.0d0*u(i,j+1,k) )
     .                    + z(k)*hinv*( 4.0d0/3.0d0* u(i,j,k+3)
     .                    - 0.25d0*u(i,j,k+4)
     .                    - 3.0d0*u(i,j,k+2) + 4.0d0*u(i,j,k+1) ) )


            j = ny
            k = 1

               res(i,j,k) =rhs(i,j,k) - (
     .                    (-z(k)*hinv *25.0d0/12.0d0
     .                      +y(j)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k))
     .                    + y(j)*hinv*(-4.0d0/3.0d0* u(i,j-3,k)
     .                    + 0.25d0*u(i,j-4,k)
     .                    + 3.0d0*u(i,j-2,k) - 4.0d0*u(i,j-1,k) )
     .                    + z(k)*hinv*( 4.0d0/3.0d0* u(i,j,k+3)
     .                    - 0.25d0*u(i,j,k+4)
     .                    - 3.0d0*u(i,j,k+2) + 4.0d0*u(i,j,k+1) ) )


            j = 1
            k = nz

               res(i,j,k) =rhs(i,j,k) - (
     .                    ( z(k)*hinv *25.0d0/12.0d0
     .                      -y(j)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k))
     .                    + y(j)*hinv*( 4.0d0/3.0d0* u(i,j+3,k)
     .                    - 0.25d0*u(i,j+4,k)
     .                    - 3.0d0*u(i,j+2,k) + 4.0d0*u(i,j+1,k) )
     .                    + z(k)*hinv*(-4.0d0/3.0d0* u(i,j,k-3)
     .                    + 0.25d0*u(i,j,k-4)
     .                    + 3.0d0*u(i,j,k-2) - 4.0d0*u(i,j,k-1) ) )



            j = ny
            k = nz

               res(i,j,k) =rhs(i,j,k) - ( 
     .                     ( z(k)*hinv *25.0d0/12.0d0
     .                      +y(j)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + x(i)*0.5d0*hinv*(u(i+1,j,k)-u(i-1,j,k))
     .                    + y(j)*hinv*(-4.0d0/3.0d0* u(i,j-3,k)
     .                    + 0.25d0*u(i,j-4,k)
     .                    + 3.0d0*u(i,j-2,k) - 4.0d0*u(i,j-1,k) )
     .                    + z(k)*hinv*(-4.0d0/3.0d0* u(i,j,k-3)
     .                    + 0.25d0*u(i,j,k-4)
     .                    + 3.0d0*u(i,j,k-2) - 4.0d0*u(i,j,k-1) ) )


         end do
c--------------------------------j lines
         do j = 2 , ny-1
            i = 1
            k = 1

               res(i,j,k) =rhs(i,j,k) - (
     .                    (-z(k)*hinv *25.0d0/12.0d0
     .                      -x(i)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k))
     .                    + x(i)*hinv*( 4.0d0/3.0d0* u(i+3,j,k)
     .                    - 0.25d0*u(i+4,j,k)
     .                    - 3.0d0*u(i+2,j,k) + 4.0d0*u(i+1,j,k) )
     .                    + z(k)*hinv*( 4.0d0/3.0d0* u(i,j,k+3)
     .                    - 0.25d0*u(i,j,k+4)
     .                    - 3.0d0*u(i,j,k+2) + 4.0d0*u(i,j,k+1) ) )

            i = nx
            k = 1

               res(i,j,k) =rhs(i,j,k) - (
     .                      (-z(k)*hinv *25.0d0/12.0d0
     .                      +x(i)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k))
     .                    + x(i)*hinv*(-4.0d0/3.0d0* u(i-3,j,k)
     .                    + 0.25d0*u(i-4,j,k)
     .                    + 3.0d0*u(i-2,j,k) - 4.0d0*u(i-1,j,k) )
     .                    + z(k)*hinv*( 4.0d0/3.0d0* u(i,j,k+3)
     .                    - 0.25d0*u(i,j,k+4)
     .                    - 3.0d0*u(i,j,k+2) + 4.0d0*u(i,j,k+1) ) )

            i = 1
            k = nz


               res(i,j,k) =rhs(i,j,k) - (
     .                     ( z(k)*hinv *25.0d0/12.0d0
     .                      -x(i)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k))
     .                    + x(i)*hinv*( 4.0d0/3.0d0* u(i+3,j,k)
     .                    - 0.25d0*u(i+4,j,k)
     .                    - 3.0d0*u(i+2,j,k) + 4.0d0*u(i+1,j,k) )
     .                    + z(k)*hinv*(-4.0d0/3.0d0* u(i,j,k-3)
     .                    + 0.25d0*u(i,j,k-4)
     .                    + 3.0d0*u(i,j,k-2) - 4.0d0*u(i,j,k-1) ) )


            i = nx
            k = nz


               res(i,j,k) = rhs(i,j,k) - (
     .                     ( z(k)*hinv *25.0d0/12.0d0
     .                      +x(i)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + y(j)*0.5d0*hinv*(u(i,j+1,k)-u(i,j-1,k))
     .                    + x(i)*hinv*(-4.0d0/3.0d0* u(i-3,j,k)
     .                    + 0.25d0*u(i-4,j,k)
     .                    + 3.0d0*u(i-2,j,k) - 4.0d0*u(i-1,j,k) )
     .                    + z(k)*hinv*(-4.0d0/3.0d0* u(i,j,k-3)
     .                    + 0.25d0*u(i,j,k-4)
     .                    + 3.0d0*u(i,j,k-2) - 4.0d0*u(i,j,k-1) ) )


         end do
c---------------------------do the k lines

         do k = 2 , nz-1
            i = 1
            j = 1

               res(i,j,k) =rhs(i,j,k) - (
     .                    (-y(j)*hinv *25.0d0/12.0d0
     .                      -x(i)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1))
     .                    + x(i)*hinv*( 4.0d0/3.0d0* u(i+3,j,k)
     .                    - 0.25d0*u(i+4,j,k)
     .                    - 3.0d0*u(i+2,j,k) + 4.0d0*u(i+1,j,k) )
     .                    + y(j)*hinv*( 4.0d0/3.0d0* u(i,j+3,k)
     .                    - 0.25d0*u(i,j+4,k)
     .                    - 3.0d0*u(i,j+2,k) + 4.0d0*u(i,j+1,k) ) )



            i = nx
            j = 1

               res(i,j,k) =rhs(i,j,k) - (
     .                    (-y(j)*hinv *25.0d0/12.0d0
     .                      +x(i)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1))
     .                    + x(i)*hinv*(-4.0d0/3.0d0* u(i-3,j,k)
     .                    + 0.25d0*u(i-4,j,k)
     .                    + 3.0d0*u(i-2,j,k) - 4.0d0*u(i-1,j,k) )
     .                    + y(j)*hinv*( 4.0d0/3.0d0* u(i,j+3,k)
     .                    - 0.25d0*u(i,j+4,k)
     .                    - 3.0d0*u(i,j+2,k) + 4.0d0*u(i,j+1,k) ) )

            i = 1
            j = ny

               res(i,j,k) =rhs(i,j,k) -  (
     .                    ( y(j)*hinv *25.0d0/12.0d0
     .                      -x(i)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1))
     .                    + x(i)*hinv*( 4.0d0/3.0d0* u(i+3,j,k)
     .                    - 0.25d0*u(i+4,j,k)
     .                    - 3.0d0*u(i+2,j,k) + 4.0d0*u(i+1,j,k) )
     .                    + y(j)*hinv*(-4.0d0/3.0d0* u(i,j-3,k)
     .                    + 0.25d0*u(i,j-4,k)
     .                    + 3.0d0*u(i,j-2,k) - 4.0d0*u(i,j-1,k) ) )


            i = nx
            j = ny

               res(i,j,k) =rhs(i,j,k) - (
     .                    ( y(j)*hinv *25.0d0/12.0d0
     .                      +x(i)*hinv *25.0d0/12.0d0 -beta )*u(i,j,k)
     .                    + z(k)*0.5d0*hinv*(u(i,j,k+1)-u(i,j,k-1))
     .                    + x(i)*hinv*(-4.0d0/3.0d0* u(i-3,j,k)
     .                    + 0.25d0*u(i-4,j,k)
     .                    + 3.0d0*u(i-2,j,k) - 4.0d0*u(i-1,j,k) )
     .                    + y(j)*hinv*(-4.0d0/3.0d0* u(i,j-3,k)
     .                    + 0.25d0*u(i,j-4,k)
     .                    + 3.0d0*u(i,j-2,k) - 4.0d0*u(i,j-1,k) ) )



         end do

c now we must do the eight points ............

      i = 1
      j = 1
      k = 1
               res(i,j,k) = rhs(i,j,k) -
     .                       lub_e(nx,ny,nz,i,j,k,h,x,y,z,beta,u)

      i = 1
      j = 1
      k = nz


               res(i,j,k) = rhs(i,j,k) - 
     .                      lub_e(nx,ny,nz,i,j,k,h,x,y,z,beta,u)


      i = 1
      j = ny
      k = 1


               res(i,j,k) = rhs(i,j,k) -
     .                      lub_e(nx,ny,nz,i,j,k,h,x,y,z,beta,u)

      i = 1
      j = ny
      k = nz

               res(i,j,k) = rhs(i,j,k) -
     .                      lub_e(nx,ny,nz,i,j,k,h,x,y,z,beta,u)

      i = nx
      j = 1
      k = 1

               res(i,j,k) = rhs(i,j,k) -
     .                      lub_e(nx,ny,nz,i,j,k,h,x,y,z,beta,u)


      i = nx
      j = 1
      k = nz

               res(i,j,k) = rhs(i,j,k) -
     .                      lub_e(nx,ny,nz,i,j,k,h,x,y,z,beta,u)

      i = nx
      j = ny
      k = 1
               res(i,j,k) = rhs(i,j,k) -
     .                      lub_e(nx,ny,nz,i,j,k,h,x,y,z,beta,u)


      i = nx
      j = ny
      k = nz


               res(i,j,k) = rhs(i,j,k) -
     .                      lub_e(nx,ny,nz,i,j,k,h,x,y,z,beta,u)

      return
      end
c==============================================================================
      real*8   function lub_e( nx,ny,nz,i,j,k,h,x,y,z,beta,u)
      implicit none
      integer nx,ny,nz
      integer i , j, k
      real*8   h
      real*8   x(nx),y(ny),z(nz)
      real*8   beta
      real*8   u(nx,ny,nz)
      real*8   num
      real*8   hinv

      hinv = 1.0d0/h
      num = 0.0d0

      if (i.eq.1) then
         num = num + x(i)*hinv*( 4.0d0/3.0d0*u(i+3,j,k)
     .             -25.0d0/12.0d0*u(i,j,k) - 0.25d0*u(i+4,j,k)
     .             - 3.0d0*u(i+2,j,k) + 4.0d0*u(i+1,j,k) )
      else if (i.eq.nx) then
         num = num + x(i)*hinv*(-4.0d0/3.0d0*u(i-3,j,k)
     .             +25.0d0/12.0d0*u(i,j,k) + 0.25d0*u(i-4,j,k)
     .             + 3.0d0*u(i-2,j,k) - 4.0d0*u(i-1,j,k) )
      else
         num = num + 0.5d0*hinv * x(i) * ( u(i+1,j,k) - u(i-1,j,k) )
      end if

      if (j.eq.1) then
         num = num + y(j)*hinv*( 4.0d0/3.0d0*u(i,j+3,k)
     .             -25.0d0/12.0d0*u(i,j,k) - 0.25d0*u(i,j+4,k)
     .             - 3.0d0*u(i,j+2,k) + 4.0d0*u(i,j+1,k) )
      else if (j.eq.ny) then
         num = num + y(j)*hinv*(-4.0d0/3.0d0*u(i,j-3,k)
     .             +25.0d0/12.0d0*u(i,j,k) + 0.25d0*u(i,j-4,k)
     .             + 3.0d0*u(i,j-2,k) - 4.0d0*u(i,j-1,k) )
      else
         num = num + 0.5d0*hinv * y(j)*( u(i,j+1,k) - u(i,j-1,k) )
      end if

      if (k.eq.1) then
         num = num + z(k)*hinv*( 4.0d0/3.0d0*u(i,j,k+3)
     .             -25.0d0/12.0d0*u(i,j,k) - 0.25d0*u(i,j,k+4)
     .             - 3.0d0*u(i,j,k+2) + 4.0d0*u(i,j,k+1) )
      else if (k.eq.nz) then
         num = num + z(k)*hinv*(-4.0d0/3.0d0*u(i,j,k-3)
     .             +25.0d0/12.0d0*u(i,j,k) + 0.25d0*u(i,j,k-4)
     .             + 3.0d0*u(i,j,k-2) - 4.0d0*u(i,j,k-1) )
      else
         num = num + 0.5d0*hinv*z(k) * ( u(i,j,k+1) - u(i,j,k-1) )
      end if

      lub_e = num - beta*u(i,j,k)


      return
      end
c here we transfer the boundaries ......
c
c
C treat the 6 faces of the cube as 2d domains, and use `typical'
c 2d transfers on these faces
c
c
      subroutine restrictb( ac,af,nxc, nyc, nzc, nxf, nyf, nzf)

      implicit none
      integer nxc, nyc , nzc
      integer nxf, nyf, nzf
      real*8     ac(nxc,nyc,nzc)
      real*8     af(nxf,nyf,nzf)

      real*8            half,       twelth
      parameter    ( half = 0.5d0,
     *               twelth = 8.333333333333333d-02 )


      integer   ic , jc , kc
      integer   ifp, jfp, kfp
      integer   ic1,ic2,jc1,jc2,kc1,kc2
      integer   ifp1,ifp2,jfp1,jfp2,kfp1,kfp2

c faces 1,2 .....

      kc1   = 1
      kfp1  = 1
      kc2   = nzc
      kfp2  = nzf

      do ic = 2, nxc-1

         ifp = 2*ic - 1

         do jc=2,nyc-1

            jfp = 2*jc - 1

            ac(ic,jc,kc1) = 0.5d0*af(ifp,jfp,kfp1)              +
     .                 0.125d0*(  af(ifp+1,jfp,kfp1)            +
     .                 af(ifp-1,jfp,kfp1) + af(ifp,jfp+1,kfp1)  +
     .                 af(ifp,jfp-1,kfp1)                      )

            ac(ic,jc,kc2) = 0.5d0*af(ifp,jfp,kfp2)              +
     .                 0.125d0*(  af(ifp+1,jfp,kfp2)            +
     .                 af(ifp-1,jfp,kfp2) + af(ifp,jfp+1,kfp2)  +
     .                 af(ifp,jfp-1,kfp2)                       )

         end do
      end do

c faces 3,4 .....

      jc1   = 1
      jfp1  = 1
      jc2   = nyc
      jfp2  = nyf

      do ic = 2, nxc-1

         ifp = 2*ic - 1

         do kc=2,nzc-1

            kfp = 2*kc - 1

            ac(ic,jc1,kc) = 0.5d0*af(ifp,jfp1,kfp)              +
     .                 0.125d0*(  af(ifp+1,jfp1,kfp)            +
     .                 af(ifp-1,jfp1,kfp) + af(ifp,jfp1,kfp+1)  +
     .                 af(ifp,jfp1,kfp-1)                       )

            ac(ic,jc2,kc) = 0.5d0*af(ifp,jfp2,kfp)              +
     .                 0.125d0*(  af(ifp+1,jfp2,kfp)            +
     .                 af(ifp-1,jfp2,kfp) + af(ifp,jfp2,kfp+1)  +
     .                 af(ifp,jfp2,kfp-1)                       )

         end do
      end do

c faces 5,6 ....

      ic1   = 1
      ifp1  = 1
      ic2   = nxc
      ifp2  = nxf

      do jc = 2, nyc-1

         jfp = 2*jc - 1

         do kc=2,nzc-1

            kfp = 2*kc - 1

            ac(ic1,jc,kc) = 0.5d0*af(ifp1,jfp,kfp)              +
     .                 0.125d0*(  af(ifp1,jfp+1,kfp)            +
     .                 af(ifp1,jfp-1,kfp) + af(ifp1,jfp,kfp+1)  +
     .                 af(ifp1,jfp,kfp-1)                       )

            ac(ic2,jc,kc) = 0.5d0*af(ifp2,jfp,kfp)              +
     .                 0.125d0*(  af(ifp2,jfp+1,kfp)            +
     .                 af(ifp2,jfp-1,kfp) + af(ifp2,jfp,kfp+1)  +
     .                 af(ifp2,jfp,kfp-1)                       )

         end do
      end do


      return
      end
