dnl ======================================================================
dnl autoconf macros for the BBH project
dnl
dnl Copyright 1995 Matthew William Choptuik
dnl The University of Texas at Austin
dnl
dnl ======================================================================

dnl ----------------------------------------------------------------------
dnl BBH_SYS_GETSYSTEM attempts to determine what flavour of UNIX
dnl is being used
dnl ----------------------------------------------------------------------

dnl ----------------------------------------------------------------------
dnl Template for new macros
dnl ----------------------------------------------------------------------

AC_DEFUN(BBH_TEMPLATE,[
])

dnl ----------------------------------------------------------------------
dnl BBH_SYS_GETSYSTEM attempts to determine what flavour of UNIX
dnl is being used
dnl 
dnl Sets Output Variable BBH_SYSTEM and possibly others, dependent on 
dnl the pecularities of the given OS.
dnl ----------------------------------------------------------------------

AC_DEFUN(BBH_SYS_GETSYSTEM,[
   if test "X$BBH_SYSTEM" = "X"; then
      AC_MSG_CHECKING(for Unix flavour)
      AC_CACHE_VAL(ac_cv_bbh_system,
      [
         changequote(,)dnl
         __UNAME=`uname -a | tr '[a-z]' '[A-Z]'`
         changequote([, ])dnl
         for s in $__UNAME
         do
            case $s in
            IRIX*)                        BBH_SYSTEM=IRIX;        
                                          F77_TRANSFORM="touch";
                                                                  break;;
            SUNOS)                        BBH_SYSTEM=SUNOS;       
                                          F77_TRANSFORM="touch";
                 LDFLAGS="$LDFLAGS -L/usr/ccs/lib";               
                                                                  break;;
            AIX)                          BBH_SYSTEM=AIX;        
                                          F77_TRANSFORM="touch";
                                                                  break;;
            A/UX)                         BBH_SYSTEM=A/UX;       
                                          F77_TRANSFORM="touch";
                                                                  break;;
            OSF1)                         BBH_SYSTEM=OSF1;        
                                          F77_TRANSFORM="touch";
                 BBH_DEFS="$BBH_DEFS -std1 -DDEC_ALPHA";
                                                                  break;;
            UNICOS|CRAY)                  BBH_SYSTEM=UNICOS;      
                                          F77_TRANSFORM="../bin/f77transcray";
                 BBH_RNPL_FLAGS="-csufE";
                                                                  break;;
            HP-UX)                        BBH_SYSTEM=HP-UX;       
                                          F77_TRANSFORM="touch";
                 BBH_DEFS="$BBH_DEFS -DHP9000";
                 CFLAGS="$CFLAGS -Aa -D_HPUX_SOURCE -w";        
                                                                  break;;
            *)                            BBH_SYSTEM=UNKNOWN           ;;
            esac
         done
         ac_cv_bbh_system="$BBH_SYSTEM"
         ac_cv_bbh_f77_transform="$F77_TRANSFORM";
      ]
      )
      AC_MSG_RESULT($ac_cv_bbh_system)
      BBH_SYSTEM=$ac_cv_bbh_system
      F77_TRANSFORM=$ac_cv_bbh_f77_transform
      AC_SUBST(BBH_SYSTEM)
      AC_SUBST(F77_TRANSFORM)
		case $BBH_SYSTEM in
		SUNOS)
			AC_CHECK_LIB(nsl, clnt_create,
			  LIBS="$LIBS -lnsl"
			  BBH_RNPLAPP_FLIBS="$BBH_RNPLAPP_FLIBS -lnsl"
			  BBH_RNPLAPP_CLIBS="$BBH_RNPLAPP_CLIBS -lnsl"
			  )
																						break;;
		esac
   fi
])

dnl ----------------------------------------------------------------------
dnl Locate lib, include paths for auxiliary software needed by RNPL
dnl ----------------------------------------------------------------------

AC_DEFUN(BBH_CHECK_RNPL_LIBS,[
   BBH_CHECK_FATAL="no"
   BBH_RNPLAPP_LIBPATHS="$BBH_RNPLAPP_LIBPATHS $prefix/lib"
   BBH_RNPLAPP_INCPATHS="$BBH_RNPLAPP_INCPATHS $prefix/include"
   BBH_RNPLAPP_FLIBS="$BBH_RNPLAPP_FLIBS -lrnpl "
   BBH_RNPLAPP_CLIBS="$BBH_RNPLAPP_CLIBS -lrnpl "
   BBH_RNPLBLD_CLIBS="$BBH_RNPLBLD_CLIBS -lrnpl $LIBS"
   BBH_CHECK_INC_PATHS="$INCLUDE_PATHS $HOME/include $HOME/hdf/include"
   BBH_CHECK_INC_PATHS="$BBH_CHECK_INC_PATHS /usr/local/include /local/include"
   BBH_CHECK_INC_PATHS="$BBH_CHECK_INC_PATHS /usr/local/hdf/include"
   BBH_CHECK_INC_PATHS="$BBH_CHECK_INC_PATHS /usr/local/lib/hdf/include"
   BBH_CHECK_INC_PATHS="$BBH_CHECK_INC_PATHS $prefix/include"
   BBH_CHECK_INC_PATHS="$BBH_CHECK_INC_PATHS $prefix/hdf/include"
   BBH_CHECK_LIB_PATHS="$LIB_PATHS $HOME/lib $HOME/hdf/lib"
   BBH_CHECK_LIB_PATHS="$BBH_CHECK_LIB_PATHS /usr/local/lib /local/lib" 
   BBH_CHECK_LIB_PATHS="$BBH_CHECK_LIB_PATHS /usr/local/hdf/lib"
   BBH_CHECK_LIB_PATHS="$BBH_CHECK_LIB_PATHS /usr/local/lib/hdf/lib"
   BBH_CHECK_LIB_PATHS="$BBH_CHECK_LIB_PATHS $prefix/lib $prefix/hdf/lib"
   AC_MSG_CHECKING(for netcdf headers)
   BBH_FINDFILE(BBH_CHECK_INC_PATHS,"mfhdf.h",BBH_INC_PATH)
   if test $BBH_INC_PATH = "no"; then
      AC_MSG_RESULT(not found)
      AC_MSG_WARN(Can't find netcdf headers ... see message below)
      BBH_CHECK_FATAL="yes"
   else 
      AC_MSG_RESULT($BBH_INC_PATH/mfhdf.h)
      BBH_PUSHUNIQUE(BBH_RNPLAPP_INCPATHS,$BBH_INC_PATH)
   fi
   AC_MSG_CHECKING(for hdf headers) 
   BBH_FINDFILE(BBH_CHECK_INC_PATHS,"df.h",BBH_INC_PATH)
   if test $BBH_INC_PATH = "no"; then
      AC_MSG_RESULT(not found)
      AC_MSG_WARN(Can't find hdf headers ... see message below)
      BBH_CHECK_FATAL="yes"
   else 
      BBH_PUSHUNIQUE(BBH_RNPLAPP_INCPATHS,$BBH_INC_PATH)
      AC_MSG_RESULT($BBH_INC_PATH/df.h)
   fi
   AC_MSG_CHECKING(for netcdf library)
   BBH_FINDLIB(BBH_CHECK_LIB_PATHS,"netcdf",BBH_LIB_PATH)
   if test $BBH_LIB_PATH = "no"; then
      AC_MSG_RESULT(not found)
      AC_MSG_WARN(Can't find netcdf library ... see message below)
      BBH_CHECK_FATAL="yes"
   else 
      AC_MSG_RESULT($BBH_LIB_PATH/libnetcdf.a)
      BBH_PUSHUNIQUE(BBH_RNPLAPP_LIBPATHS,$BBH_LIB_PATH)
      BBH_RNPLAPP_FLIBS="$BBH_RNPLAPP_FLIBS -lnetcdf"
      BBH_RNPLAPP_CLIBS="$BBH_RNPLAPP_CLIBS -lnetcdf"
      BBH_RNPLBLD_CLIBS="$BBH_RNPLBLD_CLIBS -lnetcdf"
      BBH_DEFS="$BBH_DEFS -DHAVE_LIBNETCDF=1"
   fi
   BBH_FINDLIB(BBH_CHECK_LIB_PATHS,"df",BBH_LIB_PATH)
   AC_MSG_CHECKING(for hdf library)
   if test $BBH_LIB_PATH = "no"; then
      AC_MSG_RESULT(not found)
      AC_MSG_WARN(Can't find hdf library ... see message below)
      BBH_CHECK_FATAL="yes"
   else 
      AC_MSG_RESULT($BBH_LIB_PATH/libdf.a)
      BBH_PUSHUNIQUE(BBH_RNPLAPP_LIBPATHS,$BBH_LIB_PATH)
      BBH_RNPLAPP_FLIBS="$BBH_RNPLAPP_FLIBS -ldf"
      BBH_RNPLAPP_CLIBS="$BBH_RNPLAPP_CLIBS -ldf"
      BBH_RNPLBLD_CLIBS="$BBH_RNPLBLD_CLIBS -ldf"
      BBH_DEFS="$BBH_DEFS -DHAVE_LIBDF=1"
   fi
   AC_MSG_CHECKING(for vs library)
   BBH_FINDLIB(BBH_CHECK_LIB_PATHS,"vs",BBH_LIB_PATH)
   if test $BBH_LIB_PATH != "no"; then
      AC_MSG_RESULT($BBH_LIB_PATH/libvs.a)
      BBH_PUSHUNIQUE(BBH_RNPLAPP_LIBPATHS,$BBH_LIB_PATH)
      BBH_RNPLAPP_FLIBS="$BBH_RNPLAPP_FLIBS -lvs"
      BBH_RNPLAPP_CLIBS="$BBH_RNPLAPP_CLIBS -lvs"
      BBH_RNPLBLD_CLIBS="$BBH_RNPLBLD_CLIBS -lvs"
      BBH_DEFS="$BBH_DEFS -DHAVE_LIBVS=1"
   else
      AC_MSG_RESULT(not found)
   fi
   BBH_RNPLAPP_CLIBS="$BBH_RNPLAPP_CLIBS -lm"
   BBH_RNPLBLD_CLIBS="$BBH_RNPLBLD_CLIBS -lm"
   if test $BBH_CHECK_FATAL = "yes"; then
      BBH_MSG_NEED_HDF
   fi
])

AC_DEFUN(BBH_MSG_NEED_HDF,[
cat<<.

Installation of RNPL requires HDF and NetCDF software which could 
not be found in any of the usual locations.  If this software 
exists on your system, set the following environment variables

INCLUDE_PATHS
LIB_PATHS

to appropriate values and re-configure.

.
AC_MSG_ERROR(Exiting)
])

dnl ----------------------------------------------------------------------
dnl BBH_RNPL_SET_VARS defines various environment variables used 
dnl in RNPL application Makefiles
dnl ----------------------------------------------------------------------

AC_DEFUN(BBH_RNPL_SET_VARS,[
   for el in $BBH_RNPLAPP_LIBPATHS; do 
      BBH_PUSHUNIQUE(BBH_RNPLAPP_LPATHS,"-L$el")
   done
   for el in $BBH_RNPLAPP_INCPATHS; do 
      BBH_PUSHUNIQUE(BBH_RNPLAPP_CINC,"-I$el")
      BBH_PUSHUNIQUE(BBH_RNPLAPP_FINC,"${F77INCSTEM}${el}")
   done
   AC_SUBST(BBH_RNPLAPP_CINC)
   AC_SUBST(BBH_RNPLAPP_FINC)
   AC_SUBST(BBH_RNPLAPP_DEFS)
   AC_SUBST(BBH_RNPLAPP_LPATHS)
   AC_SUBST(BBH_RNPLAPP_FLIBS)
   AC_SUBST(BBH_RNPLAPP_CLIBS)
   AC_SUBST(BBH_RNPLBLD_CLIBS)
   AC_SUBST(BBH_RNPL_FLAGS)
   RNPL=$prefix/bin/rnpl
   AC_SUBST(RNPL)
])

dnl ----------------------------------------------------------------------
dnl BBH_F77_CONFIGURE invokes various macros to set up f77 environment
dnl for rnpl
dnl ----------------------------------------------------------------------

AC_DEFUN(BBH_F77_CONFIGURE,[
   BBH_PROG_F77
	if test $F77 != "None"; then
		BBH_PROG_F77FLAGS
		BBH_PROG_F77INCSTEM
		BBH_F77_HAS_SYSTEM
	else
		AC_MSG_WARN(No Fortran compiler found on this system)
	fi
])

dnl ----------------------------------------------------------------------
dnl BBH_PROG_F77 attempts to determine what the Fortran 77 compiler 
dnl is called 
dnl
dnl Sets Output Variable: F77
dnl ----------------------------------------------------------------------

AC_DEFUN(BBH_PROG_F77,[
   BBH_SYS_GETSYSTEM
   AC_MSG_CHECKING(for Fortran 77 compiler)
   AC_CACHE_VAL(ac_cv_bbh_prog_f77,[
      case $BBH_SYSTEM in
      IRIX)       F77='f77';                                              ;;
      SUNOS)      F77='f77';                                              ;; 
      AIX)        F77='xlf';                                              ;;
      A/UX)       F77='f77';                                              ;;
      OSF1)       F77='f90';                                              ;;
      UNICOS)     F77='cf77';                                             ;;
      *)          F77='f77';                                             ;;
      esac
		BBH_CHECK_PROG_NOMESS(ac_cv_bbh_prog_f77,$F77,$F77,"None")
      if test $ac_cv_bbh_prog_f77 != "None"; then
      cat <<END > _foo.f
            write(*,*) 'hello world!'
            stop
            end
END
         if $F77 _foo.f -o _foo > /dev/null 2>&1; then
            ac_cv_bbh_prog_f77=$F77
            rm _foo.f > /dev/null 2>&1 
            rm _foo > /dev/null 2>&1 
         else
            ac_cv_bbh_prog_f77="None"
            rm _foo.f > /dev/null 2>&1
            rm _foo > /dev/null 2>&1
         fi
      fi
   ])
   AC_MSG_RESULT($ac_cv_bbh_prog_f77)
   F77=$ac_cv_bbh_prog_f77
   AC_SUBST(F77)
])

dnl ----------------------------------------------------------------------
dnl BBH_CHECK_PROG_NOMESS: Version of AC_CHECK_PROG which disables 
dnl messages ...
dnl ----------------------------------------------------------------------
AC_DEFUN(BBH_CHECK_PROG_NOMESS,
[# Extract the first word of "$2", so it can be a program name with args.
set dummy $2; ac_word=[$]2
AC_CACHE_VAL(ac_cv_prog_$1,
[if test -n "[$]$1"; then
  ac_cv_prog_$1="[$]$1" # Let the user override the test.
else
  IFS="${IFS=   }"; ac_save_ifs="$IFS"; IFS="${IFS}:"
  for ac_dir in $PATH; do
    test -z "$ac_dir" && ac_dir=.
    if test -f $ac_dir/$ac_word; then
      ac_cv_prog_$1="$3"
      break
    fi
  done
  IFS="$ac_save_ifs"
dnl If no 4th arg is given, leave the cache variable unset,
dnl so AC_CHECK_PROGS will keep looking.
ifelse([$4], , , [  test -z "[$]ac_cv_prog_$1" && ac_cv_prog_$1="$4"
])dnl
fi])dnl
$1="$ac_cv_prog_$1"
AC_SUBST($1)dnl
])

dnl ----------------------------------------------------------------------
dnl BBH_PROG_F77FLAGS attempts to determine what  Fortran 77 flags 
dnl should be used
dnl
dnl Sets Output Variable: F77FLAGS
dnl ----------------------------------------------------------------------

AC_DEFUN(BBH_PROG_F77FLAGS,[
   BBH_SYS_GETSYSTEM
   AC_MSG_CHECKING(for Fortran 77 flags)
   AC_CACHE_VAL(ac_cv_bbh_prog_f77flags,[
      case $BBH_SYSTEM in
      IRIX)       ac_cv_bbh_prog_f77flags='-O';                           ;;
      SUNOS)      ac_cv_bbh_prog_f77flags='-O';                           ;; 
      AIX)        ac_cv_bbh_prog_f77flags='-qextname';                    ;;
      A/UX)       ac_cv_bbh_prog_f77flags='-O';                           ;;
      OSF1)       ac_cv_bbh_prog_f77flags='-O';                           ;;
      UNICOS)     ac_cv_bbh_prog_f77flags=' ';                            ;;
      *)          ac_cv_bbh_prog_f77flags=' ';                            ;;
      esac
   ])
   AC_MSG_RESULT($ac_cv_bbh_prog_f77flags)
   F77FLAGS=$ac_cv_bbh_prog_f77flags
   AC_SUBST(F77FLAGS)
])

dnl ----------------------------------------------------------------------
dnl BBH_PROG_F77INCSTEM sets stem for 'include' files 
dnl
dnl Sets Output Variable: F77INCSTEM
dnl ----------------------------------------------------------------------

AC_DEFUN(BBH_PROG_F77INCSTEM,[
   BBH_SYS_GETSYSTEM
   AC_MSG_CHECKING(for Fortran 77 include stem)
   AC_CACHE_VAL(ac_cv_bbh_prog_f77incstem,[
      case $BBH_SYSTEM in
      IRIX)       ac_cv_bbh_prog_f77incstem='-Wf,-I'                     ;;
      SUNOS)      ac_cv_bbh_prog_f77incstem='-I';                        ;; 
      AIX)        ac_cv_bbh_prog_f77incstem='-I';                        ;;
      A/UX)       ac_cv_bbh_prog_f77incstem='-I';                        ;;
      OSF1)       ac_cv_bbh_prog_f77incstem='-I';                        ;;
      UNICOS)     ac_cv_bbh_prog_f77incstem='-I';                        ;;
      *)          ac_cv_bbh_prog_f77incstem=' ';                         ;;
      esac
   ])
   AC_MSG_RESULT($ac_cv_bbh_prog_f77incstem)
   F77INCSTEM=$ac_cv_bbh_prog_f77incstem
   AC_SUBST(F77INCSTEM)
])

dnl ----------------------------------------------------------------------
dnl BBH_LIB_HDF attempts to locate the HDF and NETCDF include files 
dnl and libraries 
dnl Sets Output Variables: BBH_HDF_INCLUDEDIR, BBH_HDF_LIBDIR
dnl ----------------------------------------------------------------------

AC_DEFUN(BBH_LIB_HDF,[
])

dnl ----------------------------------------------------------------------
dnl BBH_F77_HAS_SYSTEM checks to see whether Fortran
dnl accepts 'call system()'
dnl Sets Output Variables: F77_HAS_SYSTEM (iff 'call system()' works)
dnl ----------------------------------------------------------------------

AC_DEFUN(BBH_F77_HAS_SYSTEM,[
   AC_MSG_CHECKING(whether Fortran 77 supports 'system()')
   AC_CACHE_VAL(ac_cv_bbh_f77_has_system,[
      if test $F77 != None; then
         cat <<END > _foo.f
            call system('touch _foo.f')
            stop
            end
END
         if $F77 $F77FLAGS _foo.f -o _foo > _output 2>&1; then
dnl         This hack necessary since UNICOS only generates warning for 
dnl         unsatisfied external references
            if grep -i Unsatisfied _output > /dev/null; then
               ac_cv_bbh_f77_has_system="no"
            else
               ac_cv_bbh_f77_has_system="yes"
            fi
            rm _output > /dev/null 2>&1
         else
            ac_cv_bbh_f77_has_system="no"
         fi
      fi
      rm _foo.f > /dev/null 2>&1 
      rm _foo > /dev/null 2>&1 
   ])
   AC_MSG_RESULT($ac_cv_bbh_f77_has_system)
   if test $ac_cv_bbh_f77_has_system = "yes"; then
      BBH_DEFS="$BBH_DEFS -DF77_HAS_SYSTEM"
   fi
])

dnl ----------------------------------------------------------------------
dnl BBH_RNPL_DEFAULT_PATH sets CPP variable BBH_RNPL_DEFAULT_PATH
dnl ----------------------------------------------------------------------

AC_DEFUN(BBH_RNPL_DEFAULT_PATH,[
   BBH_DEFS="$BBH_DEFS -DBBH_RNPL_DEFAULT_PATH='\"$prefix/lib/rnpl\"'"
])

dnl ----------------------------------------------------------------------
dnl Locate lib, include paths for auxiliary software needed by 
dnl Klasky's interpolator
dnl ----------------------------------------------------------------------

AC_DEFUN(BBH_CHECK_INTERP_LIBS,[
   BBH_CHECK_FATAL="no"
	BBH_INTERP_LIBS="-lbbhinterp"

	case $BBH_SYSTEM in
	UNICOS) 
		AC_MSG_WARN(Unicos system: Assuming 'linpack', 'blas' in 'libsci.a');
	;;
	*)
		BBH_CHECK_LIB_PATHS="$LIB_PATHS $HOME/lib $HOME/linpack/lib"
		BBH_CHECK_LIB_PATHS="$BBH_CHECK_LIB_PATHS /usr/lib" 
		BBH_CHECK_LIB_PATHS="$BBH_CHECK_LIB_PATHS /usr/local/lib /local/lib" 
		BBH_CHECK_LIB_PATHS="$BBH_CHECK_LIB_PATHS $prefix/lib $prefix/linpack/lib"

		AC_MSG_CHECKING(for linpack library)
		BBH_FINDLIB(BBH_CHECK_LIB_PATHS,"linpack",BBH_LIB_PATH)
		if test $BBH_LIB_PATH = "no"; then
			AC_MSG_RESULT(not found)
			AC_MSG_WARN(Can't find linpack library ... see message below)
			BBH_CHECK_FATAL="yes"
		else 
			AC_MSG_RESULT($BBH_LIB_PATH/liblinpack.a)
			BBH_PUSHUNIQUE(BBH_INTERP_LIBPATHS,$BBH_LIB_PATH)
			BBH_INTERP_LIBS="$BBH_INTERP_LIBS -llinpack"
		fi

		AC_MSG_CHECKING(for blas library)
		BBH_FINDLIB(BBH_CHECK_LIB_PATHS,"blas",BBH_LIB_PATH)
		if test $BBH_LIB_PATH = "no"; then
			AC_MSG_RESULT(not found)
			AC_MSG_WARN(Can't find blas library ... see message below)
			BBH_CHECK_FATAL="yes"
		else 
			AC_MSG_RESULT($BBH_LIB_PATH/libblas.a)
			BBH_PUSHUNIQUE(BBH_INTERP_LIBPATHS,$BBH_LIB_PATH)
			BBH_INTERP_LIBS="$BBH_INTERP_LIBS -lblas"
		fi
	;;
	esac

   if test $BBH_CHECK_FATAL = "yes"; then
      BBH_MSG_NEED_LINPACK_BLAS
   fi
])

dnl ----------------------------------------------------------------------
dnl BBH_INTERP_SET_VARS defines various environment variables used
dnl in RNPL application Makefiles
dnl ----------------------------------------------------------------------

AC_DEFUN(BBH_INTERP_SET_VARS,[
   for el in $BBH_INTERP_LIBPATHS; do
      BBH_PUSHUNIQUE(BBH_INTERP_LPATHS,"-L$el")
   done
	case $BBH_SYSTEM in
	AIX)        F77FLAGS='-O';                           ;;
	esac
   AC_SUBST(F77FLAGS)
   AC_SUBST(BBH_INTERP_LPATHS)
	AC_SUBST(BBH_INTERP_LIBS)
])

AC_DEFUN(BBH_MSG_NEED_LINPACK_BLAS,[
cat<<.

Installation of 'bbhinterp' requires LINPACK (liblinpack.a)
and BLAS (libblas.a).  These could not be found in any of the 
usual locations.  If you know that this software exists on your 
system, set the following environment variable

LIB_PATHS

to an appropriate value and re-configure.

.
AC_MSG_ERROR(Exiting)
])

dnl ----------------------------------------------------------------------
dnl Some utility macros
dnl ----------------------------------------------------------------------

AC_DEFUN(BBH_PUSHUNIQUE,[
   _pu_push="yes"
   for _pu_el in ${$1}; do
      if test $_pu_el = $2; then
         _pu_push="no"
      fi
   done
   if test $_pu_push = "yes"; then
      $1="${$1} "" $2"
   fi
])

AC_DEFUN(BBH_FINDLIB,[
   $3="no"
   for _stem in ${$1}; do
      if test -f $_stem/lib$2.a; then
         $3=$_stem
         break
      fi
   done
])

AC_DEFUN(BBH_FINDFILE,[
   $3="no"
   for _stem in ${$1}; do
      if test -f $_stem/$2; then
         $3=$_stem
         break
      fi
   done
])
