/****************************************************************************/
/*  Program reads a byte data cube for marching cubes
/****************************************************************************/

#include <stdio.h>
#include "math.h"
#include "march.h"
/* #include "nan.h" */


#define THRESHOLD 200.0		/* default threshold */
#define ZCHUNK 84

#define PI         3.1415926535
#define TWOPI      (PI * 2.0)
#define PI8        (PI / 8.0)
#define PI16       (PI / 16.0)
#define PI32       (PI / 32.0)
#define PI64       (PI / 64.0)
#define PI128      (PI / 128.0)
#define PI256      (PI / 256.0)
#define PI512      (PI / 512.0)

#define dot_prod(p,q) ((p)[0]*(q)[0]+(p)[1]*(q)[1]+(p)[2]*(q)[2])

char ch = '\'';

float ***grid;
unsigned  xsize, ysize, zsize;

main()
{

   int i,j,k,l,n,o;
   int m = 0;
   Big p_edges, temp_count, march();
   FILE *fp,*fopen();
   float  number;
   char s;
   char filename[64];
   unsigned char px[256][256];
   float pr;
   unsigned char header[256];
   int threshold = THRESHOLD;
   int lastchunk,iz;

/*  marching cubes variables  */
   long num_connections, num_corners;
   Big *connections;
   double *corner_coords[3], *corner_norms[3];
   int return_val;


   /** input data x y and z dimensions  **/
   printf(" Input X dimension : ");
   scanf ("%d", &xsize);
   printf("Input Y dimension: ");
   scanf ("%d", &ysize);
   printf("Input Z dimension: ");
   scanf("%d", &zsize);
   printf("Input threshold: ");
   scanf("%d", &threshold);
    
/* allocate the grid for marching()  dynamically   **/
 
   if((grid = (float ***) Calloc(xsize, sizeof(float **))) == NULL) 
     perror("grid"); 
   if((grid[0] = (float **) Calloc(xsize*ysize, sizeof(float *))) == NULL) 
     perror("grid[]"); 
   if((grid[0][0] = (float *) Calloc(xsize*ysize*ZCHUNK, sizeof(float))) == NULL) 
     perror("grid[][]"); 

   for(i = 0; i < xsize; i++)
     {
       grid[i] = *grid + i*ysize;
       for(j = 0; j < ysize; j++) 
	 grid[i][j] = **grid + ((ysize*i)+j)*ZCHUNK;  
     }
   /**************************************************************************/
   /* initialize first slice to zero inorder to get rid of edges extending   */
   /* to extents of the working space                                        */
   
   for(i=0;i<xsize;i++)
     for(j=0;j<ysize;j++)
       grid[i][j][0] = 0.0;
   
/***************************************************************************/


   printf("Input filename:  ");
   gets(filename);
   gets(filename);
   if((fp = fopen( filename, "rb")) == NULL) {
	printf("\nfile NOT found\n");
	exit(0);
	}



  for(m=4;m<zsize + 1;m=m+(ZCHUNK-1))
   {
   for(iz=0;iz<(ZCHUNK-1);iz++)
   {
    
   for(i=0;i<ysize;i++) {
     for(j=0;j<xsize;j++) {
         fscanf(fp,"%f",&pr);
        grid[j][i][iz+1] = pr;
     }
   }
}

/***************************************************************/

   temp_count = (Big) march( (unsigned long)xsize,
			        (unsigned long)ysize, (unsigned long) ZCHUNK,
                 (Input)threshold, (Input ***)grid, 
                 corner_coords, corner_norms,
                 &connections, &num_corners, &num_connections, &p_edges);

   /* copy the second row to the first row */

      printf("offset = %d\n", m-1);
      
      save(num_connections,num_corners,corner_coords, connections,temp_count,corner_norms,m-1);
      Cfree(corner_coords, p_edges, sizeof(corner_coords[3] ));
      Cfree(corner_norms, p_edges , sizeof(corner_norms[3] ));
      Cfree(connections, (3*temp_count), sizeof(Big));
      for(l=0;l<ysize;l++)
	for(n=0;n<xsize;n++)
	  grid[n][l][0] = grid[n][l][ZCHUNK-1];
      
      printf("finished row %d\n", m);
 }

/*
   lastchunk = (zsize%(ZCHUNK-1));
   for(iz=0;iz<lastchunk;iz++)
   {
   fread(px, sizeof(char), xsize*ysize, fp);  
    
   for(i=0;i<ysize;i++)
     for(j=0;j<xsize;j++)
       grid[j][i][iz+1] = (float)px[i][j];
   }

   for(i=0;i<xsize;i++)
    for(j=0;j<ysize;j++)
      grid[i][j][lastchunk] = 0.0;

   temp_count = (Big) march( (unsigned long)xsize,
			        (unsigned long)ysize, (unsigned long) (lastchunk+1), (Input)threshold, (Input ***)grid, 
   corner_coords, 
   corner_norms,
   &connections, &num_corners, &num_connections, &p_edges);


      printf("offset = %d\n", m-1);
      save(num_connections,num_corners,corner_coords, connections,temp_count,corner_norms,m-1);

*/
 }/*end  of main   */

/********************/
/********************/


/*=================================================================*/
/* subroutines begin here                                          */
/*=================================================================*/

save(num_connections,num_corners,corner_coords, connections, temp_count,corner_norms,offset)
     
     Big connections[];
     Output *corner_coords[3],*corner_norms[3];
     Big temp_count;
     int offset;
   long num_connections, num_corners;

{
  int n = 0;
  long j = 3;
  float normvec1[3];
  float normvec2[3];
  float normvec3[3];
  FILE *fopen();
  FILE *vrml_fp;
  int tmp1,tmp2,tmp3;
  Output temp1, temp2, temp3, temp4, temp5, temp6;
  short nanflag = 0;
  int temp7;
  float v0[3], v1[3], v2[3], n0[3], n1[3], n2[3];

/*********************************************************/
 /* the vrml output */
/********************************************************/
 
  if ( (vrml_fp = fopen ( "sphere.wrl", "w" )) == NULL ) {
      fprintf ( stderr, "Can't create vrml file.\n");
      exit (1);
    }

  fprintf ( vrml_fp, "#VRML V1.0 ascii\n\n" );

  fprintf ( vrml_fp, "Separator {\n\n" );

/********************************************************/ 
/* write coordinate points to vrml_fp */

  fprintf ( vrml_fp, "\t\tCoordinate3 {\n" );
  fprintf ( vrml_fp, "\t\t\tpoint\t[\n" );
  printf("The total number of corners = %d \n",num_corners);
  
  for(n=1; n<=num_corners; n++)
     { temp1 = corner_coords[0][n];
       temp2 = corner_coords[1][n];
       temp3 = (corner_coords[2][n] + offset);
       fprintf(vrml_fp, "\t\t\t%.3f %.3f %.3f,\n",temp1,temp2,temp3);
     } 

  fprintf ( vrml_fp, "\t\t\t]\n" );
  fprintf ( vrml_fp, "\t\t}\n\n\n" );

/******************************************************/
/* write normals to vrml_fp */
   
  fprintf ( vrml_fp, "\tNormal {\n" );
  fprintf ( vrml_fp, "\t\tvector [\n" );

  for(n=1; n<=num_corners; n++) 
     { temp4 = -corner_norms[0][n];
       temp5 = -corner_norms[1][n];
       temp6 = -corner_norms[2][n]; 
       fprintf(vrml_fp, "\t\t\t%.3f %.3f %.3f,\n",temp4,temp5,temp6);
     }
  
  fprintf ( vrml_fp, "\t\t\t]\n" );
  fprintf ( vrml_fp, "\t}\n\n\n" );
  fprintf ( vrml_fp, "\tNormalBinding { value PER_VERTEX_INDEXED }\n\n\n" );

/*******************************************************/      
/* write face connectivity */

  fprintf ( vrml_fp, "\tIndexedFaceSet {\n" );
  fprintf ( vrml_fp, "\t\tcoordIndex [\n" );

  for ( n = 0; n<num_connections; n++ ) {
       tmp1 = connections[n]-1;
       tmp2 = connections[n+1]-1;
       tmp3 = connections[n+2]-1;
       n+=2;
       fprintf ( vrml_fp, "\t\t\t%d, %d, %d, -1,\n", tmp1, tmp2, tmp3 );
     }     
  fprintf ( vrml_fp, "\t\t\t]\n" );

/**********************************************************/
/* write material indexing */

 /* fprintf ( vrml_fp, "\t\t\tmaterialIndex [\n" );
       
  for (n=0;n<num_corners-2; n++){
       tmp1=n;
       tmp2=n+1;
       tmp3=n+2;
       fprintf(vrml_fp, "\t\t\t%d, %d, %d, -1\n", tmp1, tmp2, tmp3);
     }
  tmp1 = num_corners-2;
  tmp2 = num_corners-1;
  tmp3 = num_corners;
  fprintf(vrml_fp, "\t\t\t %d, %d, %d, -1\n", tmp1, tmp2, tmp3);
  fprintf ( vrml_fp, "\t\t\t\t]\n" );
  fprintf ( vrml_fp, "\t}\n\n\n" );*/
  fprintf ( vrml_fp, "}\n" ); 
  fprintf ( vrml_fp, "}\n" );
  fclose(vrml_fp);

/************************************************************/
/* end of vrml output */
/************************************************************/
  
} /* save */


/* ********************************************************************** */
norm0 (v1, v2, v3, n0, n1, n2)

double	v1[], v2[], v3[], n0[], n1[], n2[];

{
	double	v4[3], v5[3], v6[3], length;

	*v4 = *v2 - *v1;
	*(v4+1) = *(v2+1) - *(v1+1);
	*(v4+2) = *(v2+2) - *(v1+2);

	*v5 = *v3 - *v1;
	*(v5+1) = *(v3+1) - *(v1+1);
	*(v5+2) = *(v3+2) - *(v1+2);

	*v6     = *(v4+1) * *(v5+2) - *(v5+1) * *(v4+2);
	*(v6+1) = *v4 * *(v5+2) - *v5 * *(v4+2);
	*(v6+2) = *v4 * *(v5+1) - *v5 * *(v4+1);

	if (dot_prod (v6, n0) < 0.0)
	{
		*n0 = -*n0;
		*(n0+1) = -*(n0+1);
		*(n0+2) = -*(n0+2);
		*n1 = -*n1;
		*(n1+1) = -*(n1+1);
		*(n1+2) = -*(n1+2);
		*n2 = -*n2;
		*(n2+1) = -*(n2+1);
		*(n2+2) = -*(n2+2);
	}
	length = sqrt(*v6 * *v6 + *(v6+1) * *(v6+1) + *(v6+2) * *(v6+2));

	return (length < 1e-30);
}

/* ********************************************************************** */

