/*****************************************************************************/
/*           P.I.V.S  Platform Independent Visualization System              */
/*           ==================================================              */
/*              P.I.V.S is copyright (C) 1996, Jason McIlvain                */
/*                                                                           */
/*   This code was written at NPAC Northeast Parallel Architecture Center as */
/* an undergraduate research project funded by NSF, NPAC, Syracuse Univ., and*/
/* General Electric. The advising of this project was done by Scott Klasky,  */
/* research scientist at NPAC and the coding was done by Jason McIlvain,     */
/* an undergraduate in the NSF REU program.                                  */
/*   The marching cubes algorithm used in this code was redesigned from the  */
/* SCRY Visualization System, (C) 1988, Regents of the University of         */
/* California, Lawrence Berkeley Laboratory.                                 */
/*****************************************************************************/

import java.awt.*;
import java.applet.*;
import java.net.*;
import java.io.*;
import java.util.*;
import java.lang.*;


public class march extends java.applet.Applet{

     static char vertpats[]= {
                 (char)( 0 ),
                 (char)( (1<<0)  | (0) ),
                 (char)( (1<<0)  | (1<<1)  | (0) ),
                 (char)( (1<<0)  | (1<<3)  | (0) ),
                 (char)( (1<<0)  | (1<<7)  | (0) ),
                 (char)( (1<<1)  | (1<<5)  | (1<<4)  | (0) ),
                 (char)( (1<<0)  | (1<<1)  | (1<<7)  | (0) ),
                 (char)( (1<<2)  | (1<<1)  | (1<<7)  | (0) ),
                 (char)( (1<<0)  | (1<<1)  | (1<<5)  | (1<<4)  ),
                 (char)( (1<<0)  | (1<<4)  | (1<<5)  | (1<<6)  ),
                 (char)( (1<<0)  | (1<<2)  | (1<<5)  | (1<<7)  ),
                 (char)( (1<<0)  | (1<<4)  | (1<<5)  | (1<<7)  ),
                 (char)( (1<<2)  | (1<<4)  | (1<<5)  | (1<<1)  ),
                 (char)( (1<<0)  | (1<<3)  | (1<<5)  | (1<<6)  ),
                 (char)( (1<<1)  | (1<<5)  | (1<<4)  | (1<<6)  )  };


   static char edgepats[][]=
        {{ (char)0,                       (char)0,                        (char)0,                       (char)0 },
         { (char)((1<<0)|(1<<1)|(1<<2)),  (char)0,                        (char)0,                       (char)0 },
         { (char)((1<<1)|(1<<2)|(1<<5)),  (char)((1<<2)|(1<<10)|(1<<5)),  (char)0,                       (char)0 },
         { (char)((1<<0)|(1<<1)|(1<<2)),  (char)((1<<4)|(1<<5)|(1<<6)),   (char)0,                       (char)0 },
         { (char)((1<<0)|(1<<1)|(1<<2)),  (char)((1<<6)|(1<<9)|(1<<12)),  (char)0,                       (char)0 },
         { (char)((1<<0)|(1<<2)|(1<<13)), (char)((1<<0)|(1<<13)|(1<<9)),  (char)((1<<0)|(1<<9)|(1<<5)),  (char)0 },
         { (char)((1<<1)|(1<<2)|(1<<5)),  (char)((1<<2)|(1<<10)|(1<<5)),  (char)((1<<6)|(1<<9)|(1<<12)), (char)0 },
         { (char)((1<<1)|(1<<4)|(1<<14)), (char)((1<<0)|(1<<5)|(1<<10)),  (char)((1<<6)|(1<<9)|(1<<12)), (char)0 },
         { (char)((1<<1)|(1<<5)|(1<<9)),  (char)((1<<1)|(1<<13)|(1<<9)),  (char)0,                       (char)0 },
         { (char)((1<<1)|(1<<0)|(1<<14)), (char)((1<<0)|(1<<9)|(1<<10)),  (char)((1<<9)|(1<<12)|(1<<14)),(char)((1<<14)|(1<<0)|(1<<9))   },
         { (char)((1<<0)|(1<<2)|(1<<14)), (char)((1<<0)|(1<<14)|(1<<4)),  (char)((1<<8)|(1<<10)|(1<<6)), (char)((1<<8)|(1<<12)|(1<<6))   },
         { (char)((1<<0)|(1<<10)|(1<<6)), (char)((1<<0)|(1<<1)|(1<<13)),  (char)((1<<13)|(1<<12)|(1<<6)),(char)((1<<0)|(1<<13)|(1<<6))   },
         { (char)((1<<0)|(1<<2)|(1<<13)), (char)((1<<0)|(1<<13)|(1<<9)),  (char)((1<<0)|(1<<9)|(1<<5)),  (char)((1<<4)|(1<<1)|(1<<14))   },
         { (char)((1<<0)|(1<<1)|(1<<2)),  (char)((1<<4)|(1<<5)|(1<<6)),   (char)((1<<8)|(1<<9)|(1<<10)), (char)((1<<12)|(1<<13)|(1<<14)) },
         { (char)((1<<0)|(1<<2)|(1<<14)), (char)((1<<14)|(1<<12)|(1<<9)), (char)((1<<0)|(1<<14)|(1<<9)), (char)((1<<0)|(1<<9)|(1<<5))    }
        };

   static short cubedges[][] = new short[256][4];
   static char numtri[] = new char[256];
   static char edgecorns[] = new char[256];
   static char numedg[] = new char[256];
   static char edgecornsi[] = new char[256];
   static char edgecornsj[] = new char[256];
   static char edgecornsk[] = new char[256];
   static char edgecornst[] = new char[256];

   static char edgenums[][] ={{1,0,0},
                              {0,1,0},
                              {0,0,1},
                              {0,0,0},
                              {1,2,0},
                              {2,1,0},
                              {2,2,1},
                              {0,0,0},
                              {1,0,2},
                              {2,1,2},
                              {2,0,1},
                              {0,0,0},
                              {1,2,2},
                              {0,1,2},
                              {0,2,1}};


   static char revedges[][][] ={{{(char)255,(char)255,(char)255},{(char)255,(char)255,(char)255},{(char)255,(char)255,(char)255}},
                                {{(char)255,(char)255,(char)255},{(char)255,(char)255,(char)255},{(char)255,(char)255,(char)255}},
                                {{(char)255,(char)255,(char)255},{(char)255,(char)255,(char)255},{(char)255,(char)255,(char)255}}};

   static short pat[][] = {{0,1,2,1},
                           {0,0,2,2}};

   static short edge, axis, num;
   static int xsize,ysize,zsize,offset,smoothpass;
   static double threshold,value;
   static String filename = new String();
   static double[][][] grid;
   static long p_edges, num_corners, num_connections;
   static long[] connections;
   static long[] depths;
   static double[][] corner_coords;
   static double[][] corner_norms;
   static long triangles_done = 0;
   static cube[][] cubes; //or close to this
   //interface vars
   Button mbutton;
   Checkbox abox,bbox;
   TextField xtext;
   TextField ytext;
   TextField ztext;
   TextField ttext;
   TextField ftext;
   TextField stext;
   Label lbl0,lbl1,lbl2,lbl3,lbl4,lbl5,lbl6,lbl7,lbls;
   int ascii=0;
   int binary=0;

/*****************************************************************************************/
public static void init_marchcubes() throws java.io.IOException {
        short h,i,j,k,tmp,patt;
        short aftx, afty, aftz, finl=0,e0,e1,e2,e3;//from rotcube
        short xr, yr, zr, onum;//from rotcube

        for(i=0;i<15;i++)
           revedges[(int)edgenums[i][0]][(int)edgenums[i][1]][(int)edgenums[i][2]]=(char)i;
        revedges[0][0][0]= (char)255;
        for(i=0;i<256;i++)
           for(j=0;j<4;j++)
              cubedges[i][j]=0;
//rotcubes here
    i=0;j=0;
    xr=0; yr=0; zr=0; onum=0;
    for(i=0; i<8; i++)
       if(((((255)&(vertpats[onum])) & (1<<(i)))>0)&&(((((255)&(vertpats[onum]))& (1<<(i)))& (1<<7)) !=128)) {
         aftx = (short)(i ^ ((xr==2) ? 6: ((((i & (1<<1))>0?1:0) == ((i &(1<<2))>0?1:0)) ?
                    ((xr==3) ? 4: ((xr==1)?2:0)): ((xr==3)? 2: ((xr==1)? 4:0)))));
         afty = (short)(aftx ^ ((yr==2) ? 5: ((((aftx &(1<<0))>0?1:0)== ((aftx &(1<<2))>0?1:0)) ?
                       ((yr==3) ? 1: ((yr==1)? 4:0)): ((yr==3) ? 4: ((yr==1)? 1:0)))));
         aftz = (short)(afty ^ ((zr==2) ? 3: ((((afty &(1<<0))>0?1:0)== ((afty &(1<<01))>0?1:0)) ?
                       ((zr==3) ? 2: ((zr==1)? 1:0)): ((zr==3) ? 1: ((zr==1)? 2:0)))));
         finl |= (short)(1<<(aftz));
       }
    if(cubedges[finl][0]==0)
    //rotedges here
    for(i=0; i<4; i++){
       cubedges[finl][i] = 0;
       if((edgepats[onum][i])>0){
         for(j=0; j<15; j++){
            if(((short)(edgepats[onum][i] & (1<<j)))>0){
                e0 = (short)(edgenums[j][0] | (edgenums[j][1]<<2) | (edgenums[j][2]<<4));
                edge=e0; axis=0; num=xr;
                e1 = rotate();
                edge=e1; axis=1; num=yr;
                e2 = rotate();
                edge=e2; axis=2; num=zr;
                e3 = rotate();
                cubedges[finl][i] |= (1<<(revedges[e3 & 3][(e3 &(3<<2))>>2][(e3 &(3<<4))>>4]));
            }
         }
       } else break;//could have problem here was return when a seperate procedure
    }

    for(h=1; h<15; h++)
       for(i=0; i<4; i++)
          for(j=0; j<4; j++)
             for(k=0; k<4; k++){
                finl=0;
                xr=i; yr=j; zr=k; onum=h;
                for(int i1=0; i1<8; i1++)
                   if(((((255)&(vertpats[onum])) & (1<<(i1)))>0)&&(((((255)&(vertpats[onum]))& (1<<(i1)))& (1<<7)) !=128)) {
                      aftx = (short)(i1 ^ ((xr==2) ? 6: ((((i1 & (1<<1))>0?1:0) == ((i1 &(1<<2))>0?1:0)) ?
                                         ((xr==3) ? 4: ((xr==1)?2:0)): ((xr==3)? 2: ((xr==1)? 4:0)))));
                      afty = (short)(aftx ^ ((yr==2) ? 5: ((((aftx &(1<<0))>0?1:0)== ((aftx &(1<<2))>0?1:0)) ?
                                            ((yr==3) ? 1: ((yr==1)? 4:0)): ((yr==3) ? 4: ((yr==1)? 1:0)))));
                      aftz = (short)(afty ^ ((zr==2) ? 3: ((((afty &(1<<0))>0?1:0)== ((afty &(1<<01))>0?1:0)) ?
                                            ((zr==3) ? 2: ((zr==1)? 1:0)): ((zr==3) ? 1: ((zr==1)? 2:0)))));
                      finl |= (short)(1<<(aftz));
                   }
                   if(cubedges[finl][0]==0)
                      //rotedges here
                     for(int i2=0; i2<4; i2++){
                        cubedges[finl][i2] = 0;
                        if((edgepats[onum][i2])>0){
                          for(int j2=0; j2<15; j2++){
                             if(((short)(edgepats[onum][i2] & (1<<j2)))>0){
                               e0 = (short)(edgenums[j2][0] | (edgenums[j2][1]<<2) | (edgenums[j2][2]<<4));
                               edge=e0; axis=0; num=xr;
                               e1 = rotate();
                               edge=e1; axis=1; num=yr;
                               e2 = rotate();
                               edge=e2; axis=2; num=zr;
                               e3 = rotate();
                               cubedges[finl][i2] |= (1<<(revedges[e3 & 3][(e3 &(3<<2))>>2][(e3 &(3<<4))>>4]));
                             }
                          }
                        } else break;//could have problem here was return when a seperate procedure
                     }

             }
             for(i=0; i<255; i++)
                if((cubedges[i ^ 255][0])>0) ; else
                  for(j=0; j<4; j++)
                      cubedges[i ^ 255][j] = cubedges[i][j];

    for(i=0; i<256; i++){
       tmp = 0;
       for(j=0; j<4; j++)
          if(cubedges[i][j]>0) tmp++; else break;
       numtri[i]=(char)(tmp & 255);
       tmp=0;
       patt = (short)(cubedges[i][0] | cubedges[i][1] | cubedges[i][2] | cubedges[i][3]);
       for(j=0; j<3; j++) {
          if((patt & (1<<(j)))>0)
            tmp++;
       }
       edgecorns[i] = (char)(tmp & 255);
       tmp=0;
       for(j=0; j<15; j++){
          if((patt & (1<<(j)))>0)
            tmp++;
       }
       numedg[i] = (char)(tmp & 255);
       edgecornsi[i] = (char)(((((patt & (1<<5))>0?1:0) + ((patt & (1<<10))>0?1:0))&(255)) & 255);
       edgecornsj[i] = (char)(((((patt & (1<<4))>0?1:0) + ((patt & (1<<14))>0?1:0))&(255)) & 255);
       edgecornsk[i] = (char)(((((patt & (1<<8))>0?1:0) + ((patt & (1<<13))>0?1:0))&(255)) & 255);
       edgecornst[i] = (char)((edgecorns[i] + edgecornsi[i]) & 255);
    }

   }

/*****************************************************************************************/
 public static short rotate() throws java.io.IOException {

        short index=0, newedge = 0, parity, inc;
        boolean y;

        if(num==0) return edge;
        y = (axis == 1);
          if((edge & (1<<(axis*2)))>0){
            parity=1;
            if(((axis==2)?(edge & (3<<2))>>2 :(edge &(3<<4))>>4)==2)
              if(((axis==0)?(edge&(3<<2))>>2 : (edge&3))==2)
                index = (short)((y) ? 3 : 2);
              else index = (short)((y) ? 2 : 3);
            else
                if(((axis==0)?(edge&(3<<2))>>2:(edge&3))==2)
                  index = (short)((y)?0:1);
                else index = (short)((y)?1:0);
          }
          else {
            parity = 0;
            switch((axis==2)?(edge&(3<<2))>>2:(edge&(3<<4))>>4) {
                case 0:
                       index=0; break;
                case 1:
                       if(((axis==0)?(edge&(3<<2))>>2:(edge&3))>0)
                          index=(short)((y)?3:1);
                       else index = (short)((y)?1:3); break;
                case 2:
                       index = 2; break;
            }
        }
        if(axis==01) inc = -1; else inc = 1;

        newedge|=((axis==2)?(pat[parity][(index+num)&3])<<2:(pat[parity][(index+num)&3])<<4);
        newedge|=((axis>0)?(pat[parity][(index+num+inc)&3]):(pat[parity][(index+num+inc)&3])<<2);
        newedge|=(edge & (3 << (axis*2)));
        return newedge;
   }



/*************************************************************************************/
public void VRML_out() throws java.io.IOException
{
    int n=0;
    double temp,t1,t2,t3;
    long tmp1,tmp2,tmp3;
    PrintStream f = null;
   // OutputStream f = null;

   URL outputfile = new URL(getDocumentBase(),"out.wrl");
   URLConnection uc = outputfile.openConnection();   
   uc.setDoOutput(true);

System.out.println("before printstream");
//uc.connect();
   f = new PrintStream(uc.getOutputStream());
//   f= uc.getOutputStream();  

System.out.println("after printstream");

//int val= 123;
//f.write(val);


    f.print("#VRML V1.0 ascii\n\n");
    f.print("Separator {\n\n");
    f.print("\t\tCoordinate3 {\n");
    f.print("\t\t\tpoint\t[\n");

    for(n=1;n<=num_corners;n++)
       {
         temp = corner_coords[2][n] + offset;
         f.print("\t\t\t");
         f.print("" +corner_coords[0][n]);
         f.print(", ");
         f.print("" +corner_coords[1][n]);
         f.print(", ");
         f.print("" +temp);
         f.print(",\n");
       }

System.out.println(" in print statements");
    f.print("\t\t\t]\n");
    f.print("\t\t}\n\n\n");
    f.print("\tNormal {\n");
    f.print("\t\tvector [\n");

    for(n=1;n<= num_corners; n++)
       {
         t1= -corner_norms[0][n];
         t2= -corner_norms[1][n];
         t3= -corner_norms[2][n];
         f.print("\t\t\t");
         f.print("" +t1);
         f.print(", ");
         f.print("" +t2);
         f.print(", ");
         f.print("" +t3);
         f.print(",\n");
       }

    f.print("\t\t\t]\n");
    f.print("\t}\n\n\n");
    f.print("\tNormalBinding { value PER_VERTEX_INDEXED }\n\n\n");

    f.print("\tIndexedFaceSet {\n");
    f.print("\t\tcoordIndex [\n");

    for(n=0;n< num_connections; n+=3)
       { tmp1 = connections[n] -1;
         tmp2 = connections[n+1] -1;
         tmp3 = connections[n+2] -1;
         f.print("\t\t\t");
         f.print("" +tmp1);
         f.print(", ");
         f.print("" +tmp2);
         f.print(", ");
         f.print("" +tmp3);
         f.print(", -1,\n");
       }
    f.print("\t\t\t]\n");
    f.print("}\n");
    f.print("}\n");

f.close();
System.out.println(" at end of VRML");

}

/*******************************************************************************************/
public static void gouraud_tess()throws java.io.IOException {

    int i,j,l,m,n,d=0,od,nd,q,ia,ib,ic,dx,dy,dz;
    short tmp;
    int vec[] = new int[3];
    int lx,ly,lz,tri=0,ct;
    int sides_done=0, tri_done=0, num_edges=0;
    int t[] = new int[3];
    int fl[] = new int[3];
    int fg[] = new int[3];
    int cl[] = new int[3];
    int cg[] = new int[3];
    double s[] = new double[3];
    double b[] = new double[3];
    double u[] = new double[3];
    int f[] = new int[3];
    int c[] = new int[3];
    double mag;

    dx= xsize;
    dy= ysize;
    dz= zsize;
    t[0]= dx-1;
    t[1]= dy-1;
    t[2]= dz-1;

    corner_coords = new double[3][(int)p_edges];
    corner_norms = new double[3][(int)p_edges];

    long nextslice[][][] = new long[dy][dz][3];
    long curslice[][][] = new long[dy][dz][3];
    long tmpslice[][][] = new long[dy][dz][3];

    for(i=0;i<t[0];i++) {
        if(depths[i]>0)
           for(j=0;j<depths[i];j++) {
               tri = (int)numtri[cubes[i][j].cube];
               lx = i;
               ly = cubes[i][j].y;
               lz = cubes[i][j].z;
               for(l=0;l<tri;l++) {
                   tmp = (short)cubedges[cubes[i][j].cube][l];
                   for(ct=0,m=0;m<15;m++)
                       if(((short)(tmp & (1<<(m))))>0) {
                          vec[0]= lx;
                          vec[1]= ly;
                          vec[2]= lz;
                          for(n=0;n<3;n++){
                              if(edgenums[m][n] == 1) d=n;
                              else if (edgenums[m][n] == 2) vec[n]++;}
                           if(((vec[0] == i)&&(curslice[vec[1]][vec[2]][d]==0)) || ((vec[0] != i)&&(nextslice[vec[1]][vec[2]][d]==0)))
                           {
                            if(vec[0] == i) curslice[vec[1]][vec[2]][d] = ++num_edges;
                            else nextslice[vec[1]][vec[2]][d] = ++num_edges;
                            if(d==0){od = 1; nd = 2;}
                            else if(d==1){od = 0; nd = 2;}
                            else { od = 0; nd = 1;}

                            if(vec[d] < t[d])
                               s[d] = corner_coords[d][num_edges]=
                               (double)((vec[d]) - ((((value - grid[vec[0]][vec[1]][vec[2]])/(grid[vec[0]][vec[1]][vec[2]] - grid[((d==0)?vec[0]+1:vec[0])] [((d==1)?vec[1]+1:vec[1])]  [((d==2)?vec[2]+1:vec[2])] ))
                               < -1.0) ? -0.5  :    (((value - grid[vec[0]][vec[1]][vec[2]])/(grid[vec[0]][vec[1]][vec[2]] - grid[((d==0)?vec[0]+1:vec[0])] [((d==1)?vec[1]+1:vec[1])]  [((d==2)?vec[2]+1:vec[2])] ))
                               >  1.0) ?  0.5  :     ((value - grid[vec[0]][vec[1]][vec[2]])/(grid[vec[0]][vec[1]][vec[2]] - grid[((d==0)?vec[0]+1:vec[0])] [((d==1)?vec[1]+1:vec[1])]  [((d==2)?vec[2]+1:vec[2])] ))));

                           else s[d] = corner_coords[d][num_edges] = (double)(vec[d] + 0.5);

                           s[od] = corner_coords[od][num_edges] = (double)vec[od];
                           s[nd] = corner_coords[nd][num_edges] = (double)vec[nd];
                           //normfrag begins here
                           for(q=0;q<3;q++){
                              f[q]= (int)(Math.floor((double)s[q]));
                              c[q]= (int)(Math.ceil((double)s[q]));
                              if(f[q] != c[q]){
                                b[q]= (Math.floor((double)s[q]));
                                u[q]= (Math.ceil((double)s[q]));
                                if(f[q] <= 0) fl[q]= 0; else fl[q]= (f[q]-1);
                                if(c[q] <= 0) cl[q]= 0; else cl[q]= (c[q]-1);
                                if(f[q] >= t[q]) fg[q]= t[q]; else fg[q]= (f[q]+1);
                                if(c[q] >= t[q]) cg[q]= t[q]; else cg[q]= (c[q]+1);
                                if(f[q] < 0) f[q]= 0;
                                if(c[q] < 0) c[q]= 0;
                                if(c[q] >= t[q]) c[q]= f[q]= t[q];
                              }
                              else{
                                if(f[q] <= 0) fl[q]= 0; else fl[q]= (f[q]-1);
                                if(f[q] >= t[q]) fg[q]= t[q]; else fg[q]= (f[q]+1);
                              }
                           }
                           if(f[0] == c[0])
                             s[0]= corner_norms[0][num_edges]=
                                     grid[fg[0]][f[1]][f[2]] - grid[fl[0]][f[1]][f[2]];
                           else
                             s[0]= corner_norms[0][num_edges]=
                                     ((grid[fg[0]][f[1]][f[2]] - grid[fl[0]][f[1]][f[2]])*
                                      (s[0] - b[0]) + (u[0] - s[0])*
                                      (grid[cg[0]][c[1]][c[2]] - grid[cl[0]][c[1]][c[2]]));
                           if(f[1] == c[1])
                             s[1]= corner_norms[1][num_edges]=
                                     grid[f[0]][fg[1]][f[2]] - grid[f[0]][fl[1]][f[2]];
                           else
                             s[1]= corner_norms[1][num_edges]=
                                     ((grid[f[0]][fg[1]][f[2]] - grid[f[0]][fl[1]][f[2]])*
                                      (s[1] - b[1]) + (u[1] - s[1])*
                                      (grid[c[0]][cg[1]][c[2]] - grid[c[0]][cl[1]][c[2]]));
                           if(f[2] == c[2])
                             s[2]= corner_norms[2][num_edges]=
                                     grid[f[0]][f[1]][fg[2]] - grid[f[0]][f[1]][fl[2]];
                           else
                             s[2]= corner_norms[2][num_edges]=
                                     ((grid[f[0]][f[1]][fg[2]] - grid[f[0]][f[1]][fl[2]])*
                                      (s[2] - b[2]) + (u[2] - s[2])*
                                      (grid[c[0]][c[1]][cg[2]] - grid[c[0]][c[1]][cl[2]]));
                           mag= Math.sqrt(s[0]*s[0] + s[1]*s[1] + s[2]*s[2]);
                           corner_norms[0][num_edges] /= mag;
                           corner_norms[1][num_edges] /= mag;
                           corner_norms[2][num_edges] /= mag;
                           //end of normfrag

                         }
                         connections[sides_done++] = ((vec[0]==i)?curslice[vec[1]][vec[2]][d]:
                                                   nextslice[vec[1]][vec[2]][d]);
                         if(++ct>= 3) break;
                       }
                       tri_done++;
               }}
               for(ia=0;ia<dy;ia++){
                     for(ib=0;ib<dz;ib++){
                     for(ic=0;ic<3;ic++){
                        tmpslice[ia][ib][ic] = nextslice[ia][ib][ic];}}}

               for(ia=0;ia<dy;ia++){
                     for(ib=0;ib<dz;ib++){
                     for(ic=0;ic<3;ic++){
                        nextslice[ia][ib][ic]= curslice[ia][ib][ic];}}}

               for(ia=0;ia<dy;ia++){
                     for(ib=0;ib<dz;ib++){
                     for(ic=0;ic<3;ic++){
                        curslice[ia][ib][ic]= tmpslice[ia][ib][ic];}}}

               //memset set
                     for(ia=0;ia<dy;ia++){
                     for(ib=0;ib<dz;ib++){
                     for(ic=0;ic<3;ic++){
                        nextslice[ia][ib][ic]=0;}}}
           }
           triangles_done = tri_done;
           num_corners = num_edges;


            System.out.println(" end of gt: num edges =" +num_edges);//debug
            System.out.println("num triangles =" +tri_done);//debug
            System.out.println("end of g_tess fooobar");
           

    }


/******************************************************************************************/
public static long process_array() throws java.io.IOException{

    int i,j,k=0,ic,bb,bs,dx,dy,dz;
    int tmp;
    int tx,ty,tz,pd;
    int tmptr[];
    long num_tri = 0, edges_predicted = 0;
    int num_cubes = 0;

    dx= xsize;
    dy= ysize;
    dz= zsize;
    tx = dx-1;
    ty = dy-1;
    tz = dz-1;
    pd = (dz/32) + ((dz%32)>0 ? 1:0);

    int prevslice[] = new int[dy*pd];
    int curslice[] = new int[dy*pd];
    cube[][][] array = new cube[tx][ty][tz];
    for(i=0;i<tx;i++)
       for(j=0;j<(ty);j++)
          for(k=0;k<(tz);k++)
             array[i][j][k]= new cube();

    for(j=0;j<dy;j++)
       for(k=0;k<dz;k++)
          if(grid[0][j][k] >= value) curslice[((j)*pd)+((k)/32)] |= (1<<(k&31));

    tmptr = curslice;
    curslice = prevslice;
    prevslice = tmptr;
    for(ic=0;ic<(dy*pd);ic++)
       curslice[ic]= 0;

    for(i=0; i<tx; i++){
        num_cubes = 0;
        if(i>1) { array[i]= array[i & 1];
                  for(bb=0;bb<ty;bb++)
                  for(bs=0;bs<tz;bs++){
                     array[i][bb][bs].cube=0;
                     array[i][bb][bs].y=0;
                     array[i][bb][bs].z=0;}
        }
        for(j=0;j<dy;j++)
           if(grid[i+1][j][0] >= value) curslice[((j)*pd)+((0)/32)] |= (1<<(0&31));

        for(k=1;k<dz;k++)
           if(grid[i+1][0][k] >= value) curslice[((0)*pd)+((k)/32)] |= (1<<((k)&31));

        for(j=0;j<ty;j++){
           for(k=0;k<tz;k++){
           if(grid[i+1][j+1][k+1] >= value) curslice[((j+1)*pd)+((k+1)/32)] |= (1<<((k+1)&31));

           tmp  = 0;
           tmp |= ((prevslice[((j)  *pd)+((k)  /32)] & (1<<(k&31)))>0 ? 01: 0);
           tmp |= ((curslice [((j)  *pd)+((k)  /32)] & (1<<(k&31)))>0 ? 02: 0);
           tmp |= ((prevslice[((j+1)*pd)+((k)  /32)] & (1<<(k&31)))>0 ? 04: 0);
           tmp |= ((curslice [((j+1)*pd)+((k)  /32)] & (1<<(k&31)))>0 ? 010: 0);
           tmp |= ((prevslice[((j)  *pd)+((k+1)/32)] & (1<<((k+1)&31)))>0 ? 020: 0);
           tmp |= ((curslice [((j)  *pd)+((k+1)/32)] & (1<<((k+1)&31)))>0 ? 040: 0);
           tmp |= ((prevslice[((j+1)*pd)+((k+1)/32)] & (1<<((k+1)&31)))>0 ? 0100: 0);
           tmp |= ((curslice [((j+1)*pd)+((k+1)/32)] & (1<<((k+1)&31)))>0 ? 0200: 0);

           array[i][j][k].cube = tmp;
           if((tmp != 0) && (tmp != 255)){
              num_cubes++;
              if(i==(tx-2))
                 edges_predicted += edgecornst[tmp];
              else
                 edges_predicted += edgecorns[tmp];
           }
           }
        }
        depths[i] = num_cubes;

        if(num_cubes>0) {
            //cubes[i] = num_cubes
            num_cubes = 0;
            for(j=0;j<ty;j++)
               for(k=0;k<tz;k++) {
                  if(((tmp = array[i][j][k].cube) != 0)&&(tmp != 255)) {
                    num_tri += numtri[tmp];
                    cubes[i][num_cubes].cube = tmp;
                    cubes[i][num_cubes].y = j;
                    cubes[i][num_cubes].z = k;
                    num_cubes++;
                  }
               }
            for(j=0;j<ty-1;j++)
               if(i==(tx-2))
                 edges_predicted += numedg[array[i][j][tz-1].cube];
               else
                 edges_predicted += edgecornsk[array[i][j][tz-1].cube];

            for(k=0;k<tz-1;k++)
               if(i==(tx-2))
                 edges_predicted += numedg[array[i][ty-1][k].cube];
               else
                 edges_predicted += edgecornsj[array[i][ty-1][k].cube];

            edges_predicted += numedg[array[i][ty-1][tz-1].cube];

        }

        tmptr = curslice;
        curslice = prevslice;
        prevslice = tmptr;
        for(ic=0;ic<(dy*pd);ic++)
           curslice[ic]=0;
    }

System.out.println("cube vertices calculated, "+num_tri+" triangles, "+edges_predicted+" vertices");

    p_edges = edges_predicted;
    num_connections = 3*num_tri;


    return num_tri;
}



/***************************************************************************/
public void mrch() throws java.io.IOException{
  long temp_count;
  int i=0,j=0,k=0;
  int fl=0,pt=0;
  InputStream isa = null;
  DataInputStream disb = null;
  URL url_dataset;
 

    offset=4;
    grid = new double[xsize][ysize][zsize];
    value = threshold;
    depths = new long[xsize-1];
    cubes = new cube[xsize-1][(xsize*ysize)];
        for(i=0;i<(xsize-1);i++)
           for(j=0;j<(xsize*ysize);j++){
              cubes[i][j]= new cube();
              cubes[i][j].cube =0;
              cubes[i][j].y = 0;
              cubes[i][j].z = 0;
           }

//read in data set from server
    fl=xsize*ysize*zsize;
     byte s[]= new byte[fl];

    //ascii
    if((ascii == 1)&&(binary == 0)){
       url_dataset= new URL(getDocumentBase(),filename);
       isa = url_dataset.openStream();
       isa.read(s,0,fl);
       isa.close();
       pt = 0;
       for(k=0;k<zsize;k++)
        for(i=0;i<ysize;i++)
         for(j=0;j<xsize;j++){
            if(s[pt] == '1'){
                 grid[j][i][k] = 1.0;
                 pt++;
            }
            else {
                 grid[j][i][k] = 0.0;
                 pt++;
            }
         }
     }
    //binary
    else if((binary == 1)&&(ascii == 0)){
  
     /*url_dataset= new URL(getDocumentBase(),filename);
       disb = url_dataset.openStream();
       for(int index=0;index<fl;index++)
          for(k=0;k<zsize;k++)
             for(i=0;i<ysize;i++)
                for(j=0;j<xsize;j++)
          grid[j][i][k]= disb.readDouble();
       disb.close();*/
    }//else make them redo the params


//smoothing here
for (int l=0;l<smoothpass;l++) {

   for(i=1; i<xsize-1; i++) {
      for(j=1; j<ysize-1; j++) {
         for(k=1; k<zsize-1; k++) { grid[i][j][k]=0.5*grid[i][j][k]+0.083333333333333333*
            (grid[i-1][j][k]+grid[i+1][j][k]+grid[i][j+1][k]+
             grid[i][j-1][k]+grid[i][j][k+1]+grid[i][j][k-1]);
			}
		}
	}

   for(i=xsize-2; i>=1;i--) {
      for(j=ysize-2; j>=1;j--) {
         for(k=zsize-2; k>=1;k--){ grid[i][j][k]= 0.5 * grid[i][j][k] + 0.083333333333333333*
             (grid[i-1][j][k] + grid[i+1][j][k] + grid[i][j+1][k]+
              grid[i][j-1][k] + grid[i][j][k+1] + grid[i][j][k-1]);
			}
		}
	}

}

System.out.println("after smoothing");


  init_marchcubes();
  temp_count = process_array();
  p_edges++;
  connections= new long[(int)num_connections];
System.out.print("before gouraud tess");
  gouraud_tess();
System.out.println("after gouraud tess");
  //make an output file on the server
  VRML_out();


}

/***************************************************************************/
public void init() {

 mbutton = new Button("Begin Marching");

 abox = new Checkbox("ASCII");
 bbox = new Checkbox("Binary");

 xtext= new TextField(20);
 ytext= new TextField(20);
 ztext= new TextField(20);
 ttext= new TextField(20);
 ftext= new TextField(20);
 stext= new TextField(20);

 lbl0 = new Label("enter the following parameters and click the march button");
 lbl1 = new Label("enter the number of points for each dimension");
 lbl2 = new Label("X dimension");
 lbl3 = new Label("Y dimension");
 lbl4 = new Label("Z dimension");
 lbl5 = new Label("enter the threshold");
 lbls = new Label("enter smoothing passes");
 lbl6 = new Label("enter the filename of your dataset");
 lbl7 = new Label("choose which type of data is to be read");

 //setLayout(new BorderLayout());
 add("North",lbl0);
 add("North",lbl1);

 Panel pnl= new Panel();
 pnl.setLayout(new GridLayout(6,2));

 pnl.add(lbl2);
 pnl.add(xtext);

 pnl.add(lbl3);
 pnl.add(ytext);

 pnl.add(lbl4);
 pnl.add(ztext);

 pnl.add(lbl5);
 pnl.add(ttext);
 
 pnl.add(lbls);
 pnl.add(stext);

 pnl.add(lbl6);
 pnl.add(ftext);


 add(pnl);
 add("Center",lbl7);
 add("Center",abox);
 add("Center",bbox);

 add("South",mbutton);


}
/***************************************************************************/
public boolean action(Event evt, Object what){
    if(evt.target == mbutton){

      System.out.println("were here");
        xsize= Integer.valueOf(xtext.getText()).intValue();
        ysize= Integer.valueOf(ytext.getText()).intValue();
        zsize= Integer.valueOf(ztext.getText()).intValue();
        threshold= Double.valueOf(ttext.getText()).doubleValue();
        smoothpass= Integer.valueOf(stext.getText()).intValue();
        if(abox.getState()) ascii=1;
        if(bbox.getState()) binary=1;
        filename= ftext.getText();
        try{
           mrch();
        }
        catch(IOException e){}

        return true;
    } else return false;

}

}

