/* ++++++++++++++++++++++++++++++++++++++++++++++
   +++   SHOW  +++  PICTURE  +++  PROGRAM     +++
   ++++++++++++++++++++++++++++++++++++++++++++++ */

/* +++  Libraries +++ */

#include <fcntl.h>
#include <termio.h>
#include <sys/signal.h>

#define STDIN 0
#define STDOUT 1
#define STDERR 2

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif


/* +++ Main Program +++ */
int ttfile;
/****************************************************************************/
/*  This program renders triangles which have been written out by one of the*/
/*  CHPC's marching cubes programs.                                         */
/****************************************************************************/

#include <stdio.h>
#include "gl.h"
#include "device.h"
#include "math.h"
#include "march.h"

#define PI         3.1415926535
#define TWOPI      (PI * 2.0)
#define PI8        (PI / 8.0)
#define PI16       (PI / 16.0)
#define PI32       (PI / 32.0)
#define PI64       (PI / 64.0)
#define PI128      (PI / 128.0)
#define PI256      (PI / 256.0)
#define PI512      (PI / 512.0)


#define ZCUT	65
#define ZSTEP	4
#define YCUT	256
#define ZPAUSE	38
#define ZSCALE	1.0

#define FPS        30          /* frames per second                */
#define FPR        10          /* frames per recording             */
#define SCENE      1           /* scene number                     */
#define FIRST      1           /* first frame number of scene      */
#define LAST       1           /* last recorded frame of old scene */

/* definitions for hardware shading */

Matrix idmat = {1.0,0.0,0.0,0.0, /* identity matrix */
		0.0,1.0,0.0,0.0,
		0.0,0.0,1.0,0.0,
		0.0,0.0,0.0,1.0};
		
float blackvec[3] = {0.4, 0.4, 0.4};
float greyvec[3] = {0.4, 0.4, 0.4};
float whitevec[3] = {0.4, 0.4, 0.4};
float greenvec[3] = {0.4, 0.4, 0.4};
float redvec[3] = {0.4, 0.4, 0.4};

typedef struct { /* structure fro a 3-D vertex */
   Coord x;
   Coord y;
   Coord z;
   } POINT;
 
 typedef struct { /* 3 vertex lighted polygon struct */
    POINT vertex[4];
    POINT normal[4];
    } POLYGON;

    POLYGON *polygon;

  float inner_surface_1[] = 
   {SPECULAR, 0.9, 0.4, 0.3,
   DIFFUSE,   0.8, 0.2, 0.3,
   ALPHA,    1.0,
   SHININESS, 7.0,
   LMNULL};
  
float outer_surface[] = {
EMISSION,  0.0, 0.0, 0.0,
AMBIENT,   0.2, 0.05, 0.0,
DIFFUSE,   0.2, 0.05, 0.0,
SPECULAR,  0.2, 0.05, 0.0,
SHININESS, 1.4,
ALPHA,     0.7,
LMNULL};

float material_2[] = {
EMISSION,  0.0, 0.0, 0.0,
AMBIENT,   0.0, 0.0, 0.0,
DIFFUSE,   0.0, 0.0, 0.0,
SPECULAR, 0.0, 0.0, 0.0,
SHININESS, 0.0,
ALPHA, 0.0,
LMNULL};

float material_3[] = {
EMISSION, 0.0, 0.3, 0.8,
AMBIENT, 0.0, 0.3, 0.8,
DIFFUSE, 0.0, 0.3, 0.8,
SPECULAR, 0.0, 0.0, 0.0,
SHININESS, 0.0,
ALPHA, 1.0,
LMNULL};

float glass_material[] = {
   EMISSION, 0.0, 0.0, 0.0,
   AMBIENT, 0.0, 0.8, 0.7,
   DIFFUSE, 0.0, 0.8, 0.7,
   SPECULAR, 0.0, 0.0, 0.0,
   SHININESS, 1.0,
   ALPHA, 0.08,
   LMNULL};

  float blue_light[]=
  { LCOLOR, 0.8,0.4,0.4,
    POSITION, -150.0, 0.0, 570.0, 0.0,
    LMNULL};

  float light_model[]=
  { AMBIENT, 0.7, 0.3, 0.3,
    LOCALVIEWER, 1.0,
    ATTENUATION, 4.0, 4.5,
    LMNULL
  }; 

/* hardware shading ends here */
  /* globals */
   
/****/
/****/
/*=================================================================*/
/* subroutines begin here                                          */
/*=================================================================*/
def_light_calc()
{
    lmdef(DEFMATERIAL, 1, 19, outer_surface);
    lmdef(DEFMATERIAL, 5, 13, inner_surface_1);
    lmdef(DEFMATERIAL, 2, 19, material_2);
    lmdef(DEFMATERIAL, 3, 19, material_3);
    lmdef(DEFMATERIAL, 4, 19, glass_material);
    lmdef(DEFLIGHT, 1, 10, blue_light);
    lmdef(DEFLMODEL, 1, 9, light_model);
}

use_simple_light_calc()
{
    lmbind(LIGHT0, 1);
    lmbind(LMODEL, 1);
}


/*=================================================================*/
/* main                                                            */
/*=================================================================*/

 main()


 {
  
  FILE *fp3;
  short grid_flag = 1;
  short stop_flag = 0;
  int xboxsize, yboxsize, zboxsize;
  short val;
  long dev;
  float offset;
  float x,y,z,x1,y1,x2,y2,z1,z2,xtmp;
  short stop;
  float twist;
  float  xrot = 797.0;
/*
  float  xrot = 817.0;
*/
  float  yrot = 1988.0 ;
  float  zrot = 1658.0;
  char buff[50];
  char stx = 0x02;
  char etx = 0x03;
  char *ch ="GS";
  FILE *viewcoord,*fopen();
  char filename[64];
  float thisrot,oldy,oldx,oldz;
  float newx=0.0,newy=0.0,newz=0.0;
  int ycut, zcut;



  printf("Input filename: ");
  gets(filename);



  printf(" Input X bounding box dimension: ");
  scanf("%d", &xboxsize);
  printf("Input Y bounding box dimension: ");
  scanf("%d", &yboxsize);
  printf("Input Z bounding box dimension: ");
  scanf("%d", &zboxsize);

/* ntsc */
  prefposition(0,680,0,500);
  keepaspect(1,1);
  prefsize(680,500); 
/*
  prefposition(0, 1280,0,1024);
  keepaspect(1,1);
  prefsize(1270,984);
*/

  vtalk(); 
  winopen(" M A R C H ");
  RGBmode();
/*  doublebuffer(); */
  gconfig();
/*  lRGBrange(5,5,5,200,200,200,200,0);
  lsetdepth((long)0, (long) 100);
  depthcue(TRUE);
 */ zbuffer(TRUE);
  color(BLACK);
  clear();
  zclear();
  mmode(MVIEWING); 

  color (BLACK);
  x = 0.52*((float) xboxsize);
  y = 0.52*((float) yboxsize);
  z = 0.52*((float) zboxsize);
  x2 = ((float)(xboxsize))/2.0;
  y2 = ((float)(yboxsize))/2.0;
  z2 = ((float)(zboxsize))/2.0;
  ortho(-x, x, -x, x, 0.09, 2000.0);
  ycut = YCUT;
  zcut = ZCUT;

  loadmatrix(idmat);
  blanktime(0);

 /* blendfunction(BF_SA, BF_MSA);
*/
  def_light_calc();
  use_simple_light_calc();
  stop = 0;
/*
  viewcoord = fopen("viewcoord", "r");
  fscanf(viewcoord," %f %f %f", &x, &y, &z);
  fscanf(viewcoord," %f %f %f %f", &x2, &y2, &z2, &twist);
  fclose(viewcoord);
*/
  qdevice(KEYBD);

 /* qreset();
*/ 
/*                                                                     */
/* This section of code polls the keyboard for the selection           */
/* of control keys by the user. The x, y and z variables control       */
/* the x, y and z of the abstract eye point. The vars xyang and xzang  */
/* control the direction of the line of sight emanating from the       */
/* eye point.                                                          */
/*                                                                     */
  ringbell();
  offset = 1.0;
  while (!stop)
   {
    while(qtest())
     {

      dev = qread(&val);

      if(val == 'w')
       {
	viewcoord = fopen("viewcoord","w");
	fprintf(viewcoord," %f %f %f", xrot, yrot, zrot);
	fclose(viewcoord);
       }
      }  /* while qtest */

    if (zcut >= 130) {
       zrot+=5.0;
       xrot+=0.5;
       yrot+=1.5;
       }


     c3f(whitevec);
     clear();
     zclear();
    
         pushmatrix();
          lookat(x,y,z,x2,y2,z2,twist);
          rot(xrot,'X');
          rot(yrot,'Y');
	  rot(zrot,'Z');
	  pushmatrix();
	   translate(-x2,-y2,-z2);
           readfileo(filename,ycut, zcut);

	   readfile();

	   draw_grid(xboxsize,yboxsize,zboxsize);
          popmatrix();
         popmatrix();

	zcut += ZSTEP;
	if (zcut >220) zcut = 230;
      ycut = 256;


     /* control the Panasonic equipment  */

     write(ttfile, &stx, 1);
     write(ttfile, ch, 2);
     write(ttfile, &etx, 1);
     sleep(3);
      swapbuffers();

    if(twist < -3600)
     twist = twist + 3600;
    if(twist > 3600)
     twist = twist - 3600;
    

   } /* while (!stop) */
     
  system("gclear");     /* undo the color mapping before quit */

} /* main */
/*                                                              */
/*                        readfileo - read ther old format      */
/*                                                              */

readfileo(filename,ycut,zcut)
char *filename;
int ycut, zcut;

{
  int i, k,pointer=4;
  float  n;
  POLYGON parray, parray1;
  float normvec1[3];
  float normvec2[3];
  float normvec3[3];
  float vec1[3];
  float vec2[3];
  float vec3[3];
  FILE *fp3,*fopen();
  float temparray[3];

#ifdef FRONT_BUF
 frontbuffer(TRUE); 
#endif


     if((fp3 =
     fopen(filename,"rb"))== NULL)
      {
	fprintf(stderr,"fopen: couldn't open input files\n");
	exit(0);
      }
printf("%s\n", filename);

  /* define material properties of the surface */
     lmbind(MATERIAL,5);
  blendfunction(BF_SA, BF_ONE);
  while( pointer != 0 ) 
   {

      fread(&parray.vertex[0].x,sizeof(float), 1, fp3);
      fread(&parray.vertex[0].y,sizeof(float), 1, fp3);
      fread(&parray.vertex[0].z,sizeof(float), 1, fp3);
      fread(&parray.vertex[1].x,sizeof(float), 1, fp3);
      fread(&parray.vertex[1].y,sizeof(float), 1, fp3);
      fread(&parray.vertex[1].z,sizeof(float), 1, fp3);
      fread(&parray.vertex[2].x,sizeof(float), 1, fp3);
      fread(&parray.vertex[2].y,sizeof(float), 1, fp3);
      fread(&parray.vertex[2].z,sizeof(float), 1, fp3);

      for(i=0;i<3;i++)
      {
       parray.vertex[i].z =(ZSCALE * parray.vertex[i].z);
      }

      fread(&normvec1[0],sizeof(float), 1, fp3);
      fread(&normvec1[1],sizeof(float), 1, fp3);
      fread(&normvec1[2],sizeof(float), 1, fp3);
      fread(&normvec2[0],sizeof(float), 1, fp3);
      fread(&normvec2[1],sizeof(float), 1, fp3);
      fread(&normvec2[2],sizeof(float), 1, fp3);
      fread(&normvec3[0],sizeof(float), 1, fp3);
      fread(&normvec3[1],sizeof(float), 1, fp3);
      pointer = fread(&normvec3[2],sizeof(float), 1, fp3);
       if(parray.vertex[0].y < ((float)ycut) && parray.vertex[0].z < ((float)zcut))
	 {
          for(k=0;k<3;k++)
	   {
	    normvec1[k] = -normvec1[k];
	    normvec2[k] = -normvec2[k];
	    normvec3[k] = -normvec3[k];
	   } 
      lmbind(MATERIAL,1);
      blendfunction(BF_SA, BF_ZERO);
            parray1.vertex[0].x =parray.vertex[0].x + (.2*normvec1[0]);
            parray1.vertex[0].y =parray.vertex[0].y + (.2*normvec1[1]);
            parray1.vertex[0].z =parray.vertex[0].z + (.2*normvec1[2]);
            parray1.vertex[1].x =parray.vertex[1].x + (.2*normvec2[0]);
            parray1.vertex[1].y =parray.vertex[1].y + (.2*normvec2[1]);
            parray1.vertex[1].z =parray.vertex[1].z + (.2*normvec2[2]);
            parray1.vertex[2].x =parray.vertex[2].x + (.2*normvec3[0]);
            parray1.vertex[2].y =parray.vertex[2].y + (.2*normvec3[1]);
            parray1.vertex[2].z =parray.vertex[2].z + (.2*normvec3[2]);

          bgnpolygon();
               n3f(normvec1);
               v3f(&parray1.vertex[0]);
               n3f(normvec2);
               v3f(&parray1.vertex[1]);
               n3f(normvec3);
               v3f(&parray1.vertex[2]);
          endpolygon();
     lmbind(MATERIAL,5);
     blendfunction(BF_SA, BF_ZERO);
          for(k=0;k<3;k++)
	   {
	    normvec1[k] = -normvec1[k];
	    normvec2[k] = -normvec2[k];
	    normvec3[k] = -normvec3[k];
	   } 
          bgnpolygon();
               n3f(normvec1);
               v3f(&parray.vertex[0]);
               n3f(normvec2);
               v3f(&parray.vertex[1]);
               n3f(normvec3);
               v3f(&parray.vertex[2]);
          endpolygon();
         }
   }
fclose(fp3);
#ifdef FRONT_BUF
 frontbuffer(FALSE);
#endif


}
/* ********************************************************************** */

/*                        draw_grid                              */
readfile()

{
  int flag, vert1, vert2, vert3,verts = 0;
  static int flag1 = 1;
  float  vertic[3600*18];
  POLYGON parray;
  float normvec1[3];
  float normvec2[3];
  float normvec3[3];
  float vec1[3];
  float vec2[3];
  float vec3[3];
  FILE *fopen(), *fp1, *fp2;

  /* define material properties of the surface */
     lmbind(MATERIAL,1);

  /* open files */
     if((fp1 =
     fopen("uniques","r"))== NULL)
      {
	fprintf(stderr,"fopen: couldn't open input files\n");
	exit(0);
      }	
     if((fp2 =
     fopen("conns","r"))== NULL)
      {
	fprintf(stderr,"fopen: couldn't open input files\n");
	exit(0);
      }

  while(( flag != EOF)&(flag1)) 
   {

    flag = fscanf(fp1,"%f",&vertic[verts]);
/*    if (verts++ % 600 == 0)
     fprintf(stderr,"Inserted unique vertice #%d\n", verts);
 */ verts++; }
   flag1 = 0;

   flag++;
   
 /*  fprintf(stderr," ****Now reading connectivity information**\n");
  */ verts = 0;

   while( flag != EOF )
    {
      flag = fscanf(fp2, " %d %d %d", &vert1, &vert2, &vert3 );
      parray.vertex[0].x = vertic[6*vert1];
      parray.vertex[0].y = vertic[(6*vert1) + 1];
      parray.vertex[0].z = vertic[(6*vert1) + 2];
      parray.vertex[1].x = vertic[6*vert2];
      parray.vertex[1].y = vertic[(6*vert2) + 1];
      parray.vertex[1].z = vertic[(6*vert2) + 2];
      parray.vertex[2].x = vertic[6*vert3];
      parray.vertex[2].y = vertic[(6*vert3) + 1];
      parray.vertex[2].z = vertic[(6*vert3) + 2];
      normvec1[0] = vertic[(6*vert1)+3];
      normvec1[1] = vertic[(6*vert1)+4];
      normvec1[2] = vertic[(6*vert1)+5];
      normvec2[0] = vertic[(6*vert2)+3];
      normvec2[1] = vertic[(6*vert2)+4];
      normvec2[2] = vertic[(6*vert2)+5];
      normvec3[0] = vertic[(6*vert3)+3];
      normvec3[1] = vertic[(6*vert3)+4];
      normvec3[2] = vertic[(6*vert3)+5];

          bgnpolygon();
               n3f(normvec1);
               v3f(&parray.vertex[0]);
               n3f(normvec2);
               v3f(&parray.vertex[1]);
               n3f(normvec3);
               v3f(&parray.vertex[2]);
          endpolygon();

   }
fclose(fp1);
fclose(fp2);
} /* readfile */
/* ********************************************************************** */
/*                                                              */

draw_grid(xboxsize,yboxsize,zboxsize)
  int xboxsize, yboxsize, zboxsize;

{
  float vert1[3];
  float vert2[3];
  float vert3[3];
  float vert4[3];
  float vert5[3];
  float vert6[3];
  float vert7[3];
  float vert8[3];
  char buff[50];
      lmbind(MATERIAL,4); 
      vert1[0] = 0.0;
      vert1[1] = 0.0;
      vert1[2] = 0.0;
      vert2[0] = (float)xboxsize;
      vert2[1] = 0.0;
      vert2[2] = 0.0;
      vert3[0] = (float)xboxsize;
      vert3[1] = (float)yboxsize;
      vert3[2] = 0.0;
      vert4[0] = 0.0;
      vert4[1] = (float)yboxsize;
      vert4[2] = 0.0;
      vert5[0] = 0.0;
      vert5[1] = 0.0;
      vert5[2] = ZSCALE*(float)zboxsize;
      vert6[0] = (float)xboxsize;
      vert6[1] = 0.0;
      vert6[2] = ZSCALE*(float)zboxsize;
      vert7[0] = (float)xboxsize;
      vert7[1] = (float)yboxsize;
      vert7[2] = ZSCALE*(float)zboxsize;
      vert8[0] = 0.0;
      vert8[1] = (float)yboxsize;
      vert8[2] = ZSCALE*(float)zboxsize;
         linewidth(3);


          bgnclosedline();
	       v3f(vert1);
	       v3f(vert2);
	       v3f(vert3);
	       v3f(vert4);
	       v3f(vert1);
	       v3f(vert5);
	       v3f(vert6);
	       v3f(vert2);
	       v3f(vert6);
	       v3f(vert7);
	       v3f(vert3);
	       v3f(vert7);
	       v3f(vert8);
	       v3f(vert4);
	       v3f(vert8);
	       v3f(vert5);
          endclosedline();

	  lmbind(MATERIAL,1);
          bgnpolygon();
	       v3f(vert1);
	       v3f(vert4);
	       v3f(vert1);
	       v3f(vert5);
	       v3f(vert1);
	       v3f(vert2);
          endpolygon();
	  
     lmbind(MATERIAL,1);
     sprintf(buff,"X");
     cmov(((Coord)xboxsize),((Coord)0),((Coord)0)); charstr(buff);
     sprintf(buff,"Y");
     cmov(((Coord)0),((Coord)yboxsize),((Coord)0)); charstr(buff);
     sprintf(buff,"Z");
     cmov(((Coord)0),((Coord)0),((Coord)zboxsize)); charstr(buff);


   }      
 /* draw_grid  */
/*************************************************************/
/*************************************************************/
vtalk()
{
static char lldev[20]="/dev/ttyd2";
register temp;
char cc;
char *filename1;
struct termio termdata,prevdata;
filename1=lldev;
/* try to open the device file */
ttfile = open( filename1, O_RDWR );
if ( ttfile < 0 )
  {
  fprintf( stderr, "Can't open file '%s'.\n", filename1 );
  exit( 1 );
  }

/* fetch current terminal modes for tty device */
temp = ioctl( ttfile, TCGETA, &termdata );
if ( temp == -1 )
  {
  fprintf( stderr, "Can't get modes for '%s'.\n", filename1 );
  exit( 1 );
  }

/* change to the correct modes */
termdata.c_iflag = 0;  /* no special input processing */
termdata.c_oflag = 0;  /* no special output processing */
termdata.c_cflag = 0
                 | B1200    /* 1200 baud */
                 | CS8      /* 8 data bits */
                 & ~CSTOPB  /* one stop bit */
                 | CREAD    /* enable receiver */
                 & ~PARENB  /* no parity */
                 | CLOCAL;  /* local line (not dialup) */
termdata.c_lflag = 0;       /* no processing */
termdata.c_cc[VMIN] = 1;

/* set the modes */
temp = ioctl( ttfile, TCSETA, &termdata );
if ( temp == -1 )
  {
  fprintf( stderr, "Can't set modes for '%s'.\n", filename1 );
  exit( 1 );
  }
}
/* ********************************************************************** */
