/* ++++++++++++++++++++++++++++++++++++++++++++++   
             SURFACE PLOT PROGRAM  
   ++++++++++++++++++++++++++++++++++++++++++++++ */

/* +++  Libraries +++ */

#include <gl.h>
#include <math.h>
#include <stdio.h>
#include <device.h>

/* +++ Function Definitions +++ */
 
#define MAX(X,Y)    ( (X) > (Y) ? (X) : (Y) )
#define MIN(X,Y)    ( (X) < (Y) ? (X) : (Y) )
#define ABS(X)      ( (X) < 0 ? -(X) : (X) )

#define NX 64
#define NY 64 

/* +++ Structures +++ */

typedef struct {               
        Coord x;
        Coord y;
        Coord z;
} POINT;

typedef struct {               
        POINT vertex[4];
        POINT normal; 
} POLYGON;

typedef struct {
	POINT esquina[8];
	POINT donde[6];
} BOX;

/* +++ Pointers ++ */

POLYGON  *polygon;              
BOX      *caja;                 
FILE     *filein;              

/* +++ Global Data declarations +++ */

int pict_r = 0;                       
int pict_n;                
int corx[4] = {0.0, 1.0, 1.0, 0.0};
int cory[4] = {0.0, 0.0, 1.0, 1.0};       
int TOT_POLY;

char in_file[20];

float raw_data[NX][NY]; 
float blackvec[3]  = {0.0, 0.0, 0.0}; 

Matrix idmat = {1.0, 0.0, 0.0, 0.0,   /* identity matrix */
                0.0, 1.0, 0.0, 0.0,
                0.0, 0.0, 1.0, 0.0,
                0.0, 0.0, 0.0, 1.0};  

/* +++ Material Definitions +++ */

float material_1[] = {
      EMISSION,  0.0, 0.0, 0.0,
      AMBIENT,   0.4, 0.0, 0.1,   
      DIFFUSE,   0.4, 0.2, 0.1,   
      SPECULAR,  0.0, 0.0, 0.0,   
      SHININESS, 0.0,            
      ALPHA,     0.3,
      LMNULL};

float material_2[] = {
      EMISSION,  0.0, 0.0, 0.0,
      AMBIENT,   0.6, 0.4, 0.2,   
      DIFFUSE,   0.6, 0.4, 0.2,   
      SPECULAR,  0.0, 0.0, 0.0,   
      SHININESS, 30.0,            
      ALPHA,     1.0,
      LMNULL};

float material_3[] = {
      EMISSION,  0.0, 0.0, 0.0,
      AMBIENT,   0.2, 0.2, 0.2,   
      DIFFUSE,   0.4, 0.4, 0.4,   
      SPECULAR,  0.8, 0.8, 0.8,   
      SHININESS, 30.0,            
      ALPHA,     1.0,
      LMNULL};

float purple_material[] = {
      SPECULAR,  0.3, 0.3, 0.3,   
      DIFFUSE,   0.8, 0.0, 0.8,   
      SHININESS, 3.0,            
      AMBIENT,   0.2, 0.0, 0.2,   
      LMNULL};

float shiny_material[] = {
      SPECULAR,  0.0, 0.8, 0.8,   
      DIFFUSE,   0.0, 0.4, 0.4,   
      SHININESS, 30.0,            
      LMNULL};

float glass_material[] = {
      EMISSION,  0.0, 0.0, 0.0,
      AMBIENT,   1.0, 1.0, 1.0,   
      DIFFUSE,   1.0, 1.0, 1.0,   
      SPECULAR,  0.0, 0.0, 0.0,   
      SHININESS, 1.0,            
      ALPHA,     0.08,
      LMNULL};

float blue_light[] = {
      LCOLOR,    0.0, 0.0, 0.6,
      POSITION,  1.0, 1.0, 1.0, 0.0,
      LMNULL};

float light_1[] = {
      AMBIENT,   0.2, 0.2, 0.2,
      LCOLOR,    0.9, 0.9, 0.9,
      POSITION,  1.0, 1.0, 1.0, 0.0,
      LMNULL};

float ambient_1[] = {
      AMBIENT,     0.2, 0.2, 0.2,
      LOCALVIEWER, 0.0,
      ATTENUATION, 1.0, 1.0,
      LMNULL};  

/* +++ Main Program +++ */

/************************************************
** Driver of the plotting program 
************************************************/

main()
{
   int movie = 0;
   Coord x_v = 3.0;
   Coord y_v = 1.6;
   Coord z_v = 3.0;
   float x_a = 0.0;
   float y_a = 0.0;
   float z_a = 0.0;

/* Ask for input file of raw data */

   printf("\t Hi Pablo. Please enter:\n");
   printf("\t\t Input file name = ");
   scanf("%s", in_file);
   printf("\t\t No. of pictures = ");
   scanf("%d", &pict_n);

/* Make box */

   make_box();

/* Set graphics environment */

   graph_environment();

/* Let there be light !!!!! */

   def_light_calc();
   use_light_calc();

/* Plot and rotate figure */

   make_picture();
   for ( ; ; ) {
       c3f(blackvec);                 /* set backgroud color */
       clear();                       /* clear frame */
       zclear();                      /* clear buffer */
       pushmatrix();                  /* make a frame */
          lookat(x_v,y_v,z_v,0.0,0.0,0.0,0);       
	  rot(x_a,'X');
	  rot(y_a,'Y');
	  rot(z_a,'Z');
	  draw_frame(); 
/*	  glass_box();   */
       popmatrix();                   /* pop next picture in buffer */
       swapbuffers();                 /* swap buffers */  

           if (getbutton(RIGHTARROWKEY)) x_v += 0.2; 
           if (getbutton(LEFTARROWKEY))  x_v -= 0.2; 
           if (getbutton(UPARROWKEY))    y_v += 0.2; 
           if (getbutton(DOWNARROWKEY))  y_v -= 0.2; 
           if (getbutton(XKEY))          z_v += 0.2; 
           if (getbutton(ZKEY))          z_v -= 0.2; 

           if (getbutton(PAD6))      x_a += 1.0;
           if (getbutton(PAD4))      x_a -= 1.0;
           if (getbutton(PAD8))      y_a += 1.0;
           if (getbutton(PAD2))      y_a -= 1.0;
           if (getbutton(PERIODKEY)) z_a += 1.0;
           if (getbutton(COMMAKEY))  z_a -= 1.0;

	   if (getbutton(SKEY)) movie = 0; 
	   if (getbutton(AKEY)) movie = 1; 
	   if (getbutton(FKEY)) movie = 2;

       switch (movie)
           {
	   case 0: break;
           case 1: make_picture();
		   break;
           case 2: make_picture();
		   movie = 0;
		   break;
           }

     printf("\n View = (%3.1f,%3.1f,%3.1f) Rot = (%3.1f,%3.1f,%3.1f) Pict = %d",
	    x_v,y_v,z_v,x_a,y_a,z_a,pict_r);

    }
}
    
/* +++ Subroutines +++ */

/**********************************************************
** Set graphics environment. Use mmode() to set up 
** projection and viewing matrices for lighting 
***********************************************************/
graph_environment()
{
   prefposition(100,1100,50,850);          /* set window size */
   winopen("PABLO's PLOT");                /* open window */
   RGBmode();                              /* set RGB mode on */
   doublebuffer();                         /* set double buffer */
   gconfig();                              /* reconfigure */
   lsetdepth(0, 0x7FFFFF);                 /* set 3-D viewport */
   zbuffer(TRUE);                          /* start z-buffer mode */
   mmode(MVIEWING);                        /* set current matrix mode */
   perspective(500,1.25,1.0,15.0);         /* perspective porj. trans. */
   loadmatrix(idmat);                      /* load identity matrix */
}

/******************************************************************
** def_light_calc()
** Tell de Graphics Library to DEFINE a simple lighting calculation
** that accounts for diffuse and ambient reflection. This simple
** lighting calculation happens to use the default lighting
** parameters in the Graphis Library.
********************************************************************/
def_light_calc() 
{
    lmdef(DEFMATERIAL, 1, 21, material_1);
    lmdef(DEFMATERIAL, 2, 21, material_2);
    lmdef(DEFMATERIAL, 3, 21, material_3);
    lmdef(DEFMATERIAL, 4, 21, glass_material);
    lmdef(DEFMATERIAL, 5, 11, shiny_material);
    lmdef(DEFMATERIAL, 6, 15, purple_material);

    lmdef(DEFLIGHT, 1, 0, NULL);
    lmdef(DEFLIGHT, 2, 14, light_1);
    lmdef(DEFLIGHT, 3, 10, blue_light);

    lmdef(DEFLMODEL, 1, 0, NULL); 
    lmdef(DEFLMODEL, 2, 10, ambient_1); 
}

/*************************************************
** use_light_calc() 
** Tell the Graphics LIbrary to USE the simple 
** lighting calculation that we defined earlier.
***************************************************/
use_light_calc()
{
    lmbind(LIGHT0, 1);     
    lmbind(LIGHT1, 2);     
    lmbind(LMODEL, 1);   
}

/****************************************************
** Draw contour label in the picture
*****************************************************/
draw_frame()
{
   POLYGON *ss;             /* pointer into the picture list */
   int p_i;                 /* polygon counter */
   int v_i;                 /* vertex counter */
   
   ss = polygon;
   lmbind(MATERIAL, 5);
   for (p_i = 0; p_i < TOT_POLY; p_i++) {
       bgnpolygon();
       for (v_i = 0; v_i < 4; v_i++) {
           n3f(&ss->normal);   
           v3f(&ss->vertex[v_i]);  
       }
       endpolygon();
       ss++;
   }
   free(ss);
}

/********************************************************
** Glass box 
*********************************************************/
glass_box()
{
   lmbind(MATERIAL, 4);

   bgnpolygon();
      n3f(&caja->donde[0]);
      v3f(&caja->esquina[0]);
      n3f(&caja->donde[0]);
      v3f(&caja->esquina[1]);
      n3f(&caja->donde[0]);
      v3f(&caja->esquina[3]);
      n3f(&caja->donde[0]);
      v3f(&caja->esquina[2]);
   endpolygon();

   bgnpolygon();
      n3f(&caja->donde[1]);
      v3f(&caja->esquina[4]);
      n3f(&caja->donde[1]);
      v3f(&caja->esquina[5]);
      n3f(&caja->donde[1]);
      v3f(&caja->esquina[7]);
      n3f(&caja->donde[1]);
      v3f(&caja->esquina[6]);
   endpolygon();

   bgnpolygon();
      n3f(&caja->donde[2]);
      v3f(&caja->esquina[0]);
      n3f(&caja->donde[2]);
      v3f(&caja->esquina[4]);
      n3f(&caja->donde[2]);
      v3f(&caja->esquina[5]);
      n3f(&caja->donde[2]);
      v3f(&caja->esquina[1]);
   endpolygon();

   bgnpolygon();
      n3f(&caja->donde[3]);
      v3f(&caja->esquina[1]);
      n3f(&caja->donde[3]);
      v3f(&caja->esquina[5]);
      n3f(&caja->donde[3]);
      v3f(&caja->esquina[7]);
      n3f(&caja->donde[3]);
      v3f(&caja->esquina[3]);
   endpolygon();

   bgnpolygon();
      n3f(&caja->donde[4]);
      v3f(&caja->esquina[3]);
      n3f(&caja->donde[4]);
      v3f(&caja->esquina[7]);
      n3f(&caja->donde[4]);
      v3f(&caja->esquina[6]);
      n3f(&caja->donde[4]);
      v3f(&caja->esquina[2]);
   endpolygon();

   bgnpolygon();
      n3f(&caja->donde[5]);
      v3f(&caja->esquina[2]);
      n3f(&caja->donde[5]);
      v3f(&caja->esquina[6]);
      n3f(&caja->donde[5]);
      v3f(&caja->esquina[4]);
      n3f(&caja->donde[5]);
      v3f(&caja->esquina[0]);
   endpolygon();

   linewidth(0);

}

/****************************************************
** Make box
*****************************************************/
make_box()
{
   int i;

   caja = (BOX *) malloc(sizeof(BOX));

   caja->esquina[0].x = -1.0;
   caja->esquina[0].y = -1.0;
   caja->esquina[0].z = -1.0;

   caja->esquina[1].x = 1.0;
   caja->esquina[1].y = -1.0;
   caja->esquina[1].z = -1.0;

   caja->esquina[2].x = -1.0;
   caja->esquina[2].y = 1.0;
   caja->esquina[2].z = -1.0;

   caja->esquina[3].x = 1.0;
   caja->esquina[3].y = 1.0;
   caja->esquina[3].z = -1.0;

   caja->esquina[4].x = -1.0;
   caja->esquina[4].y = -1.0;
   caja->esquina[4].z = 1.0;

   caja->esquina[5].x = 1.0;
   caja->esquina[5].y = -1.0;
   caja->esquina[5].z = 1.0;

   caja->esquina[6].x = -1.0;
   caja->esquina[6].y = 1.0;
   caja->esquina[6].z = 1.0;

   caja->esquina[7].x = 1.0;
   caja->esquina[7].y = 1.0;
   caja->esquina[7].z = 1.0;

   caja->donde[0].x =  0.0;
   caja->donde[0].y =  0.0;
   caja->donde[0].z = -1.0;

   caja->donde[1].x =  0.0;
   caja->donde[1].y =  0.0;
   caja->donde[1].z =  1.0;

   caja->donde[2].x =  0.0;
   caja->donde[2].y = -1.0;
   caja->donde[2].z =  0.0;

   caja->donde[3].x =  1.0;
   caja->donde[3].y =  0.0;
   caja->donde[3].z =  0.0;

   caja->donde[4].x =  0.0;
   caja->donde[4].y =  1.0;
   caja->donde[4].z =  0.0;

   caja->donde[5].x = -1.0;
   caja->donde[5].y =  0.0;
   caja->donde[5].z =  0.0;

}
    
make_surface()
{
   int i, j, indx, xx, yy;
   float n0, n1, n2, norm;
   float SCLX = 2.0/(NX-1);
   float SCLY = 2.0/(NY-1);
   float SCLZ = 2.0/255.0;
   float OFFX = (NX-1)/2.0;
   float OFFY = (NY-1)/2.0;
   float OFFZ = 255.0/2.0;

   POLYGON *pp;
   pp = polygon = (POLYGON *) malloc(NX*NY*sizeof(POLYGON));
   TOT_POLY = 0;
   for (i = 0; i < NX-1; i++) {
   for (j = 0; j < NY-1; j++) {
       for ( indx = 0; indx < 4; indx++ ) {
           xx = i+corx[indx];
           yy = j+cory[indx];
           pp->vertex[indx].z = SCLX*(xx-OFFX);
           pp->vertex[indx].x = SCLY*(yy-OFFY);
           pp->vertex[indx].y = SCLZ*(raw_data[xx][yy]-OFFZ);
       }
       n0 = raw_data[i][j]-raw_data[i+1][j];
       n1 = raw_data[i][j]-raw_data[i][j+1];
       n2 = 1.0;
       norm = sqrt(n0*n0+n1*n1+n2*n2);
       pp->normal.z = n0/norm;
       pp->normal.x = n1/norm;
       pp->normal.y = n2/norm;
       pp++;
       TOT_POLY++;
   }
   }
}

make_picture()
{
    int i,j;
    int size = NX*NY;     
    char indat[NX][NY];          
    if (pict_r > pict_n) return 0;
    if (pict_r == 0) {
       if ((filein = fopen(in_file,"rb")) == NULL)
          printf("\t !!! No puedo abrir el pinche file !!!");
    }
    if (polygon != NULL) free(polygon);
    fread(indat, size, 1, filein);
    pict_r++;
    for (i = 0; i < NX; i++) {
    for (j = 0; j < NY; j++) {
        raw_data[i][j] = ABS((float) indat[i][j]);
    }
    }
    make_surface();
}
