      program converge
c
      implicit none
c
      integer outbound
      parameter (outbound=2) ! outbound=1->Dirichlet, outbound=2->Robin
      real xmax, ymax, zmax
      parameter (xmax=5.,ymax=5.,zmax=5.)
c
      integer nxl, nyl, nzl, il, jl, kl
      parameter (nxl=11,nyl=11,nzl=11)
      real dxl, dyl, dzl
      real xl(nxl), yl(nyl), zl(nzl), ul(nxl,nyl,nzl), uan(nxl,nyl,nzl)
      real rel(nxl,nyl,nzl), rem(nxl,nyl,nzl)
c
      integer nxm, nym, nzm, im, jm, km
      parameter (nxm=21,nym=21,nzm=21)
      real um(nxm,nym,nzm)
c
      integer nxh, nyh, nzh, ih, jh, kh
      parameter (nxh=41,nyh=41,nzh=41)
      real uh(nxh,nyh,nzh)
c
      integer imin, jmin, kmin, imax, jmax, kmax
      real r, norml, normm, normh
      real sigma(nxl,nyl,nzl), sigmaave, sigmasigma, sigmanorm
c
      dxl = 2.*xmax/real(nxl-1)
      dyl = 2.*ymax/real(nyl-1)
      dzl = 2.*zmax/real(nzl-1)
c
      call initcoord(nxl,nyl,nzl,dxl,dyl,dzl,xl,yl,zl)
c
c Read in function at each resolution:
c
      open(10,file='u.n11.out',status='unknown',form='unformatted',
     &     access='sequential')
      read(10) ul
      close(10)
      write(*,*) ul(2,2,2)
c
      open(10,file='u.n21.out',status='unknown',form='unformatted',
     &     access='sequential')
      read(10) um
      close(10)
c
      open(10,file='u.n41.out',status='unknown',form='unformatted',
     &     access='sequential')
      read(10) uh
      close(10)
c
c Specify analytic value of u:
c
      do kl = 1,nzl
         do jl = 1,nyl
            do il=1,nxl
               r = sqrt(xl(il)**2+yl(jl)**2+zl(kl)**2)
               uan(il,jl,kl) = exp(-r**2)
            enddo
         enddo
      enddo
c
c Compute pointwise convergence rate of solution error:
c (Don't look at end points for Dirichlet boundary condition.)
c
      if (outbound.eq.1) then
         imin = 2
         jmin = 2
         kmin = 2
         imax = nxl-1
         jmax = nyl-1
         kmax = nzl-1
      elseif (outbound.eq.2) then
         imin = 1
         jmin = 1
         kmin = 1
         imax = nxl
         jmax = nyl
         kmax = nzl
      endif
c
      do kl = kmin,kmax
         km = 2*kl-1
         do jl = jmin,jmax
            jm = 2*jl-1
            do il = imin,imax
               im = 2*il-1
               sigma(il,jl,kl) = log((um(im,jm,km)-uan(il,jl,kl))/
     &              (ul(il,jl,kl)-uan(il,jl,kl)))/log(0.5)
            enddo
         enddo
      enddo
c
      sigmaave = 0.
      do kl = kmin,kmax
         do jl = jmin,jmax
            do il = imin,imax
               sigmaave = sigmaave + sigma(il,jl,kl)
            enddo
         enddo
      enddo
      sigmaave = sigmaave/real((imax-imin+1)*(jmax-jmin+1)*(kmax-kmin+1))
c
      sigmasigma = 0.
      do kl = kmin,kmax
         do jl = jmin,jmax
            do il = imin,imax
               sigmasigma = sigmasigma + (sigma(il,jl,kl)-sigmaave)**2
            enddo
         enddo
      enddo
      sigmasigma = sqrt(sigmasigma/real((imax-imin+1)*(jmax-jmin+1)*
     &     (kmax-kmin+1)-1))
c
      write(*,*) 'Average convergence rate: ', sigmaave
      write(*,*) 'Standard deviation of convergence rate: ', sigmasigma
c
c Compute convergence rate of L2 norm of solution error:
c
      norml = 0.
      normm = 0.
      do kl = kmin,kmax
         km = 2*kl-1
         do jl = jmin,jmax
            jm = 2*jl-1
            do il = imin,imax
               im = 2*il-1
               norml = norml + (ul(il,jl,kl)-uan(il,jl,kl))**2
               normm = normm + (um(im,jm,km)-uan(il,jl,kl))**2
            enddo
         enddo
      enddo
      norml = sqrt(norml)
      normm = sqrt(normm)
      sigmanorm = log(normm/norml)/log(0.5)
      write(*,*) 'Convergence rate of L2 norm: ', sigmanorm
c
c Compute Richardson extrapolants:
c
      do kl = kmin,kmax
         km = 2*kl-1
         kh = 2*km-1
         do jl = jmin,jmax
            jm = 2*jl-1
            jh = 2*jm-1
            do il = imin,imax
               im = 2*il-1
               ih = 2*im-1
               rel(il,jl,kl) = (4.*um(im,jm,km)-ul(il,jl,kl))/3.
               rem(il,jl,kl) = (4.*uh(ih,jh,kh)-um(im,jm,km))/3.
            enddo
         enddo
      enddo
c
      open(10,file='rel.xline',status='unknown')
      do il = imin,imax
         write(10,*) xl(il), rel(il,(nyl+1)/2,(nzl+1)/2)-
     &        uan(il,(nyl+1)/2,(nzl+1)/2)
      enddo
      close(10)
c
      open(10,file='rem.xline',status='unknown')
      do il = imin,imax
         write(10,*) xl(il), rem(il,(nyl+1)/2,(nzl+1)/2)-
     &        uan(il,(nyl+1)/2,(nzl+1)/2)
      enddo
      close(10)
c
c Compute convergence rate of Richardson extrapolants:
c
      do kl = kmin,kmax
         do jl = jmin,jmax
            do il = imin,imax
               sigma(il,jl,kl) = log((rem(il,jl,kl)-uan(il,jl,kl))/
     &              (rel(il,jl,kl)-uan(il,jl,kl)))/log(0.5)
            enddo
         enddo
      enddo
c
      sigmaave = 0.
      do kl = kmin,kmax
         do jl = jmin,jmax
            do il = imin,imax
               sigmaave = sigmaave + sigma(il,jl,kl)
            enddo
         enddo
      enddo
      sigmaave = sigmaave/real((imax-imin+1)*(jmax-jmin+1)*(kmax-kmin+1))
c
      sigmasigma = 0.
      do kl = kmin,kmax
         do jl = jmin,jmax
            do il = imin,imax
               sigmasigma = sigmasigma + (sigma(il,jl,kl)-sigmaave)**2
            enddo
         enddo
      enddo
      sigmasigma = sqrt(sigmasigma/real((imax-imin+1)*(jmax-jmin+1)*
     &     (kmax-kmin+1)-1))
c
      write(*,*) 'Average convergence rate of RE: ', sigmaave
      write(*,*) 'Standard deviation of convergence rate: ', sigmasigma
c
c Compute convergence rate of L2 norm of extrapolants:
c
      norml = 0.
      normm = 0.
      do kl = kmin,kmax
         do jl = jmin,jmax
            do il = imin,imax
               norml = norml + (rel(il,jl,kl)-uan(il,jl,kl))**2
               normm = normm + (rem(il,jl,kl)-uan(il,jl,kl))**2
            enddo
         enddo
      enddo
      norml = sqrt(norml)
      normm = sqrt(normm)
      sigmanorm = log(normm/norml)/log(0.5)
      write(*,*) 'Convergence rate of L2 norm: ', sigmanorm
c
      stop
      end
