      subroutine outerbc(nx,ny,nz,nb,itint,dx,dy,dz,x,y,z,psi,rhs,res,
     &     dirichlet)
c
c This routine applies the outer boundary condition for the
c Jacobi iteration.
c
      implicit none
c
      integer nx, ny, nz, nb, itint
      real dx, dy, dz
      real x(nx), y(ny), z(nz), psi(nx,ny,nz)
      real rhs(nb), res(nb)
      logical dirichlet
c
c Local variables:
c
      integer i, j, k, it, itmax, nnb
      parameter (itmax=1000)
      real resnorm, dlam, tol
      parameter (tol=1.e-13)
c
      if (dirichlet) then
         return
      else
         call obres(nx,ny,nz,nb,itint,dx,dy,dz,x,y,z,psi,res,rhs)
c
         resnorm = 0.
         do i = 1,nb
            if (abs(res(i)).gt.resnorm) resnorm = abs(res(i))
         enddo
c         write(*,*) 'outerbc:', resnorm
         if (resnorm.lt.tol) return
c
c Relaxation loop:
c
         dlam = dx/50.
c
         do it = 1,itmax
            nnb = 0
            i = 1
               do j = 1,ny
                  do k = 1,nz
                     nnb = nnb+1
                     if (mod(i+j+k,2).eq.mod(itint,2)) then
                        psi(i,j,k) = psi(i,j,k) + dlam*res(nnb)
                     endif
                  enddo
               enddo

            do i = 2,nx-1
               j = 1
                  do k = 1,nz
                     nnb = nnb+1
                     if (mod(i+j+k,2).eq.mod(itint,2)) then
                        psi(i,j,k) = psi(i,j,k) + dlam*res(nnb)
                     endif
                  enddo

               do j = 2,ny-1
                  k = 1
                     nnb = nnb+1
                     if (mod(i+j+k,2).eq.mod(itint,2)) then
                        psi(i,j,k) = psi(i,j,k) + dlam*res(nnb)
                     endif
                  k = nz
                     nnb = nnb+1
                     if (mod(i+j+k,2).eq.mod(itint,2)) then
                        psi(i,j,k) = psi(i,j,k) + dlam*res(nnb)
                     endif
               enddo

               j = ny
                  do k = 1,nz
                     nnb = nnb+1
                     if (mod(i+j+k,2).eq.mod(itint,2)) then
                        psi(i,j,k) = psi(i,j,k) + dlam*res(nnb)
                     endif
                  enddo
            enddo

            i = nx
               do j = 1,ny
                  do k = 1,nz
                     nnb = nnb+1
                     if (mod(i+j+k,2).eq.mod(itint,2)) then
                        psi(i,j,k) = psi(i,j,k) + dlam*res(nnb)
                     endif
                  enddo
               enddo
c
c Compute the residual of the new boundary:
c
            call obres(nx,ny,nz,nb,itint,dx,dy,dz,x,y,z,psi,res)
c
c Still using max norm:
c
            resnorm = 0.
            do i = 1,nb
               if (abs(res(i)).gt.resnorm) resnorm = abs(res(i))
            enddo
c            write(*,*) 'outerbc:', resnorm
c
            if (resnorm.lt.tol) then
c               write(*,*) 'outerbc converged in ', it, 'iterations.'
               return
            endif
         enddo
      endif
      write(*,*) 'outerbc failed to converge in ', itmax, ' iterations.'
      stop
c
      return
      end
c
c******************************************************************************
c
      subroutine obres(nx,ny,nz,nb,itint,dx,dy,dz,x,y,z,psi,res,rhs)
c
c This routine computes the residual for the outer boundary.
c Second order accurate derivatives are used when centered
c differencing can be done, otherwise fourth order derivatives
c are used.
c
      implicit none
c
      integer nx, ny, nz, nb, itint
      real dx, dy, dz
      real x(nx), y(ny), z(nz)
      real psi(nx,ny,nz), res(nb), rhs(nb)
c
c Local variables
c
      integer i, j, k, nnb
      real px, py, pz
c
      nnb = 0
c
      i = 1
         j = 1
            k = 1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_forward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_forward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_forward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
c
               do k = 2,nz-1
                  nnb = nnb+1
                  if (mod(i+j+k,2).eq.mod(itint,2)) then
                     call fourth_forward_x(nx,ny,nz,i,j,k,dx,psi,px)
                     call fourth_forward_y(nx,ny,nz,i,j,k,dy,psi,py)
                     call second_center_z(nx,ny,nz,i,j,k,dz,psi,pz)
                     res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
                  else
                     res(nnb) = 0.
                  endif
               enddo
c
            k = nz
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_forward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_forward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_backward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
c
         do j = 2,ny-1
            k = 1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_forward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call second_center_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_forward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
c
            do k = 2,nz-1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_forward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call second_center_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call second_center_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
            enddo
c
            k = nz
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_forward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call second_center_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_backward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
         enddo
c
         j = ny
            k = 1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_forward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_backward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_forward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
c
            do k = 2,nz-1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_forward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_backward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call second_center_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
            enddo
c
            k = nz
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_forward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_backward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_backward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
c
      do i = 2,nx-1
         j = 1
            k = 1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call second_center_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_forward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_forward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
c
            do k = 2,nz-1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call second_center_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_forward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call second_center_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
            enddo
c
            k = nz
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call second_center_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_forward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_backward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
c
         do j = 2,ny-1
            k = 1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call second_center_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call second_center_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_forward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
c
            k = nz
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call second_center_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call second_center_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_backward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
         enddo
c
         j = ny
            k = 1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call second_center_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_backward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_forward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
c
            do k = 2,nz-1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call second_center_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_backward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call second_center_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
            enddo
c
            k = nz
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call second_center_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_backward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_backward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
      enddo
c
      i = nx
         j = 1
            k = 1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_backward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_forward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_forward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
c
            do k = 2,nz-1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_backward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_forward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call second_center_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
            enddo
c
            k = nz
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_backward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_forward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_backward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
c
         do j = 2,ny-1
            k = 1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_backward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call second_center_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_forward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
c
            do k = 2,nz-1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_backward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call second_center_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call second_center_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
            enddo
c
            k = nz
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_backward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call second_center_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_backward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
         enddo
c
         j = ny
            k = 1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_backward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_backward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_forward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
c
            do k = 2,nz-1
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_backward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_backward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call second_center_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
            enddo
c
            k = nz
               nnb = nnb+1
               if (mod(i+j+k,2).eq.mod(itint,2)) then
                  call fourth_backward_x(nx,ny,nz,i,j,k,dx,psi,px)
                  call fourth_backward_y(nx,ny,nz,i,j,k,dy,psi,py)
                  call fourth_backward_z(nx,ny,nz,i,j,k,dz,psi,pz)
                  res(nnb) = x(i)*px+y(j)*py+z(k)*pz+psi(i,j,k)-rhs(nnb)
               else
                  res(nnb) = 0.
               endif
c
      do nnb = 1,nb
         res(nnb) = -res(nnb)
      enddo
c
      return
      end
