      subroutine sor(nx,ny,nz,nb,dx,dy,dz,x,y,z,cc,ce,cw,cn,cs,ct,cb,
c     &     cne,cnw,cse,csw,cte,ctw,cbe,cbw,ctn,cts,cbn,cbs,
     &     rhs,u,res,obrhs,obres,obdirichlet)
c
c This routine performs Gauss-Seidel iteration on a 3D grid.
c A 19-point stencil is assumed.
c
      implicit none
c
      integer nx, ny, nz, nb
      real dx, dy, dz
c
      real x(nx), y(ny), z(nz)
c
      real cc(nx,ny,nz), ce(nx,ny,nz), cw(nx,ny,nz), cn(nx,ny,nz)
      real cs(nx,ny,nz), ct(nx,ny,nz), cb(nx,ny,nz)
c      real cne(nx,ny,nz), cnw(nx,ny,nz), cse(nx,ny,nz), csw(nx,ny,nz)
c      real cte(nx,ny,nz), ctw(nx,ny,nz), cbe(nx,ny,nz), cbw(nx,ny,nz)
c      real ctn(nx,ny,nz), cts(nx,ny,nz), cbn(nx,ny,nz), cbs(nx,ny,nz)
      real rhs(nx,ny,nz), u(nx,ny,nz), res(nx,ny,nz)
c
      real obrhs(nb), obres(nb)
      logical obdirichlet
c
c Local variables:
c
      integer i, j, k, it, itmax
      parameter (itmax=50000)
      real resnorm, tol, dlam, rjac, ome
      parameter (tol=1.e-14,rjac=0.96)
c
c First, see if solution is already close enough.
c
      resnorm = 0.
c$doacross local(k,j,i), reduction(resnorm)
      do k = 2,nz-1
         do j = 2,ny-1
            do i = 2,nx-1
               res(i,j,k) = cc(i,j,k)*u(i,j,k)+
     &              ce(i,j,k)*u(i+1,j,k)+
     &              cw(i,j,k)*u(i-1,j,k)+
     &              cn(i,j,k)*u(i,j+1,k)+
     &              cs(i,j,k)*u(i,j-1,k)+
     &              ct(i,j,k)*u(i,j,k+1)+
     &              cb(i,j,k)*u(i,j,k-1)
c     &              +cne(i,j,k)*u(i+1,j+1,k)+
c     &              cnw(i,j,k)*u(i-1,j+1,k)+
c     &              cse(i,j,k)*u(i+1,j-1,k)+
c     &              csw(i,j,k)*u(i-1,j-1,k)+
c     &              cte(i,j,k)*u(i+1,j,k+1)+
c     &              ctw(i,j,k)*u(i-1,j,k+1)+
c     &              cbe(i,j,k)*u(i+1,j,k-1)+
c     &              cbw(i,j,k)*u(i-1,j,k-1)+
c     &              ctn(i,j,k)*u(i,j+1,k+1)+
c     &              cts(i,j,k)*u(i,j-1,k+1)+
c     &              cbn(i,j,k)*u(i,j+1,k-1)+
c     &              cbs(i,j,k)*u(i,j-1,k-1)
     &              -rhs(i,j,k)
               resnorm = max(resnorm,abs(res(i,j,k)))
            enddo
         enddo
      enddo
c
c      write(*,*) 'sor: ', resnorm
      if (resnorm.lt.tol) return
c
c Relaxation loop:
c
      dlam = dx
      ome = 1.
c
      do it = 1,itmax
c$doacross local(k,j,i)
         do k = 2,nz-1
            do j = 2,ny-1
               do i = 2,nx-1
                  if (mod(i+j+k,2).eq.mod(it,2)) then
                     u(i,j,k) = u(i,j,k) - ome*res(i,j,k)/cc(i,j,k)
                  endif
               enddo
            enddo
         enddo
         call outerbc(nx,ny,nz,nb,it,dx,dy,dz,x,y,z,u,obrhs,obres,obdirichlet)
         if (it.eq.1) then
            ome = 1./(1.-0.5*rjac**2)
         else
            ome = 1./(1.-0.25*rjac**2*ome)
         endif
c
         resnorm = 0.
c$doacross local(k,j,i), reduction(resnorm)
         do k = 2,nz-1
            do j = 2,ny-1
               do i = 2,nx-1
                  res(i,j,k) = cc(i,j,k)*u(i,j,k)+
     &                 ce(i,j,k)*u(i+1,j,k)+
     &                 cw(i,j,k)*u(i-1,j,k)+
     &                 cn(i,j,k)*u(i,j+1,k)+
     &                 cs(i,j,k)*u(i,j-1,k)+
     &                 ct(i,j,k)*u(i,j,k+1)+
     &                 cb(i,j,k)*u(i,j,k-1)
c     &                 +cne(i,j,k)*u(i+1,j+1,k)+
c     &                 cnw(i,j,k)*u(i-1,j+1,k)+
c     &                 cse(i,j,k)*u(i+1,j-1,k)+
c     &                 csw(i,j,k)*u(i-1,j-1,k)+
c     &                 cte(i,j,k)*u(i+1,j,k+1)+
c     &                 ctw(i,j,k)*u(i-1,j,k+1)+
c     &                 cbe(i,j,k)*u(i+1,j,k-1)+
c     &                 cbw(i,j,k)*u(i-1,j,k-1)+
c     &                 ctn(i,j,k)*u(i,j+1,k+1)+
c     &                 cts(i,j,k)*u(i,j-1,k+1)+
c     &                 cbn(i,j,k)*u(i,j+1,k-1)+
c     &                 cbs(i,j,k)*u(i,j-1,k-1)
     &                 -rhs(i,j,k)
                  resnorm = max(resnorm,abs(res(i,j,k)))
               enddo
            enddo
         enddo
c
c         write(*,*) 'sor:', resnorm
         if (resnorm.lt.tol) then
            write(*,*) 'sor converged after ', it, ' iterations.'
            return
         endif
      enddo
c
      write(*,*) '**ERROR: sor failed to converge in ', itmax,
     &     ' iterations.'
c
      return
      end
