      program testprob4
c
c This program is a test of the solver and the Robin outer
c boundary condition. It tries to solve for the function
c exp(-x^2-y^2-z^2).
c
      implicit none
c
      integer nx, ny, nz, i, j, k, nb
      parameter (nx=21,ny=21,nz=21)
      parameter (nb=2*nx*ny+2*nx*nz+2*ny*nz-4*nx-4*ny-4*nz+8)
      real xmax, ymax, zmax, dx, dy, dz, maxerr, uan, cpui, cpuf
      parameter (xmax=1.,ymax=1.,zmax=1.)
      logical obdirichlet
      parameter (obdirichlet=.true.)
c
      real x(nx), y(ny), z(nz), r
c
      real cc(nx,ny,nz), ce(nx,ny,nz), cw(nx,ny,nz), cn(nx,ny,nz)
      real cs(nx,ny,nz), ct(nx,ny,nz), cb(nx,ny,nz)
c      real cne(nx,ny,nz), cnw(nx,ny,nz), cse(nx,ny,nz), csw(nx,ny,nz)
c      real cte(nx,ny,nz), ctw(nx,ny,nz), cbe(nx,ny,nz), cbw(nx,ny,nz)
c      real ctn(nx,ny,nz), cts(nx,ny,nz), cbn(nx,ny,nz), cbs(nx,ny,nz)
      real rhs(nx,ny,nz), u(nx,ny,nz), res(nx,ny,nz)
c
      real obrhs(nb), obres(nb)
c
c Begin timing:
c
      call start_timer(cpui)
c
c Initialize coordinates:
c
      dx = 2.*xmax/real(nx-1)
      dy = 2.*ymax/real(ny-1)
      dz = 2.*zmax/real(nz-1)
c
      call initcoord(nx,ny,nz,dx,dy,dz,x,y,z)
c
c Initialize stencil:
c
c$doacross local(k,j,i,r)
      do k = 1,nz
         do j = 1,ny
            do i = 1,nx
               cc(i,j,k) = -2.*dy**2*dz**2-2.*dx**2*dz**2-2.*dx**2*dy**2
               ce(i,j,k) = dy**2*dz**2
               cw(i,j,k) = dy**2*dz**2
               cn(i,j,k) = dx**2*dz**2
               cs(i,j,k) = dx**2*dz**2
               ct(i,j,k) = dx**2*dy**2
               cb(i,j,k) = dx**2*dy**2
c               cne(i,j,k) = 0.
c               cnw(i,j,k) = 0.
c               cse(i,j,k) = 0.
c               csw(i,j,k) = 0.
c               cte(i,j,k) = 0.
c               ctw(i,j,k) = 0.
c               cbe(i,j,k) = 0.
c               cbw(i,j,k) = 0.
c               ctn(i,j,k) = 0.
c               cts(i,j,k) = 0.
c               cbn(i,j,k) = 0.
c               cbs(i,j,k) = 0.
               r = sqrt(x(i)**2+y(j)**2+z(k)**2)
               rhs(i,j,k) = 2.*(2.*r**2-3.)*exp(-r**2)
               rhs(i,j,k) = dx**2*dy**2*dz**2*rhs(i,j,k)
            enddo
         enddo
      enddo
c
c Precondition:
c (Multiply Ax=B by the inverse of the diagonal part of A.)
c
c$doacross local(k,j,i)
      do k = 1,nz
         do j = 1,ny
            do i = 1,nx
               ce(i,j,k) = ce(i,j,k)/cc(i,j,k)
               cw(i,j,k) = cw(i,j,k)/cc(i,j,k)
               cn(i,j,k) = cn(i,j,k)/cc(i,j,k)
               cs(i,j,k) = cs(i,j,k)/cc(i,j,k)
               ct(i,j,k) = ct(i,j,k)/cc(i,j,k)
               cb(i,j,k) = cb(i,j,k)/cc(i,j,k)
c               cne(i,j,k) = cne(i,j,k)/cc(i,j,k)
c               cnw(i,j,k) = cnw(i,j,k)/cc(i,j,k)
c               cse(i,j,k) = cse(i,j,k)/cc(i,j,k)
c               csw(i,j,k) = csw(i,j,k)/cc(i,j,k)
c               cte(i,j,k) = cte(i,j,k)/cc(i,j,k)
c               ctw(i,j,k) = ctw(i,j,k)/cc(i,j,k)
c               cbe(i,j,k) = cbe(i,j,k)/cc(i,j,k)
c               cbw(i,j,k) = cbw(i,j,k)/cc(i,j,k)
c               ctn(i,j,k) = ctn(i,j,k)/cc(i,j,k)
c               cts(i,j,k) = cts(i,j,k)/cc(i,j,k)
c               cbn(i,j,k) = cbn(i,j,k)/cc(i,j,k)
c               cbs(i,j,k) = cbs(i,j,k)/cc(i,j,k)
               rhs(i,j,k) = rhs(i,j,k)/cc(i,j,k)
               cc(i,j,k) = 1.
            enddo
         enddo
      enddo
c
c Initial guess:
c
c$doacross local(k,j,i,r)
      do k = 1,nz
         do j = 1,ny
            do i = 1,nx
               r = sqrt(x(i)**2+y(j)**2+z(k)**2)
!               u(i,j,k) = exp(-r**2)
               u(i,j,k) = 0.0d0 
							 if (i.eq.1.or.j.eq.1.or.k.eq.1.or. 
     .             k.eq.nz.or.j.eq.ny.or.i.eq.nx) then
                 u(i,j,k) = exp(-r**2)
							 end if
            enddo
         enddo
      enddo
c
c Set the RHS of the outer boundary condition:
c (Do it in a subroutine since it's messy.)
c
      call obrhssub(nx,ny,nz,nb,x,y,z,obrhs)
c
      call sor(nx,ny,nz,nb,dx,dy,dz,x,y,z,cc,ce,cw,cn,cs,ct,cb,
c     &     cne,cnw,cse,csw,cte,ctw,cbe,cbw,ctn,cts,cbn,cbs,
     &     rhs,u,res,obrhs,obres,obdirichlet)
c
      open(10,file='u.n21.out',status='unknown',form='unformatted',
     &     access='sequential')
      write(10) u
      close(10)
c
      open(10,file='u.n21.xline',status='unknown')
      do i = 1,nx
         write(10,*) x(i), u(i,ny/2+1,nz/2+1)-exp(-x(i)**2)
      enddo
      close(10)
c
      maxerr = 0.
      do k = 1,nz
         do j = 1,ny
            do i = 1,nx
               r = sqrt(x(i)**2+y(j)**2+z(k)**2)
               uan = exp(-r**2)
               if (abs(u(i,j,k)-uan).gt.maxerr)
     &              maxerr = abs(u(i,j,k)-uan)
            enddo
         enddo
      enddo
      write(*,*) 'maxerr=', maxerr
c
c Finish timing and report results:
c
      call stop_timer(cpuf)
      call show_timing(cpuf-cpui)
c
      stop
      end
c
c******************************************************************************
c
      subroutine obrhssub(nx,ny,nz,nb,x,y,z,obrhs)
c
      implicit none
c
      integer nx, ny, nz, nb
      real x(nx), y(ny), z(nz), obrhs(nb)
c
c Local variables:
c
      integer i, j, k, nnb
      real r
c
      nnb = 0
      i = 1
         do j = 1,ny
            do k = 1,nz
               nnb = nnb+1
               r = sqrt(x(i)**2+y(j)**2+z(k)**2)
               obrhs(nnb) = (1.-2.*r**2)*exp(-r**2)
            enddo
         enddo

      do i = 2,nx-1
         j = 1
            do k = 1,nz
               nnb = nnb+1
               r = sqrt(x(i)**2+y(j)**2+z(k)**2)
               obrhs(nnb) = (1.-2.*r**2)*exp(-r**2)
            enddo

         do j = 2,ny-1
            k = 1
               nnb = nnb+1
               r = sqrt(x(i)**2+y(j)**2+z(k)**2)
               obrhs(nnb) = (1.-2.*r**2)*exp(-r**2)
            k = nz
               nnb = nnb+1
               r = sqrt(x(i)**2+y(j)**2+z(k)**2)
               obrhs(nnb) = (1.-2.*r**2)*exp(-r**2)
         enddo

         j = ny
            do k = 1,nz
               nnb = nnb+1
               r = sqrt(x(i)**2+y(j)**2+z(k)**2)
               obrhs(nnb) = (1.-2.*r**2)*exp(-r**2)
            enddo
      enddo

      i = nx
         do j = 1,ny
            do k = 1,nz
               nnb = nnb+1
               r = sqrt(x(i)**2+y(j)**2+z(k)**2)
               obrhs(nnb) = (1.-2.*r**2)*exp(-r**2)
            enddo
         enddo
c
      return
      end
