#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#define SPINS 100
#define TEMPS 500000
#define PRINT 1000
#define ANNEALS 50
#define EXPONENT 0.66
#define LOG 0.50

/* Globals */
double s[SPINS+2];
double h[SPINS+2];
double ms[SPINS+2];
double me;
double beta1;
double beta2;
double temp;
double min_energy,energy;
int accept,reject;

double kinetic(void){
  double k=0;
  int i;

  for(i=0;i<=SPINS;i++){
    k -= s[i] * s[i+1];
  }
  return(k);
}

double potential(void){
  double p=0;
  int i;

  for(i=0;i<=SPINS+1;i++){
    p += s[i]*h[i];
  }
  return(p);
}

void printout(void){
  int i;

  for(i=0;i<=(SPINS+1);i++){
    if(s[i] == -1) printf("-");
    else printf("+");
  }
  energy = beta1*kinetic() + beta2*potential();
  if ( energy < min_energy) {
		min_energy = energy;
	} 
  printf("\t%le\t%le\t%i\t%i\t%le\n",temp,beta1*kinetic() + beta2*potential(),accept,reject,beta1);
}

void print_field(void){
  int i;

  for(i=0;i<=SPINS+1;i++){
    printf("%6.1lf ",h[i]);
  }
  printf("\n");
}

void randomize_spin(void){
  int i;

  for(i=1;i<=SPINS;i++){
    if(drand48()>0.5){
      s[i] = 1;
    }else{
      s[i] = -1;
    }
  }
}

void randomize_field(void){
  int i;
  double factor;

  factor = 1/drand48();
  for(i=0;i<=SPINS+1;i++){
    h[i] = factor*2*(drand48()-0.5);
  }
}

double absmax(void){
  int i,maxi;
  double max=0;

  for(i=0;i<=SPINS+1;i++){
    if(fabs(h[i]) > max){
      max = fabs(h[i]);
      maxi = i;
    }
  }
  printf("max = %le, maxi = %i\n",max,maxi);
  return(max);
}

double average(){
  int i;
  double sum;

  sum = 0;
  for(i=0;i<=SPINS+1;i++){
    sum += fabs(h[i]);
  }
  sum /= (SPINS+2);
  printf("average = %le\n",sum);
  return(sum/(SPINS+2));
}

void find_min(){
  int i,j;
  double newe;

  s[0] = -1;
  s[SPINS+1] = 1;
  ms[0] = -1;
  ms[SPINS+1] = 1;
  for(i=1;i<=SPINS;i++){
    s[i] = 1;
    ms[i] = 1;
  }

  me = beta1*kinetic() + beta2*potential();

  for(i=1;i<=SPINS;i++){
    s[i] = -1;
    newe = beta1*kinetic() + beta2*potential();
    if(newe < me){
      me = newe;
      for(j=1;j<=SPINS;j++)
	ms[j] = s[j];
    }
  }

  printf("Minimum energy configuration is\n");
  for(i=0;i<=(SPINS+1);i++){
    if(ms[i] == -1) printf("-");
    else printf("+");
  }
  printf("\n");
}
 

void main(void){
  int i,j,k,l,m;
  int down=0;
  double olde, newe;  
  double tbase,mdp,adp;

  srand48(time(NULL));

  randomize_field();
  print_field();
  s[0] = -1;
  s[SPINS+1] = 1;
  mdp = absmax();
  adp = average();
  beta1 = SPINS/2*mdp;
  beta2 = 1;
  tbase = -beta1*(2*beta1+mdp)/log(LOG);
  find_min();
  min_energy = 100.0;


  for(m=1;m<=ANNEALS;m++){
    /* printf("\nANNEALING %i\n",m); */
    randomize_spin();
    accept = 0;
    reject = 0;
    temp = tbase;
    for(i=1;i<=TEMPS;i++){
      /*
      if(i%PRINT==0){
	 printout(); 
	accept = 0;
	reject = 0;
      }
      */
      
      /*    temp = tbase/pow(1+i,EXPONENT); */
      temp = exp(log(tbase) - EXPONENT*log(i));
      k = 1 + (int)((SPINS)*drand48());
      olde = s[k]*(-beta1*(s[k-1]+s[k+1]) + beta2*h[k]);
      newe = -olde;
      if(drand48() < exp(-(newe-olde)/temp)){
	s[k] *= -1;
	accept++;
      }else{
	reject++;
      }
    }
    printout();
  }
 printf ("Minimal energy configuration = %lf \n",min_energy);
}
