!/*@@
!   @file setgauge.f90
!@@*/

!/*@@
!   @routine setgauge
!   @author T.Haupt M.Huq S.Klasky 
!   @desc
!   driver for computing gauge
!   @enddesc
!   @calledby initialize
!   @calls f_lapse
!   @calls f_shift
!   @comment
!   uses module tensors
!   Modifications:
!   M.Huq    See lapse.f90 for additions to f_lapse
!
!   M.Huq    setgauge called from stub.f90. Added new set_gauge routine
!            which is called from setgauge and in addition from tstep.f90
!            More sensible interface added... need 3-metric, extrinsic 
!            curvature data, inverse metric etc.
!   @endcomment
!@@*/
subroutine setgauge(params)
use tensors
use parameters
real*8, dimension(100) :: params


! Pointers
type(GridFunction)                     :: alpha_,mask_
type(GridFunction), dimension(3)       :: beta_,beta_old_,tempv_
type(GridFunction), dimension(3,3)     :: g_,k_,gm1_,km1_,gup_,r_,tempt_
type(GridFunction), dimension(3,3,3)   :: c_,ca_
type(GridFunction), dimension(3,3,3,3) :: riemann_


INTERFACE
subroutine f_lapse(g1,g2,g3,g4,g5,g6,k1,k2,k3,k4,k5,k6,                &
                   gup1,gup2,gup3,gup4,gup5,gup6, r1,r2,r3,r4,r5,r6,   &
                   c1,c2,c3,c4,c5,c6,c7,c8,c9,c10,c11,c12,c13,c14,c15, &
                   c16,c17,c18,alpha,t,t2,t3,beta1, beta2, beta3, &
                   mask,params )
use tensors
real*8,dimension(:,:,:),target :: g1,g2,g3,g4,g5,g6
real*8,dimension(:,:,:),target :: k1,k2,k3,k4,k5,k6
real*8,dimension(:,:,:),target :: gup1,gup2,gup3,gup4,gup5,gup6
real*8,dimension(:,:,:),target :: r1,r2,r3,r4,r5,r6
real*8,dimension(:,:,:),target :: c1,c2,c3,c4,c5,c6,c7,c8,c9          
real*8,dimension(:,:,:),target :: c10,c11,c12,c13,c14,c15,c16,c17,c18
real*8,dimension(:,:,:),target :: beta1, beta2, beta3
real*8,dimension(:,:,:)        :: alpha,t,mask,t2,t3
real*8,dimension(100)          :: params
end subroutine !f_lapse

subroutine f_shift(b1,b2,b3,t1,t2,t3,t4,t5,t6,params)
use tensors
real*8, dimension(:,:,:),target :: b1,b2,b3,t1,t2,t3,t4,t5,t6 
real*8, dimension(100)          :: params
end subroutine 
subroutine set_gauge(params, g_, gup_, k_, c_, r_, t1, t2, t3, t4, & 
                     beta_, beta_old_, alpha, tempt_, mask)
  use tensors

  real*8, dimension(100) :: params
  type(GridFunction),dimension(3)      :: beta_, beta_old_
  type(GridFunction),dimension(3,3)    :: g_, gup_, k_, r_, tempt_
  type(GridFunction),dimension(3,3,3)  :: c_
  real*8,dimension(:,:,:)              :: t1, t2, t3, t4, mask, alpha
end subroutine


subroutine ginv(g_,detg_,mask_,gup_)
use tensors
implicit none
type(GridFunction),dimension(3,3)    :: g_,gup_
real*8, dimension(:,:,:)             :: detg_, mask_
real*8,parameter                 :: zero = 0.0d0
end subroutine

END INTERFACE

! Set up pointers

alpha_%d => ALPH
mask_%d => MASK

beta_(1)%d => BETA
beta_(2)%d => BETA1
beta_(3)%d => BETA2

beta_old_(1)%d => BETA_OLD
beta_old_(2)%d => BETA_OLD1
beta_old_(3)%d => BETA_OLD2

g_(1,1)%d => G
g_(1,2)%d => G1
g_(1,3)%d => G2
g_(2,1)%d => G1
g_(2,2)%d => G3
g_(2,3)%d => G4
g_(3,1)%d => G2
g_(3,2)%d => G4
g_(3,3)%d => G5

k_(1,1)%d => K
k_(1,2)%d => K1
k_(1,3)%d => K2
k_(2,1)%d => K1
k_(2,2)%d => K3
k_(2,3)%d => K4
k_(3,1)%d => K2
k_(3,2)%d => K4
k_(3,3)%d => K5

! MakeSymTensor gm1_
gm1_(1,1)%d => GM1
gm1_(1,2)%d => GM11
gm1_(1,3)%d => GM12
gm1_(2,1)%d => GM11
gm1_(2,2)%d => GM13
gm1_(2,3)%d => GM14
gm1_(3,1)%d => GM12
gm1_(3,2)%d => GM14
gm1_(3,3)%d => GM15

km1_(1,1)%d => KM1
km1_(1,2)%d => KM11
km1_(1,3)%d => KM12
km1_(2,1)%d => KM11
km1_(2,2)%d => KM13
km1_(2,3)%d => KM14
km1_(3,1)%d => KM12
km1_(3,2)%d => KM14
km1_(3,3)%d => KM15

gup_(1,1)%d => GUP
gup_(1,2)%d => GUP1
gup_(1,3)%d => GUP2
gup_(2,1)%d => GUP1
gup_(2,2)%d => GUP3
gup_(2,3)%d => GUP4
gup_(3,1)%d => GUP2
gup_(3,2)%d => GUP4
gup_(3,3)%d => GUP5

r_(1,1)%d => RICC
r_(1,2)%d => RICC1
r_(1,3)%d => RICC2
r_(2,1)%d => RICC1
r_(2,2)%d => RICC3
r_(2,3)%d => RICC4
r_(3,1)%d => RICC2
r_(3,2)%d => RICC4
r_(3,3)%d => RICC5

tempt_(1,1)%d => TEMPT
tempt_(1,2)%d => TEMPT1
tempt_(1,3)%d => TEMPT2
tempt_(2,1)%d => TEMPT1
tempt_(2,2)%d => TEMPT3
tempt_(2,3)%d => TEMPT4
tempt_(3,1)%d => TEMPT2
tempt_(3,2)%d => TEMPT4
tempt_(3,3)%d => TEMPT5

riemann_(1,1,1,1)%d => TEMPT
riemann_(1,1,1,2)%d => TEMPT
riemann_(1,1,1,3)%d => TEMPT
riemann_(1,1,2,1)%d => TEMPT
riemann_(1,1,2,2)%d => TEMPT
riemann_(1,1,2,3)%d => TEMPT
riemann_(1,1,3,1)%d => TEMPT
riemann_(1,1,3,2)%d => TEMPT
riemann_(1,1,3,3)%d => TEMPT
riemann_(1,2,1,1)%d => TEMPT
riemann_(1,2,1,2)%d => TEMPT
riemann_(1,2,1,3)%d => TEMPT1
riemann_(1,2,2,1)%d => TEMPT
riemann_(1,2,2,2)%d => TEMPT
riemann_(1,2,2,3)%d => TEMPT2
riemann_(1,2,3,1)%d => TEMPT1
riemann_(1,2,3,2)%d => TEMPT2
riemann_(1,2,3,3)%d => TEMPT
riemann_(1,3,1,1)%d => TEMPT
riemann_(1,3,1,2)%d => TEMPT1
riemann_(1,3,1,3)%d => TEMPT3
riemann_(1,3,2,1)%d => TEMPT1
riemann_(1,3,2,2)%d => TEMPT
riemann_(1,3,2,3)%d => TEMPT4
riemann_(1,3,3,1)%d => TEMPT3
riemann_(1,3,3,2)%d => TEMPT4
riemann_(1,3,3,3)%d => TEMPT
riemann_(2,1,1,1)%d => TEMPT
riemann_(2,1,1,2)%d => TEMPT
riemann_(2,1,1,3)%d => TEMPT1
riemann_(2,1,2,1)%d => TEMPT
riemann_(2,1,2,2)%d => TEMPT
riemann_(2,1,2,3)%d => TEMPT2
riemann_(2,1,3,1)%d => TEMPT1
riemann_(2,1,3,2)%d => TEMPT2
riemann_(2,1,3,3)%d => TEMPT
riemann_(2,2,1,1)%d => TEMPT
riemann_(2,2,1,2)%d => TEMPT
riemann_(2,2,1,3)%d => TEMPT
riemann_(2,2,2,1)%d => TEMPT
riemann_(2,2,2,2)%d => TEMPT
riemann_(2,2,2,3)%d => TEMPT
riemann_(2,2,3,1)%d => TEMPT
riemann_(2,2,3,2)%d => TEMPT
riemann_(2,2,3,3)%d => TEMPT
riemann_(2,3,1,1)%d => TEMPT
riemann_(2,3,1,2)%d => TEMPT2
riemann_(2,3,1,3)%d => TEMPT4
riemann_(2,3,2,1)%d => TEMPT2
riemann_(2,3,2,2)%d => TEMPT
riemann_(2,3,2,3)%d => TEMPT5
riemann_(2,3,3,1)%d => TEMPT4
riemann_(2,3,3,2)%d => TEMPT5
riemann_(2,3,3,3)%d => TEMPT
riemann_(3,1,1,1)%d => TEMPT
riemann_(3,1,1,2)%d => TEMPT1
riemann_(3,1,1,3)%d => TEMPT3
riemann_(3,1,2,1)%d => TEMPT1 
riemann_(3,1,2,2)%d => TEMPT
riemann_(3,1,2,3)%d => TEMPT4
riemann_(3,1,3,1)%d => TEMPT3
riemann_(3,1,3,2)%d => TEMPT4
riemann_(3,1,3,3)%d => TEMPT
riemann_(3,2,1,1)%d => TEMPT
riemann_(3,2,1,2)%d => TEMPT2
riemann_(3,2,1,3)%d => TEMPT4
riemann_(3,2,2,1)%d => TEMPT2
riemann_(3,2,2,2)%d => TEMPT
riemann_(3,2,2,3)%d => TEMPT5
riemann_(3,2,3,1)%d => TEMPT4
riemann_(3,2,3,2)%d => TEMPT5
riemann_(3,2,3,3)%d => TEMPT
riemann_(3,3,1,1)%d => TEMPT
riemann_(3,3,1,2)%d => TEMPT
riemann_(3,3,1,3)%d => TEMPT
riemann_(3,3,2,1)%d => TEMPT
riemann_(3,3,2,2)%d => TEMPT
riemann_(3,3,2,3)%d => TEMPT
riemann_(3,3,3,1)%d => TEMPT
riemann_(3,3,3,2)%d => TEMPT
riemann_(3,3,3,3)%d => TEMPT


c_(1,1,1)%d => CHR
c_(1,1,2)%d => CHR1
c_(1,1,3)%d => CHR2
c_(1,2,1)%d => CHR1
c_(1,2,2)%d => CHR3
c_(1,2,3)%d => CHR4
c_(1,3,1)%d => CHR2
c_(1,3,2)%d => CHR4
c_(1,3,3)%d => CHR5
c_(2,1,1)%d => CHR6
c_(2,1,2)%d => CHR7
c_(2,1,3)%d => CHR8
c_(2,2,1)%d => CHR7
c_(2,2,2)%d => CHR9
c_(2,2,3)%d => CHR10
c_(2,3,1)%d => CHR8
c_(2,3,2)%d => CHR10
c_(2,3,3)%d => CHR11
c_(3,1,1)%d => CHR12
c_(3,1,2)%d => CHR13
c_(3,1,3)%d => CHR14
c_(3,2,1)%d => CHR13
c_(3,2,2)%d => CHR15
c_(3,2,3)%d => CHR16
c_(3,3,1)%d => CHR14
c_(3,3,2)%d => CHR16
c_(3,3,3)%d => CHR17

ca_(1,1,1)%d => CHRA
ca_(1,1,2)%d => CHRA1
ca_(1,1,3)%d => CHRA2
ca_(1,2,1)%d => CHRA1
ca_(1,2,2)%d => CHRA3
ca_(1,2,3)%d => CHRA4
ca_(1,3,1)%d => CHRA2
ca_(1,3,2)%d => CHRA4
ca_(1,3,3)%d => CHRA5
ca_(2,1,1)%d => CHRA6
ca_(2,1,2)%d => CHRA7
ca_(2,1,3)%d => CHRA8
ca_(2,2,1)%d => CHRA7
ca_(2,2,2)%d => CHRA9
ca_(2,2,3)%d => CHRA10
ca_(2,3,1)%d => CHRA8
ca_(2,3,2)%d => CHRA10
ca_(2,3,3)%d => CHRA11
ca_(3,1,1)%d => CHRA12
ca_(3,1,2)%d => CHRA13
ca_(3,1,3)%d => CHRA14
ca_(3,2,1)%d => CHRA13
ca_(3,2,2)%d => CHRA15
ca_(3,2,3)%d => CHRA16
ca_(3,3,1)%d => CHRA14
ca_(3,3,2)%d => CHRA16
ca_(3,3,3)%d => CHRA17
!
! Call new set_gauge
  call set_gauge(params, g_, gup_, k_, c_, r_,chra14, chra15, chra16,chra17, &
                     beta_, beta_old_, alpha_%d, tempt_, mask_%d)

return

end subroutine !setgauge
!-----------------------------------------------------------------------
! Routine: set_gauge
! Author:  M. Huq
! Description: Frontend routine to set coordinate conditions

subroutine set_gauge(params, g_, gup_, k_, c_, r_, t1, t2, t3, t4, & 
                     beta_, beta_old_, alpha, tempt_, mask)
  use tensors

  real*8, dimension(100) :: params
  type(GridFunction),dimension(3)      :: beta_, beta_old_
  type(GridFunction),dimension(3,3)    :: g_, gup_, k_, r_, tempt_
  type(GridFunction),dimension(3,3,3)  :: c_
  real*8,dimension(:,:,:)              :: t1, t2, t3, t4, alpha, mask

  interface
    subroutine setshift(params, g_, gup_, k_, c_, r_, t1, t2, t3, t4, &
                     beta_, alpha, tempt_, mask)
      use tensors

      real*8, dimension(100) :: params
      type(GridFunction),dimension(3)      :: beta_
      type(GridFunction),dimension(3,3)    :: g_, gup_, k_, r_, tempt_
      type(GridFunction),dimension(3,3,3)  :: c_
      real*8,dimension(:,:,:)              :: t1, t2, t3, t4, mask, alpha
    end subroutine

    subroutine setlapse(params, g_, gup_, k_, c_, r_, t1, t2, t3, t4, &
                    beta_, alpha, tempt_, mask)
      use tensors

      real*8, dimension(100) :: params
      type(GridFunction),dimension(3)      :: beta_
      type(GridFunction),dimension(3,3)    :: g_, gup_, k_, r_, tempt_
      type(GridFunction),dimension(3,3,3)  :: c_
      real*8,dimension(:,:,:)              :: t1, t2, t3, t4, mask, alpha

    end subroutine
subroutine ginv(g_,detg_,mask_,gup_)
use tensors
implicit none
type(GridFunction),dimension(3,3)    :: g_,gup_
real*8, dimension(:,:,:)             :: detg_, mask_
real*8,parameter                 :: zero = 0.0d0
end subroutine


  end interface


  ! Copy current beta to old...
  beta_old_(1)%d = beta_(1)%d
  beta_old_(2)%d = beta_(2)%d
  beta_old_(3)%d = beta_(3)%d


  ! Need the inverse metric computed here....
  if (params(90).ne.(1.0d0)) call ginv(g_, t1, mask, gup_)

  call setshift(params, g_, gup_, k_, c_, r_, t1, t2, t3, t4, &
                beta_, alpha, tempt_, mask)

  call setlapse(params, g_, gup_, k_, c_, r_, t1, t2, t3, t4, &
                    beta_, alpha, tempt_, mask)

  return
end
!-----------------------------------------------------------------------
