/* binio.h */

/*
 Orignal authors of this file are
    Bill Hibbard, Brian Paul, Dave Santek, and Andre Battaiola.
*/


/*
 * Functions to do binary I/O of floats, ints, etc. with byte swapping
 * as needed.
 */

#ifndef BINIO_H
#define BINIO_H

/* Include files which define SEEK_SET, O_RD_ONLY, etc. */
/* and prototype open(), close(), lseek(), etc. */
#include <unistd.h>
#include <fcntl.h>

#ifdef __STDC__
#  define PROTO(x) x
#else
#  define PROTO(x) ()	
#endif

extern void flip4 PROTO(( unsigned int *src, unsigned int *dest, int n ));
extern void flip2 PROTO(( unsigned short *src, unsigned short *dest, int n ));

#ifdef cray
  extern void cray_to_ieee_array( /* dest, source, n */ );
  extern void ieee_to_cray_array( /* dest, source, n */ );
#endif

/******************************************************/
/*                    Read Functions                  */
/******************************************************/
extern int read_bytes PROTO(( int f, void *b, int n ));
extern int read_int2_array PROTO(( int f, short *iarray, int n ));
extern int read_uint2_array PROTO(( int f, unsigned short *iarray, int n ));
extern int read_int4 PROTO(( int f, int *i ));
extern int read_int4_array PROTO(( int f, int *iarray, int n ));
extern int read_float4 PROTO(( int f, float *x ));
extern int read_float4_array PROTO(( int f, float *x, int n ));
extern int read_block PROTO(( int f, void *data, int elements, int elsize ));

/*******************************************************/
/*               Write Functions                       */
/*******************************************************/
extern int write_bytes PROTO(( int f, void *b, int n ));
extern int write_int2_array PROTO(( int f, short *iarray, int n ));
extern int write_uint2_array PROTO(( int f, unsigned short *iarray, int n ));
extern int write_int4 PROTO(( int f, int i ));
extern int write_int4_array PROTO(( int f, int *iarray, int n ));
extern int write_float4 PROTO(( int f, float x ));
extern int write_float4_array PROTO(( int f, float *x, int n ));
extern int write_block PROTO(( int f, void *data, int elements, int elsize ));

#undef PROTO

#endif
