! This program will read in a 2dhdf file an then output a 2d slice to Scivis
! Scott A. Klasky


program hdf2ser
implicit none
character*128                             :: name,fname   !name of hdf file
integer                                  :: level,lev    !iteration number
integer,dimension(:),allocatable         :: shp          !shape
character*128,dimension(1)                :: cnames       !names of x,y,z axis 
integer                                  :: rank         !rank of grid function
real*8                                   :: time         !Time 
real*8,dimension(:), allocatable         :: otime        !Output Time Array

real*8,dimension(:),allocatable          :: coords       !coordinate values
real*8,dimension(:),allocatable        :: gf           !Grid function
integer                                  :: orank        !Output rank
real*8,dimension(:),allocatable          :: ogf          !1D gf output function
real*8,dimension(:),allocatable          :: ogx1d !1D gf output x,y funcs
integer,dimension(:),allocatable         :: nx,nxy
integer                                  :: error        !Error condition
integer                                  :: iter,siter,eiter   !Time iterations
integer                                  :: line_type, xvalue, yvalue
integer                                  :: i,j,mem_loc
integer                                  :: java_mser2d,java_mser,java_ser2d
integer                                  :: gft_read_rank, gft_read_shape
integer                                  :: gft_read_full,indlnb
integer                                  :: elevel, irlevel, stride


integer                                  :: i4arg,iargc
real*8                                   :: r8arg
character*128                            :: sarg

interface
   integer function get_total_iters(eiter,fname,shp,cnames, &
                                    rank,time,coords,gf)
   implicit none
   character*128                             :: fname   !name of hdf file
   integer,dimension(1)                     :: shp
   integer                                  :: rank, eiter
   character*128,dimension(rank)             :: cnames
   real*8                                   :: time
   real*8,dimension(shp(1))    :: coords
   real*8,dimension(shp(1))   :: gf
	end function get_total_iters

end interface



	if (iargc().lt.1) then
		write(0,*) 'Usage: hdf2jser1d filename start_iter end_iter stride'
		stop
	end if

   fname = sarg(1," ")          ! no default file name
	siter  = i4arg(2,1)              !default is 1 time iteration
	eiter  = i4arg(3,1)              !default is 1 time iteration
	stride  = i4arg(4,1)             !default is 1 for the stride 
	rank = 1
   orank = 1

	allocate(shp(rank)) 
	error = gft_read_shape(fname(1:indlnb(fname)),1,shp)
   if (error.eq.0) then
      write(6,*) 'hdf2ser1d: Can not read shape for file:',fname(1:indlnb(fname))
      stop
   end if
	write(6,*) 'shape=',shp(1)

	allocate(gf(shp(1)))
   allocate(coords(shp(1)))

	level = get_total_iters(eiter,fname,shp,cnames,rank,time,coords,gf)

	if (eiter.gt.level) eiter = level


	if (stride.eq.0) stride = 1
	iter = int((eiter - siter + 1  +(stride-1))/stride)

   allocate(ogf(iter*shp(1)))
   allocate(ogx1d(iter*shp(1)) )
   allocate(otime(iter),nx(iter))


	irlevel = 0  ! this is the integer level count 
   mem_loc = 0 
	do level = 1 , eiter
		lev = level
      error = gft_read_full(fname(1:indlnb(fname)),lev,shp,cnames,rank,time,coords,gf)
		if (error.eq.0) then
			write(6,*) 'only could read in',level,'time steps'
			stop
		end if
		if (mod(level-1,stride).ne.0) goto 100 ! we don't include this 
		if (lev.lt.siter) goto 100  !we don't include this in the output
		irlevel = irlevel + 1
		otime(irlevel) = time
		nx(irlevel) = shp(1)
		do i = 1, shp(1)
         mem_loc = mem_loc + 1
			ogf(mem_loc) = gf(i) 
			ogx1d(mem_loc) = coords(i)
		end do

100   continue
	end do

   error = java_mser(fname(1:indlnb(fname)),iter,otime,          &
                       ogx1d,ogf,nx)


	stop
	end

	integer function get_total_iters(eiter,fname,shp,cnames,rank,time,coords,gf)
	implicit none
   character*128                           :: fname   !name of hdf file
	integer,dimension(1)                     :: shp
   integer                                  :: rank
   character*32,dimension(rank)             :: cnames	
	real*8                                   :: time
	real*8,dimension(shp(1))                 :: coords
	real*8,dimension(shp(1))                 :: gf
	integer                                  :: gft_close_all, gft_read_full
	integer                                  :: indlnb, lev

	integer                                  :: iter, error, eiter
	error = gft_close_all()
	error = 1
	iter = 0

	cnames(1) = 'x'
	lev = 1
	do iter = 1 , 1000000 
      error = gft_read_full(fname(1:indlnb(fname)),lev,shp,cnames,rank,time,coords,gf)
		lev = lev + 1
		if (error.eq.0.or.iter.ge.eiter) goto 100
   end do

100 continue

	error = gft_close_all()

	get_total_iters = iter

	return
	end function get_total_iters
