! This program will read in a 3dhdf file an then output a 2d slice to Scivis
! Scott A. Klasky


program hdf2ser
implicit none
character*32                             :: name,fname   !name of hdf file
integer                                  :: level,lev    !iteration number
integer,dimension(:),allocatable         :: shp          !shape
character*32,dimension(:),allocatable    :: cnames       !names of x,y,z axis 
integer                                  :: rank         !rank of grid function
real*8                                   :: time         !Time 
real*8,dimension(:), allocatable         :: otime        !Output Time Array

real*8,dimension(:),allocatable          :: coords       !coordinate values
real*8,dimension(:,:,:),allocatable      :: gf           !Grid function
integer                                  :: orank        !Output rank
real*8,dimension(:),allocatable          :: ogf          !1D gf output function
real*8,dimension(:),allocatable          :: ogx1d,ogy1d  !1D gf output x,y funcs
integer,dimension(:),allocatable         :: nx,nxy
integer                                  :: error        !Error condition
integer                                  :: iter,siter,eiter   !Time iterations
integer                                  :: line_type, xvalue, yvalue
integer                                  :: i,j,mem_loc
integer                                  :: java_mser2d,java_mser,java_ser2d
integer                                  :: gft_read_rank, gft_read_shape
integer                                  :: gft_read_full,indlnb
integer                                  :: elevel, irlevel


integer                                  :: i4arg,iargc
real*8                                   :: r8arg
character*128                            :: sarg

interface
   subroutine check_rank_shape(orank,rank,line_type,xvalue,yvalue,shp)
   implicit none
   integer                         :: line_type, rank
   integer,dimension(rank)         :: shp
   integer                         :: orank, xvalue, yvalue
	end subroutine check_rank_shape

   integer function get_total_iters(fname,shp,cnames,rank,time,coords,gf)
   implicit none
   character*32                             :: fname   !name of hdf file
   integer,dimension(3)                     :: shp
   integer                                  :: rank
   character*32,dimension(rank)             :: cnames
   real*8                                   :: time
   real*8,dimension(shp(1)+shp(2)+shp(3))    :: coords
   real*8,dimension(shp(1),shp(2),shp(3))   :: gf
	end function get_total_iters

end interface



	if (iargc().lt.1) then
		write(0,*) 'Usage: hdf2jser filename out_rank line_type(1:3) start_iter end_iter rank xvalue <yvalue>'
		stop
	end if

   fname = sarg(1," ")          ! no default file name
   orank = i4arg(2,2)           ! default output rank = 2
   line_type = i4arg(3,1)       ! default is (x vary or z const)
	
! Now read in the shape

	siter  = i4arg(4,1)              !default is 1 time iteration
	eiter  = i4arg(5,1)              !default is 1 time iteration
   rank  = i4arg(6,3)               ! default input  rank = 3
	allocate(shp(rank),cnames(rank))
	error = gft_read_shape(fname(1:indlnb(fname)),1,shp)
   if (error.eq.0) then
      write(6,*) 'hdf2ser: Can not read shape for file:',fname(1:indlnb(fname))
      stop
   end if
	write(6,*) 'shape=',shp(1),shp(2),shp(3)
   xvalue = i4arg(7,int(shp(1)/2))  !default x value
   yvalue = i4arg(8,int(shp(1)/2))  !default y value

	if (line_type.lt.1.or.line_type.gt.3) then
		write(6,*) 'Invalid line type:',line_type
		stop
	end if

!	error = gft_read_rank(fname(1:indlnb(fname)),1,rank)
!  this read rank does not exisits.....

	if (rank.gt.3.or.orank.gt.rank) then
		write(6,*) 'Problem with the rank,orank:',rank,orank
		stop
	end if


! now check to make sure the user is not asking for something out of the shape
	call check_rank_shape(orank,rank,line_type,xvalue,yvalue,shp)

!  read in total number of iterations in input file
	allocate(gf(shp(1),shp(2),shp(3)))
   allocate(coords(shp(1)+shp(2)+shp(3)))

	level = get_total_iters(fname,shp,cnames,rank,time,coords,gf)

	if (eiter.gt.level) eiter = level


	write(6,*) 'using iters: ', siter, ':', eiter

	iter = eiter - siter + 1  ! total number of iterations.....

! Now allocate the memory

! now get the memory for the output function (special for 2d to work right
   if (orank.eq.1) then
      allocate(ogf(iter*shp(line_type)))
      allocate(ogx1d(iter*shp(line_type)))
   else if (orank.eq.2) then
      if (line_type.eq.1) then
         allocate(ogf(1*shp(1)*shp(2)))
         allocate(ogx1d(1*shp(1)*shp(2)))
         allocate(ogy1d(1*shp(1)*shp(2)))
      else if (line_type.eq.2) then
         allocate(ogf(1*shp(1)*shp(3)))
         allocate(ogx1d(1*shp(1)*shp(3)))
         allocate(ogy1d(1*shp(1)*shp(3)))
      else
         allocate(ogf(1*shp(2)*shp(3)))
         allocate(ogx1d(1*shp(2)*shp(3)))
         allocate(ogy1d(1*shp(2)*shp(3)))
      end if
   else
      write(6,*) 'output rank is not supported:',orank
      stop
   end if
   allocate(otime(iter),nx(iter),nxy(iter))





	irlevel = 0  ! this is the integer level count 
   mem_loc = 0 
	do level = 1 , eiter
		lev = level
      error = gft_read_full(fname,lev,shp,cnames,rank,time,coords,gf)
		if (error.eq.0) then
			write(6,*) 'only could read in',level,'time steps'
			stop
		end if

		if (lev.lt.siter) goto 100  !we don't include this in the output
		irlevel = irlevel + 1
		otime(irlevel) = time

		if (orank.eq.2) then
			if (line_type.eq.1) then
				nxy(irlevel) = shp(1)*shp(2)
				do i = 1, shp(1)
					do j = 1 , shp(2)
						mem_loc = mem_loc + 1
						ogf(mem_loc) = gf(i,j,xvalue)
						ogx1d(mem_loc) = coords(i)
						ogy1d(mem_loc) = coords(j)
					end do
				end do
			else if (line_type.eq.2) then
				nxy(irlevel) = shp(1)*shp(3)
				do i = 1, shp(1)
					do j = 1 , shp(3)
						mem_loc = mem_loc + 1
						ogf(mem_loc) = gf(i,xvalue,j)
						ogx1d(mem_loc) = coords(i)
						ogy1d(mem_loc) = coords(j)
					end do
				end do
			else  !line_type=3
				nxy(irlevel) = shp(2)*shp(3)
            do i = 1, shp(2)
               do j = 1 , shp(3)
                  mem_loc = mem_loc + 1                     
                  ogf(mem_loc) = gf(xvalue,i,j)
						ogx1d(mem_loc) = coords(i)
						ogy1d(mem_loc) = coords(j)
               end do
            end do
			end if
		else !orank.eq.1
			if(line_type.eq.1) then
				nx(irlevel) = shp(1)
				do i = 1, shp(1)
					mem_loc = mem_loc + 1
					ogf(mem_loc) = gf(i,xvalue,yvalue)
					ogx1d(mem_loc) = coords(i)
				end do
			else if (line_type.eq.2) then
				nx(irlevel) = shp(2)
				do i = 1, shp(2)
					mem_loc = mem_loc + 1
					ogf(mem_loc) = gf(xvalue,i,yvalue)
					ogx1d(mem_loc) = coords(i)
				end do
			else !line_type=3
				nx(irlevel) = shp(3)
				do i = 1, shp(3)
					mem_loc = mem_loc + 1
					ogf(mem_loc) = gf(xvalue,yvalue,i)
					ogx1d(mem_loc) = coords(i)
				end do
			end if
		end if
100   continue
	end do

! now we just output this into java_mser2d

	if (orank.eq.2) then

	   if (line_type.eq.1) then
	      error = java_mser2d(fname(1:indlnb(fname)),iter,otime,          &
                             ogx1d,ogy1d,ogf,nxy)
	   else if (line_type.eq.2) then
	      error = java_mser2d(fname(1:indlnb(fname)),iter,otime,          &
                             ogx1d,ogy1d,ogf,nxy) 
	   else 
	      error = java_mser2d(fname(1:indlnb(fname)),iter,otime,          &
                             ogx1d,ogy1d,ogf,nxy) 
	   end if

	else 
	   if (line_type.eq.1) then
	      error = java_mser(fname(1:indlnb(fname)),iter,otime,          &
                             ogx1d,ogf,nx) 
	   else if (line_type.eq.2) then
	      error = java_mser(fname(1:indlnb(fname)),iter,otime,          &
                             ogx1d,ogf,nx) 
	   else 
	      error = java_mser(fname(1:indlnb(fname)),iter,otime,          &
                             ogx1d,ogf,nx) 
		end if

	end if

	stop
	end

   subroutine check_rank_shape(orank,rank,line_type,xvalue,yvalue,shp)
   implicit none
   integer                         :: line_type, rank
   integer,dimension(rank)         :: shp
   integer                         :: orank, xvalue, yvalue

   if (orank.eq.1) then
      if (line_type.eq.1) then !xvalue = y, yvalue=z
         if (xvalue.lt.1.or.xvalue.gt.shp(2).or.                              &
             yvalue.lt.1.or.yvalue.gt.shp(3)) then
            write(6,*) 'input values are not in the bounding box',xvalue,yvalue
            stop
         end if
      else if (line_type.eq.2) then !xvalue=x,yvalue=z
         if (xvalue.lt.1.or.xvalue.gt.shp(1).or.                              &
             yvalue.lt.1.or.yvalue.gt.shp(3)) then
            write(6,*) 'input values are not in the bounding box',xvalue,yvalue
            stop
         end if
      else
         if (xvalue.lt.1.or.xvalue.gt.shp(1).or.                              &
             yvalue.lt.1.or.yvalue.gt.shp(2)) then
            write(6,*) 'input values are not in the bounding box',xvalue,yvalue
            stop
         end if
      end if
   else if (orank.eq.2) then
      if (line_type.eq.1) then !xvalue = zvalue
         if (xvalue.lt.1.or.xvalue.gt.shp(3)) then
            write(6,*) 'input values are not in the bounding box',xvalue
            stop
         end if
      else if (line_type.eq.2) then !xvalue=yvalue
         if (xvalue.lt.1.or.xvalue.gt.shp(2)) then
            write(6,*) 'input values are not in the bounding box',xvalue
            stop
         end if
      else if (line_type.eq.3) then !xvalue=yvalue
         if (xvalue.lt.1.or.xvalue.gt.shp(1)) then
            write(6,*) 'input values are not in the bounding box',xvalue
            stop
         end if
      end if
   end if

   end subroutine check_rank_shape

	integer function get_total_iters(fname,shp,cnames,rank,time,coords,gf)
	implicit none
   character*32                             :: fname   !name of hdf file
	integer,dimension(3)                     :: shp
   integer                                  :: rank
   character*32,dimension(rank)             :: cnames	
	real*8                                   :: time
	real*8,dimension(shp(1)+shp(2)+shp(3))    :: coords
	real*8,dimension(shp(1),shp(2),shp(3))   :: gf
	integer                                  :: gft_close_all, gft_read_full

	integer                                  :: iter, error, level

	do while(error.ne.0)
      error = gft_read_full(fname,level,shp,cnames,rank,time,coords,gf)
      iter = iter + 1
   end do


	error = gft_close_all()

	get_total_iters = iter

	return
	end function get_total_iters
