program inttest
  implicit none
  integer :: nx,ny,nz
  real*8,dimension(:,:,:),allocatable :: mask,u,r
  integer :: n
  real*8,dimension(:,:,:),allocatable :: storage
  real*8,dimension(:),allocatable :: x,y,z
  real*8,dimension(:,:),allocatable :: intlist
  real*8,dimension(:),allocatable :: solution,exact,error

  real*8,parameter  :: xmax=1.0
  real*8 :: radius,mask_radius,h
  integer :: i,j,k,ltheta,lphi,lc
  real*8 :: theta,phi
  real*8 :: pi
  real*8  :: ssum, l1norm
  ! Note that the interpolation sphere is n*n points
  pi = 4.0*atan(1.0)
  open(22,file='inttest.dat')
  read(22,*) nx,n
  read(22,*) radius,mask_radius
  close(22)

  allocate(mask(nx,nx,nx),u(nx,nx,nx),r(nx,nx,nx))
  allocate(storage(10,14,n*n))
  allocate(x(nx),y(nx),z(nx))
  allocate(intlist(n*n,3))
  allocate(solution(n*n),exact(n*n),error(n*n))

  h = (2*xmax)/(nx-1)

  do i = 1, nx
     x(i) = -xmax + h * (i-1)
  end do
  y(:) = x(:)
  z(:) = x(:)

  do k = 1 , nx
     do j = 1, nx
        do i = 1, nx
           !u(i,j,k) = (x(i)**2+y(j)**2+z(k)**2)
           r(i,j,k) = sqrt(x(i)*x(i)+y(j)*y(j)+z(k)*z(k))
           u(i,j,k) = x(i)**4
        end do
     end do
  end do

  where (r<mask_radius) 
     mask(:,:,:)=0.0
     u(:,:,:) = -100000.0
  elsewhere
     mask(:,:,:) = 1.0
  end where

  lc = 0
  do lphi = 1 , n
     phi = pi/n * (lphi-1)
     do ltheta = 1, n
        lc = lc + 1
        theta = 2*pi/n * (lphi-1)
        intlist(lc,1) = radius*cos(theta)*cos(phi)
        intlist(lc,2) = radius*sin(theta)*cos(phi)
        intlist(lc,3) = radius*sin(phi)
        !exact(lc) = (intlist(lc,1)**2 + intlist(lc,2)**2 + intlist(lc,3)**2)
        !exact(lc) = 1.0/(0.1+sqrt(intlist(lc,1)**2+intlist(lc,2)**2+intlist(lc,3)**2))
        exact(lc) = intlist(lc,1)**4
     end do
  end do
  ! just use the 26th point for this experiment

  call int3do3(nx,nx,nx,x,y,z,n*n,intlist,mask,storage)
  call eval3do3(nx,nx,nx,n*n,storage,1,u,solution)

  
  error(:) = solution(:) - exact(:)
  ssum = sum(abs(error))
  l1norm = ssum/n/n
  write(6,*) h,l1norm
  open(66,file='inttest.out')
  do i = 1 , n*n
     !if (abs(error(i))>1.0d-12) then
     write(66,100) i,intlist(i,1),intlist(i,2),intlist(i,3),solution(i),exact(i),error(i)
     !end if
  end do
  close(66)
100 format(i5,1P,6e12.2)

  end program
